@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -33,6 +33,7 @@
 #include "segments.h"
 #include "ways.h"
 #include "relations.h"
+#include "errorlog.h"
 
 #include "files.h"
 #include "visualiser.h"
@@ -45,6 +46,7 @@ static void print_node(Nodes *nodes,index_t item);
 static void print_segment(Segments *segments,index_t item);
 static void print_way(Ways *ways,index_t item);
 static void print_turn_relation(Relations *relations,index_t item,Segments *segments,Nodes *nodes);
+static void print_errorlog(ErrorLogs *errorlogs,index_t item);
 
 static void print_head_osm(int coordcount,double latmin,double latmax,double lonmin,double lonmax);
 static void print_region_osm(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,
@@ -54,6 +56,11 @@ static void print_segment_osm(Segments *segments,index_t item,Ways *ways);
 static void print_turn_relation_osm(Relations *relations,index_t item,Segments *segments,Nodes *nodes);
 static void print_tail_osm(void);
 
+static void print_node_visualiser(Nodes *nodes,index_t item);
+static void print_segment_visualiser(Segments *segments,index_t item,Ways *ways);
+static void print_turn_relation_visualiser(Relations *relations,index_t item,Segments *segments,Nodes *nodes);
+static void print_errorlog_visualiser(ErrorLogs *errorlogs,index_t item);
+
 static char *RFC822Date(time_t t);
 
 static void print_usage(int detail,const char *argerr,const char *err);
@@ -69,15 +76,17 @@ int main(int argc,char** argv)
  Segments *OSMSegments;
  Ways     *OSMWays;
  Relations*OSMRelations;
+ ErrorLogs*OSMErrorLogs=NULL;
  int       arg;
  char     *dirname=NULL,*prefix=NULL;
- char     *nodes_filename,*segments_filename,*ways_filename,*relations_filename;
+ char     *nodes_filename,*segments_filename,*ways_filename,*relations_filename,*errorlogs_filename;
  int       option_statistics=0;
  int       option_visualiser=0,coordcount=0;
  double    latmin=0,latmax=0,lonmin=0,lonmax=0;
  char     *option_data=NULL;
  int       option_dump=0;
  int       option_dump_osm=0,option_no_super=0;
+ int       option_dump_visualiser=0;
 
  /* Parse the command line arguments */
 
@@ -97,6 +106,8 @@ int main(int argc,char** argv)
        option_dump=1;
     else if(!strcmp(argv[arg],"--dump-osm"))
        option_dump_osm=1;
+    else if(!strcmp(argv[arg],"--dump-visualiser"))
+       option_dump_visualiser=1;
     else if(!strncmp(argv[arg],"--latmin",8) && argv[arg][8]=='=')
       {latmin=degrees_to_radians(atof(&argv[arg][9]));coordcount++;}
     else if(!strncmp(argv[arg],"--latmax",8) && argv[arg][8]=='=')
@@ -117,12 +128,14 @@ int main(int argc,char** argv)
        ;
     else if(!strncmp(argv[arg],"--turn-relation=",16))
        ;
+    else if(!strncmp(argv[arg],"--errorlog=",11))
+       ;
     else
        print_usage(0,argv[arg],NULL);
    }
 
- if((option_statistics + option_visualiser + option_dump + option_dump_osm)!=1)
-    print_usage(0,NULL,"Must choose --visualiser, --statistics, --dump or --dump-osm.");
+ if((option_statistics + option_visualiser + option_dump + option_dump_osm + option_dump_visualiser)!=1)
+    print_usage(0,NULL,"Must choose --visualiser, --statistics, --dump, --dump-osm or --dump-visualiser.");
 
  /* Load in the data - Note: No error checking because Load*List() will call exit() in case of an error. */
 
@@ -134,12 +147,18 @@ int main(int argc,char** argv)
 
  OSMRelations=LoadRelationList(relations_filename=FileName(dirname,prefix,"relations.mem"));
 
+ if(ExistsFile(errorlogs_filename=FileName(dirname,prefix,"errorlogs.mem")))
+    OSMErrorLogs=LoadErrorLogs(errorlogs_filename);
+ else
+    errorlogs_filename=NULL;
+
  /* Write out the visualiser data */
 
  if(option_visualiser)
    {
     Highway highway;
     Transport transport;
+    Property property;
 
     if(coordcount!=4)
        print_usage(0,NULL,"The --visualiser option must have --latmin, --latmax, --lonmin, --lonmax.\n");
@@ -171,6 +190,10 @@ int main(int argc,char** argv)
        OutputWidthLimits(OSMNodes,OSMSegments,OSMWays,OSMRelations,latmin,latmax,lonmin,lonmax);
     else if(!strcmp(option_data,"length"))
        OutputLengthLimits(OSMNodes,OSMSegments,OSMWays,OSMRelations,latmin,latmax,lonmin,lonmax);
+    else if(!strncmp(option_data,"property",8) && option_data[8]=='-' && (property=PropertyType(option_data+9))!=Property_None)
+       OutputProperty(OSMNodes,OSMSegments,OSMWays,OSMRelations,latmin,latmax,lonmin,lonmax,property);
+    else if(!strcmp(option_data,"errorlogs"))
+       OutputErrorLog(OSMErrorLogs,latmin,latmax,lonmin,lonmax);
     else
        print_usage(0,option_data,NULL);
    }
@@ -189,28 +212,37 @@ int main(int argc,char** argv)
 
     stat(nodes_filename,&buf);
 
-    printf("'%s%snodes.mem'     - %9lld Bytes\n",prefix?prefix:"",prefix?"-":"",(long long)buf.st_size);
+    printf("'%s%snodes.mem'     - %9"PRIu64" Bytes\n",prefix?prefix:"",prefix?"-":"",(uint64_t)buf.st_size);
     printf("%s\n",RFC822Date(buf.st_mtime));
     printf("\n");
 
     stat(segments_filename,&buf);
 
-    printf("'%s%ssegments.mem'  - %9lld Bytes\n",prefix?prefix:"",prefix?"-":"",(long long)buf.st_size);
+    printf("'%s%ssegments.mem'  - %9"PRIu64" Bytes\n",prefix?prefix:"",prefix?"-":"",(uint64_t)buf.st_size);
     printf("%s\n",RFC822Date(buf.st_mtime));
     printf("\n");
 
     stat(ways_filename,&buf);
 
-    printf("'%s%sways.mem'      - %9lld Bytes\n",prefix?prefix:"",prefix?"-":"",(long long)buf.st_size);
+    printf("'%s%sways.mem'      - %9"PRIu64" Bytes\n",prefix?prefix:"",prefix?"-":"",(uint64_t)buf.st_size);
     printf("%s\n",RFC822Date(buf.st_mtime));
     printf("\n");
 
     stat(relations_filename,&buf);
 
-    printf("'%s%srelations.mem' - %9lld Bytes\n",prefix?prefix:"",prefix?"-":"",(long long)buf.st_size);
+    printf("'%s%srelations.mem' - %9"PRIu64" Bytes\n",prefix?prefix:"",prefix?"-":"",(uint64_t)buf.st_size);
     printf("%s\n",RFC822Date(buf.st_mtime));
     printf("\n");
 
+    if(errorlogs_filename)
+      {
+       stat(errorlogs_filename,&buf);
+
+       printf("'%s%serrorlogs.mem' - %9"PRIu64" Bytes\n",prefix?prefix:"",prefix?"-":"",(uint64_t)buf.st_size);
+       printf("%s\n",RFC822Date(buf.st_mtime));
+       printf("\n");
+      }
+
     /* Examine the nodes */
 
     printf("Nodes\n");
@@ -269,6 +301,26 @@ int main(int argc,char** argv)
 
     printf("sizeof(TurnRelation)=%9lu Bytes\n",(unsigned long)sizeof(TurnRelation));
     printf("Number              =%9"Pindex_t"\n",OSMRelations->file.trnumber);
+
+    if(errorlogs_filename)
+      {
+       printf("\n");
+       printf("Error Logs\n");
+       printf("----------\n");
+       printf("\n");
+
+       printf("Number(total)           =%9"Pindex_t"\n",OSMErrorLogs->file.number);
+       printf("Number(geographical)    =%9"Pindex_t"\n",OSMErrorLogs->file.number_geo);
+       printf("Number(non-geographical)=%9"Pindex_t"\n",OSMErrorLogs->file.number_nongeo);
+
+       printf("\n");
+       stat(errorlogs_filename,&buf);
+#if !SLIM
+       printf("Total strings=%9lu Bytes\n",(unsigned long)buf.st_size-(unsigned long)(OSMErrorLogs->strings-(char*)OSMErrorLogs->data));
+#else
+       printf("Total strings=%9lu Bytes\n",(unsigned long)buf.st_size-(unsigned long)OSMErrorLogs->stringsoffset);
+#endif
+      }
    }
 
  /* Print out internal data (in plain text format) */
@@ -287,7 +339,7 @@ int main(int argc,char** argv)
          {
           item=atoi(&argv[arg][7]);
 
-          if(item>=0 && item<OSMNodes->file.number)
+          if(item<OSMNodes->file.number)
              print_node(OSMNodes,item);
           else
              printf("Invalid node number; minimum=0, maximum=%"Pindex_t".\n",OSMNodes->file.number-1);
@@ -301,7 +353,7 @@ int main(int argc,char** argv)
          {
           item=atoi(&argv[arg][10]);
 
-          if(item>=0 && item<OSMSegments->file.number)
+          if(item<OSMSegments->file.number)
              print_segment(OSMSegments,item);
           else
              printf("Invalid segment number; minimum=0, maximum=%"Pindex_t".\n",OSMSegments->file.number-1);
@@ -315,7 +367,7 @@ int main(int argc,char** argv)
          {
           item=atoi(&argv[arg][6]);
 
-          if(item>=0 && item<OSMWays->file.number)
+          if(item<OSMWays->file.number)
              print_way(OSMWays,item);
           else
              printf("Invalid way number; minimum=0, maximum=%"Pindex_t".\n",OSMWays->file.number-1);
@@ -329,11 +381,25 @@ int main(int argc,char** argv)
          {
           item=atoi(&argv[arg][16]);
 
-          if(item>=0 && item<OSMRelations->file.trnumber)
+          if(item<OSMRelations->file.trnumber)
              print_turn_relation(OSMRelations,item,OSMSegments,OSMNodes);
           else
              printf("Invalid turn relation number; minimum=0, maximum=%"Pindex_t".\n",OSMRelations->file.trnumber-1);
          }
+       else if(!strcmp(argv[arg],"--errorlog=all"))
+         {
+          for(item=0;item<OSMErrorLogs->file.number;item++)
+             print_errorlog(OSMErrorLogs,item);
+         }
+       else if(!strncmp(argv[arg],"--errorlog=",11))
+         {
+          item=atoi(&argv[arg][11]);
+
+          if(item<OSMErrorLogs->file.number)
+             print_errorlog(OSMErrorLogs,item);
+          else
+             printf("Invalid error log number; minimum=0, maximum=%"Pindex_t".\n",OSMErrorLogs->file.number-1);
+         }
    }
 
  /* Print out internal data (in OSM XML format) */
@@ -365,6 +431,54 @@ int main(int argc,char** argv)
     print_tail_osm();
    }
 
+ /* Print out internal data (in HTML format for the visualiser) */
+
+ if(option_dump_visualiser)
+   {
+    index_t item;
+
+    if(!option_data)
+       print_usage(0,NULL,"The --dump-visualiser option must have --data.\n");
+
+    for(arg=1;arg<argc;arg++)
+       if(!strncmp(argv[arg],"--data=node",11))
+         {
+          item=atoi(&argv[arg][11]);
+
+          if(item<OSMNodes->file.number)
+             print_node_visualiser(OSMNodes,item);
+          else
+             printf("Invalid node number; minimum=0, maximum=%"Pindex_t".\n",OSMNodes->file.number-1);
+         }
+       else if(!strncmp(argv[arg],"--data=segment",14))
+         {
+          item=atoi(&argv[arg][14]);
+
+          if(item<OSMSegments->file.number)
+             print_segment_visualiser(OSMSegments,item,OSMWays);
+          else
+             printf("Invalid segment number; minimum=0, maximum=%"Pindex_t".\n",OSMSegments->file.number-1);
+         }
+       else if(!strncmp(argv[arg],"--data=turn-relation",20))
+         {
+          item=atoi(&argv[arg][20]);
+
+          if(item<OSMRelations->file.trnumber)
+             print_turn_relation_visualiser(OSMRelations,item,OSMSegments,OSMNodes);
+          else
+             printf("Invalid turn relation number; minimum=0, maximum=%"Pindex_t".\n",OSMRelations->file.trnumber-1);
+         }
+       else if(!strncmp(argv[arg],"--data=errorlog",15))
+         {
+          item=atoi(&argv[arg][15]);
+
+          if(item<OSMErrorLogs->file.number)
+             print_errorlog_visualiser(OSMErrorLogs,item);
+          else
+             printf("Invalid error log number; minimum=0, maximum=%"Pindex_t".\n",OSMErrorLogs->file.number-1);
+         }
+   }
+
  return(0);
 }
 
@@ -382,7 +496,7 @@ static void print_node(Nodes *nodes,index_t item)
  Node *nodep=LookupNode(nodes,item,1);
  double latitude,longitude;
 
- GetLatLong(nodes,item,&latitude,&longitude);
+ GetLatLong(nodes,item,nodep,&latitude,&longitude);
 
  printf("Node %"Pindex_t"\n",item);
  printf("  firstseg=%"Pindex_t"\n",nodep->firstseg);
@@ -390,6 +504,8 @@ static void print_node(Nodes *nodes,index_t item)
  printf("  allow=%02x (%s)\n",nodep->allow,AllowedNameList(nodep->allow));
  if(IsSuperNode(nodep))
     printf("  Super-Node\n");
+ if(nodep->flags & NODE_MINIRNDBT)
+    printf("  Mini-roundabout\n");
 }
 
 
@@ -505,6 +621,35 @@ static void print_turn_relation(Relations *relations,index_t item,Segments *segm
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Print out the contents of an error log from the routing database (as plain text).
+
+  ErrorLogs *errorlogs The set of error logs to use.
+
+  index_t item The error log index to print.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void print_errorlog(ErrorLogs *errorlogs,index_t item)
+{
+ ErrorLog *errorlogp=LookupErrorLog(errorlogs,item,1);
+
+ printf("Error Log %"Pindex_t"\n",item);
+
+ if(item<errorlogs->file.number_geo)
+   {
+    double latitude,longitude;
+
+    GetErrorLogLatLong(errorlogs,item,errorlogp,&latitude,&longitude);
+
+    printf("  latoffset=%d lonoffset=%d (latitude=%.6f longitude=%.6f)\n",errorlogp->latoffset,errorlogp->lonoffset,radians_to_degrees(latitude),radians_to_degrees(longitude));
+   }
+ else
+    printf("  No geographical information\n");
+
+ printf("  '%s'\n",LookupErrorLogString(errorlogs,item));
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Print out a header in OSM XML format.
 
@@ -599,7 +744,7 @@ static void print_region_osm(Nodes *nodes,Segments *segments,Ways *ways,Relation
                 double olat,olon;
                 index_t oitem=OtherNode(segmentp,item);
 
-                GetLatLong(nodes,oitem,&olat,&olon);
+                GetLatLong(nodes,oitem,NULL,&olat,&olon);
 
                 if(olat>latmin && olat<latmax && olon>lonmin && olon<lonmax)
                    if(item>oitem)
@@ -640,7 +785,7 @@ static void print_node_osm(Nodes *nodes,index_t item)
  double latitude,longitude;
  int i;
 
- GetLatLong(nodes,item,&latitude,&longitude);
+ GetLatLong(nodes,item,nodep,&latitude,&longitude);
 
  if(nodep->allow==Transports_ALL && nodep->flags==0)
     printf("  <node id='%lu' lat='%.7f' lon='%.7f' version='1' />\n",(unsigned long)item+1,radians_to_degrees(latitude),radians_to_degrees(longitude));
@@ -655,7 +800,7 @@ static void print_node_osm(Nodes *nodes,index_t item)
        printf("    <tag k='routino:uturn' v='yes' />\n");
 
     if(nodep->flags & NODE_MINIRNDBT)
-       printf("    <tag k='highway' v='mini_roundabout' />\n");
+       printf("    <tag k='junction' v='roundabout' />\n");
 
     if(nodep->flags & NODE_TURNRSTRCT)
        printf("    <tag k='routino:turnrestriction' v='yes' />\n");
@@ -798,6 +943,184 @@ static void print_tail_osm(void)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Print out the contents of a node from the routing database (in visualiser format).
+
+  Nodes *nodes The set of nodes to use.
+
+  index_t item The node index to print.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void print_node_visualiser(Nodes *nodes,index_t item)
+{
+ Node *nodep=LookupNode(nodes,item,1);
+ double latitude,longitude;
+ int i;
+
+ GetLatLong(nodes,item,nodep,&latitude,&longitude);
+
+ if(nodep->allow==Transports_ALL && nodep->flags==0)
+    printf("&lt;routino:node id='%lu' lat='%.7f' lon='%.7f' /&gt;\n",(unsigned long)item+1,radians_to_degrees(latitude),radians_to_degrees(longitude));
+ else
+   {
+    printf("&lt;routino:node id='%lu' lat='%.7f' lon='%.7f'&gt;\n",(unsigned long)item+1,radians_to_degrees(latitude),radians_to_degrees(longitude));
+
+    if(nodep->flags & NODE_SUPER)
+       printf("&nbsp;&nbsp;&nbsp;&lt;tag k='routino:super' v='yes' /&gt;\n");
+
+    if(nodep->flags & NODE_UTURN)
+       printf("&nbsp;&nbsp;&nbsp;&lt;tag k='routino:uturn' v='yes' /&gt;\n");
+
+    if(nodep->flags & NODE_MINIRNDBT)
+       printf("&nbsp;&nbsp;&nbsp;&lt;tag k='junction' v='roundabout' /&gt;\n");
+
+    if(nodep->flags & NODE_TURNRSTRCT)
+       printf("&nbsp;&nbsp;&nbsp;&lt;tag k='routino:turnrestriction' v='yes' /&gt;\n");
+
+    for(i=1;i<Transport_Count;i++)
+       if(!(nodep->allow & TRANSPORTS(i)))
+          printf("&nbsp;&nbsp;&nbsp;&lt;tag k='%s' v='no' /&gt;\n",TransportName(i));
+
+    printf("&lt;/routino:node&gt;\n");
+   }
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Print out the contents of a segment from the routing database (as a way in visualiser format).
+
+  Segments *segments The set of segments to use.
+
+  index_t item The segment index to print.
+
+  Ways *ways The set of ways to use.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void print_segment_visualiser(Segments *segments,index_t item,Ways *ways)
+{
+ Segment *segmentp=LookupSegment(segments,item,1);
+ Way *wayp=LookupWay(ways,segmentp->way,1);
+ char *name=WayName(ways,wayp);
+ int i;
+
+ printf("&lt;routino:way id='%lu'&gt;\n",(unsigned long)item+1);
+
+ if(IsOnewayTo(segmentp,segmentp->node1))
+   {
+    printf("&nbsp;&nbsp;&nbsp;&lt;nd ref='%lu' /&gt;\n",(unsigned long)segmentp->node2+1);
+    printf("&nbsp;&nbsp;&nbsp;&lt;nd ref='%lu' /&gt;\n",(unsigned long)segmentp->node1+1);
+   }
+ else
+   {
+    printf("&nbsp;&nbsp;&nbsp;&lt;nd ref='%lu' /&gt;\n",(unsigned long)segmentp->node1+1);
+    printf("&nbsp;&nbsp;&nbsp;&lt;nd ref='%lu' /&gt;\n",(unsigned long)segmentp->node2+1);
+   }
+
+ if(IsSuperSegment(segmentp))
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='routino:super' v='yes' /&gt;\n");
+ if(IsNormalSegment(segmentp))
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='routino:normal' v='yes' /&gt;\n");
+
+ printf("&nbsp;&nbsp;&nbsp;&lt;tag k='routino:distance' v='%.3f km' /&gt;\n",distance_to_km(DISTANCE(segmentp->distance)));
+
+ if(wayp->type & Highway_OneWay)
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='oneway' v='yes' /&gt;\n");
+
+ if(wayp->type & Highway_Roundabout)
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='roundabout' v='yes' /&gt;\n");
+
+ printf("&nbsp;&nbsp;&nbsp;&lt;tag k='highway' v='%s' /&gt;\n",HighwayName(HIGHWAY(wayp->type)));
+
+ if(IsNormalSegment(segmentp) && *name)
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='name' v='%s' /&gt;\n",ParseXML_Encode_Safe_XML(name));
+
+ for(i=1;i<Transport_Count;i++)
+    if(wayp->allow & TRANSPORTS(i))
+       printf("&nbsp;&nbsp;&nbsp;&lt;tag k='%s' v='yes' /&gt;\n",TransportName(i));
+
+ for(i=1;i<Property_Count;i++)
+    if(wayp->props & PROPERTIES(i))
+       printf("&nbsp;&nbsp;&nbsp;&lt;tag k='%s' v='yes' /&gt;\n",PropertyName(i));
+
+ if(wayp->speed)
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='maxspeed' v='%d kph' /&gt;\n",speed_to_kph(wayp->speed));
+
+ if(wayp->weight)
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='maxweight' v='%.1f t' /&gt;\n",weight_to_tonnes(wayp->weight));
+ if(wayp->height)
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='maxheight' v='%.1f m' /&gt;\n",height_to_metres(wayp->height));
+ if(wayp->width)
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='maxwidth' v='%.1f m' /&gt;\n",width_to_metres(wayp->width));
+ if(wayp->length)
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='maxlength' v='%.1f m' /&gt;\n",length_to_metres(wayp->length));
+
+ printf("&lt;/routino:way&gt;\n");
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Print out the contents of a turn relation from the routing database (in visualiser format).
+
+  Relations *relations The set of relations to use.
+
+  index_t item The relation index to print.
+
+  Segments *segments The set of segments to use.
+
+  Nodes *nodes The set of nodes to use.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void print_turn_relation_visualiser(Relations *relations,index_t item,Segments *segments,Nodes *nodes)
+{
+ TurnRelation *relationp=LookupTurnRelation(relations,item,1);
+
+ Segment *segmentp_from=LookupSegment(segments,relationp->from,1);
+ Segment *segmentp_to  =LookupSegment(segments,relationp->to  ,2);
+
+ double angle=TurnAngle(nodes,segmentp_from,segmentp_to,relationp->via);
+
+ char *restriction;
+
+ if(angle>150 || angle<-150)
+    restriction="no_u_turn";
+ else if(angle>30)
+    restriction="no_right_turn";
+ else if(angle<-30)
+    restriction="no_left_turn";
+ else
+    restriction="no_straight_on";
+
+ printf("&lt;routino:relation id='%lu'&gt;\n",(unsigned long)item+1);
+ printf("&nbsp;&nbsp;&nbsp;&lt;tag k='type' v='restriction' /&gt;\n");
+ printf("&nbsp;&nbsp;&nbsp;&lt;tag k='restriction' v='%s'/&gt;\n",restriction);
+
+ if(relationp->except)
+    printf("&nbsp;&nbsp;&nbsp;&lt;tag k='except' v='%s' /&gt;\n",AllowedNameList(relationp->except));
+
+ printf("&nbsp;&nbsp;&nbsp;&lt;member type='way' ref='%lu' role='from' /&gt;\n",(unsigned long)relationp->from+1);
+ printf("&nbsp;&nbsp;&nbsp;&lt;member type='node' ref='%lu' role='via' /&gt;\n",(unsigned long)relationp->via+1);
+ printf("&nbsp;&nbsp;&nbsp;&lt;member type='way' ref='%lu' role='to' /&gt;\n",(unsigned long)relationp->to+1);
+
+ printf("&lt;/routino:relation&gt;\n");
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Print out an error log entry from the database (in visualiser format).
+
+  ErrorLogs *errorlogs The set of error logs to use.
+
+  index_t item The error log index to print.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void print_errorlog_visualiser(ErrorLogs *errorlogs,index_t item)
+{
+ char *string=LookupErrorLogString(errorlogs,item);
+
+ printf("%s\n",ParseXML_Encode_Safe_XML(string));
+}
+
+
 /*+ Conversion from time_t to date string (day of week). +*/
 static const char* const weekdays[7]={"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
 
@@ -864,10 +1187,15 @@ static void print_usage(int detail,const char *argerr,const char *err)
          "                  [--dump [--node=<node> ...]\n"
          "                          [--segment=<segment> ...]\n"
          "                          [--way=<way> ...]\n"
-         "                          [--turn-relation=<rel> ...]]\n"
+         "                          [--turn-relation=<rel> ...]\n"
+         "                          [--errorlog=<number> ...]]\n"
          "                  [--dump-osm [--no-super]\n"
          "                              [--latmin=<latmin> --latmax=<latmax>\n"
-         "                               --lonmin=<lonmin> --lonmax=<lonmax>]]\n");
+         "                               --lonmin=<lonmin> --lonmax=<lonmax>]]\n"
+         "                  [--dump-visualiser [--data=node<node>]\n"
+         "                                     [--data=segment<segment>]\n"
+         "                                     [--data=turn-relation<rel>]\n"
+         "                                     [--data=errorlog<number>]]\n");
 
  if(argerr)
     fprintf(stderr,
@@ -909,20 +1237,29 @@ static void print_usage(int detail,const char *argerr,const char *err)
             "      height      = height limits.\n"
             "      width       = width limits.\n"
             "      length      = length limits.\n"
+            "      property-*  = segments with the specified property.\n"
+            "      errorlogs   = errors logged during parsing.\n"
             "\n"
             "--dump                    Dump selected contents of the database.\n"
-            "  --node=<node>           * the node with the selected index.\n"
-            "  --segment=<segment>     * the segment with the selected index.\n"
-            "  --way=<way>             * the way with the selected index.\n"
-            "  --turn-relation=<rel>   * the turn relation with the selected index.\n"
+            "  --node=<node>             * the node with the selected index.\n"
+            "  --segment=<segment>       * the segment with the selected index.\n"
+            "  --way=<way>               * the way with the selected index.\n"
+            "  --turn-relation=<rel>     * the turn relation with the selected index.\n"
+            "  --errorlog=<number>       * the error log with the selected index.\n"
             "                          Use 'all' instead of a number to get all of them.\n"
             "\n"
             "--dump-osm                Dump all or part of the database as an XML file.\n"
-            "  --no-super              * exclude the super-segments.\n"
-            "  --latmin=<latmin>       * the minimum latitude (degrees N).\n"
-            "  --latmax=<latmax>       * the maximum latitude (degrees N).\n"
-            "  --lonmin=<lonmin>       * the minimum longitude (degrees E).\n"
-            "  --lonmax=<lonmax>       * the maximum longitude (degrees E).\n");
+            "  --no-super                * exclude the super-segments.\n"
+            "  --latmin=<latmin>         * the minimum latitude (degrees N).\n"
+            "  --latmax=<latmax>         * the maximum latitude (degrees N).\n"
+            "  --lonmin=<lonmin>         * the minimum longitude (degrees E).\n"
+            "  --lonmax=<lonmax>         * the maximum longitude (degrees E).\n"
+            "\n"
+            "--dump-visualiser         Dump selected contents of the database in HTML.\n"
+            "  --data=node<node>         * the node with the selected index.\n"
+            "  --data=segment<segment>   * the segment with the selected index.\n"
+            "  --data=turn-relation<rel> * the turn relation with the selected index.\n"
+            "  --data=errorlog<number>   * the error log with the selected index.\n");
 
  exit(!detail);
 }
