@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -48,14 +48,40 @@ static void vfprintf_last(FILE *file,const char *format,va_list ap);
 
 /* Local variables */
 
-/*+ The time that printf_first was called. +*/
-static struct timeval start_time;
+/*+ The time that program_start() was called. +*/
+static struct timeval program_start_time;
+
+/*+ The time that printf_first() was called. +*/
+static struct timeval function_start_time;
 
 /*+ The length of the string printed out last time. +*/
 static int printed_length=0;
 
-/*+ The file handle for the error log file. +*/
-static FILE *errorlogfile;
+
+/*++++++++++++++++++++++++++++++++++++++
+  Record the time that the program started.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void printf_program_start(void)
+{
+ gettimeofday(&program_start_time,NULL);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Record the time that the program started.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void printf_program_end(void)
+{
+ if(option_logtime)
+   {
+    printf("\n");
+    fprintf_elapsed_time(stdout,&program_start_time);
+    printf("Complete\n");
+    fflush(stdout);
+   }
+}
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -71,7 +97,7 @@ void printf_first(const char *format, ...)
  va_list ap;
 
  if(option_logtime)
-    gettimeofday(&start_time,NULL);
+    gettimeofday(&function_start_time,NULL);
 
  if(option_loggable)
     return;
@@ -142,7 +168,7 @@ void fprintf_first(FILE *file,const char *format, ...)
  va_list ap;
 
  if(option_logtime)
-    gettimeofday(&start_time,NULL);
+    gettimeofday(&function_start_time,NULL);
 
  if(option_loggable)
     return;
@@ -217,7 +243,7 @@ static void vfprintf_first(FILE *file,const char *format,va_list ap)
  int retval;
 
  if(option_logtime)
-    fprintf_elapsed_time(file,&start_time);
+    fprintf_elapsed_time(file,&function_start_time);
 
  retval=vfprintf(file,format,ap);
  fflush(file);
@@ -244,7 +270,7 @@ static void vfprintf_middle(FILE *file,const char *format,va_list ap)
  fputc('\r',file);
 
  if(option_logtime)
-    fprintf_elapsed_time(file,&start_time);
+    fprintf_elapsed_time(file,&function_start_time);
 
  retval=vfprintf(file,format,ap);
  fflush(file);
@@ -279,7 +305,7 @@ static void vfprintf_last(FILE *file,const char *format,va_list ap)
     fputc('\r',file);
 
  if(option_logtime)
-    fprintf_elapsed_time(file,&start_time);
+    fprintf_elapsed_time(file,&function_start_time);
 
  retval=vfprintf(file,format,ap);
 
@@ -314,61 +340,7 @@ void fprintf_elapsed_time(FILE *file,struct timeval *start)
     elapsed.tv_usec+=1000000;
    }
 
- fprintf(file,"[%2ld:%02ld.%03ld] ",elapsed.tv_sec/60,elapsed.tv_sec%60,elapsed.tv_usec/10000);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  Create the error log file.
-
-  const char *filename The name of the file to create.
-
-  int append The option to append to an existing file.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-void open_errorlog(const char *filename,int append)
-{
- errorlogfile=fopen(filename,append?"a":"w");
-
- if(!errorlogfile)
-   {
-    fprintf(stderr,"Cannot open file '%s' for writing [%s].\n",filename,strerror(errno));
-    exit(EXIT_FAILURE);
-   }
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  Close the error log file.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-void close_errorlog(void)
-{
- if(errorlogfile)
-    fclose(errorlogfile);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  Log a message to the error log file.
-
-  const char *format The format string.
-
-  ... The other arguments.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-void logerror(const char *format, ...)
-{
- va_list ap;
-
- if(!errorlogfile)
-    return;
-
- va_start(ap,format);
-
- vfprintf(errorlogfile,format,ap);
-
- va_end(ap);
+ fprintf(file,"[%2ld:%02ld.%03ld] ",elapsed.tv_sec/60,elapsed.tv_sec%60,elapsed.tv_usec/1000);
 }
 
 
