@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -50,7 +50,6 @@ Nodes *LoadNodeList(const char *filename)
  Nodes *nodes;
 #if SLIM
  size_t sizeoffsets;
- int i;
 #endif
 
  nodes=(Nodes*)malloc(sizeof(Nodes));
@@ -70,22 +69,21 @@ Nodes *LoadNodeList(const char *filename)
 
 #else
 
- nodes->fd=ReOpenFile(filename);
+ nodes->fd=SlimMapFile(filename);
 
  /* Copy the NodesFile header structure from the loaded data */
 
- ReadFile(nodes->fd,&nodes->file,sizeof(NodesFile));
+ SlimFetch(nodes->fd,&nodes->file,sizeof(NodesFile),0);
 
  sizeoffsets=(nodes->file.latbins*nodes->file.lonbins+1)*sizeof(index_t);
 
  nodes->offsets=(index_t*)malloc(sizeoffsets);
 
- ReadFile(nodes->fd,nodes->offsets,sizeoffsets);
+ SlimFetch(nodes->fd,nodes->offsets,sizeoffsets,sizeof(NodesFile));
 
  nodes->nodesoffset=sizeof(NodesFile)+sizeoffsets;
 
- for(i=0;i<sizeof(nodes->cached)/sizeof(nodes->cached[0]);i++)
-    nodes->incache[i]=NO_NODE;
+ nodes->cache=NewNodeCache();
 
 #endif
 
@@ -93,6 +91,32 @@ Nodes *LoadNodeList(const char *filename)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Destroy the node list.
+
+  Nodes *nodes The node list to destroy.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void DestroyNodeList(Nodes *nodes)
+{
+#if !SLIM
+
+ nodes->data=UnmapFile(nodes->data);
+
+#else
+
+ nodes->fd=SlimUnmapFile(nodes->fd);
+
+ free(nodes->offsets);
+
+ DeleteNodeCache(nodes->cache);
+
+#endif
+
+ free(nodes);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Find the closest node given its latitude, longitude and the profile of the
   mode of transport that must be able to move to/from this node.
@@ -372,7 +396,7 @@ index_t FindClosestSegment(Nodes *nodes,Segments *segments,Ways *ways,double lat
                       distance_t dist2,dist3;
                       double lat2,lon2,dist3a,dist3b,distp;
 
-                      GetLatLong(nodes,OtherNode(segmentp,i),&lat2,&lon2);
+                      GetLatLong(nodes,OtherNode(segmentp,i),NULL,&lat2,&lon2);
 
                       dist2=Distance(lat2,lon2,latitude,longitude);
 
@@ -380,26 +404,37 @@ index_t FindClosestSegment(Nodes *nodes,Segments *segments,Ways *ways,double lat
 
                       /* Use law of cosines (assume flat Earth) */
 
-                      dist3a=((double)dist1*(double)dist1-(double)dist2*(double)dist2+(double)dist3*(double)dist3)/(2.0*(double)dist3);
-                      dist3b=(double)dist3-dist3a;
-
-                      if((dist1+dist2)<dist3)
+                      if(dist3==0)
+                        {
+                         distp=dist1;  /* == dist2 */
+                         dist3a=dist1; /* == dist2 */
+                         dist3b=dist2; /* == dist1 */
+                        }
+                      else if((dist1+dist2)<dist3)
                         {
                          distp=0;
+                         dist3a=dist1;
+                         dist3b=dist2;
                         }
-                      else if(dist3a>=0 && dist3b>=0)
-                         distp=sqrt((double)dist1*(double)dist1-dist3a*dist3a);
-                      else if(dist3a>0)
+                      else
                         {
-                         distp=dist2;
-                         dist3a=dist3;
-                         dist3b=0;
-                        }
-                      else /* if(dist3b>0) */
-                        {
-                         distp=dist1;
-                         dist3a=0;
-                         dist3b=dist3;
+                         dist3a=((double)dist1*(double)dist1-(double)dist2*(double)dist2+(double)dist3*(double)dist3)/(2.0*(double)dist3);
+                         dist3b=(double)dist3-dist3a;
+
+                         if(dist3a>=0 && dist3b>=0)
+                            distp=sqrt((double)dist1*(double)dist1-dist3a*dist3a);
+                         else if(dist3a>0)
+                           {
+                            distp=dist2;
+                            dist3a=dist3;
+                            dist3b=0;
+                           }
+                         else /* if(dist3b>0) */
+                           {
+                            distp=dist1;
+                            dist3a=0;
+                            dist3b=dist3;
+                           }
                         }
 
                       if(distp<(double)bestd)
@@ -510,16 +545,18 @@ static int valid_segment_for_profile(Ways *ways,Segment *segmentp,Profile *profi
 
   index_t index The node index.
 
+  Node *nodep A pointer to the node if already available.
+
   double *latitude Returns the latitude.
 
   double *longitude Returns the logitude.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void GetLatLong(Nodes *nodes,index_t index,double *latitude,double *longitude)
+void GetLatLong(Nodes *nodes,index_t index,Node *nodep,double *latitude,double *longitude)
 {
- Node *nodep=LookupNode(nodes,index,4);
- ll_bin_t latbin=-1,lonbin=-1;
- ll_bin_t start,end,mid;
+ ll_bin_t latbin,lonbin;
+ ll_bin2_t bin=-1;
+ ll_bin2_t start,end,mid;
  index_t offset;
 
  /* Binary search - search key nearest match below is required.
@@ -533,76 +570,48 @@ void GetLatLong(Nodes *nodes,index_t index,double *latitude,double *longitude)
   *  # <- end    |  start or end is the wanted one.
   */
 
- /* Search for longitude */
+ /* Search for offset */
 
  start=0;
- end=nodes->file.lonbins-1;
+ end=nodes->file.lonbins*nodes->file.latbins;
 
  do
    {
     mid=(start+end)/2;                  /* Choose mid point */
 
-    offset=LookupNodeOffset(nodes,nodes->file.latbins*mid);
+    offset=LookupNodeOffset(nodes,mid);
 
     if(offset<index)                    /* Mid point is too low for an exact match but could be lower bound */
        start=mid;
     else if(offset>index)               /* Mid point is too high */
        end=mid?(mid-1):mid;
     else                                /* Mid point is correct */
-      {lonbin=mid;break;}
+      {bin=mid;break;}
    }
  while((end-start)>1);
 
- if(lonbin==-1)
+ if(bin==-1)
    {
-    offset=LookupNodeOffset(nodes,nodes->file.latbins*end);
+    offset=LookupNodeOffset(nodes,end);
 
     if(offset>index)
-       lonbin=start;
+       bin=start;
     else
-       lonbin=end;
+       bin=end;
    }
 
- while(lonbin<nodes->file.lonbins && 
-       LookupNodeOffset(nodes,lonbin*nodes->file.latbins)==LookupNodeOffset(nodes,(lonbin+1)*nodes->file.latbins))
-    lonbin++;
+ while(bin<=(nodes->file.lonbins*nodes->file.latbins) && 
+       LookupNodeOffset(nodes,bin)==LookupNodeOffset(nodes,bin+1))
+    bin++;
 
- /* Search for latitude */
-
- start=0;
- end=nodes->file.latbins-1;
-
- do
-   {
-    mid=(start+end)/2;                  /* Choose mid point */
-
-    offset=LookupNodeOffset(nodes,lonbin*nodes->file.latbins+mid);
-
-    if(offset<index)                    /* Mid point is too low for an exact match but could be lower bound */
-       start=mid;
-    else if(offset>index)               /* Mid point is too high */
-       end=mid?(mid-1):mid;
-    else                                /* Mid point is correct */
-      {latbin=mid;break;}
-   }
- while((end-start)>1);
-
- if(latbin==-1)
-   {
-    offset=LookupNodeOffset(nodes,lonbin*nodes->file.latbins+end);
-
-    if(offset>index)
-       latbin=start;
-    else
-       latbin=end;
-   }
-
- while(latbin<nodes->file.latbins &&
-       LookupNodeOffset(nodes,lonbin*nodes->file.latbins+latbin)==LookupNodeOffset(nodes,lonbin*nodes->file.latbins+latbin+1))
-    latbin++;
+ latbin=bin%nodes->file.latbins;
+ lonbin=bin/nodes->file.latbins;
 
  /* Return the values */
 
+ if(nodep==NULL)
+    nodep=LookupNode(nodes,index,4);
+
  *latitude =latlong_to_radians(bin_to_latlong(nodes->file.latzero+latbin)+off_to_latlong(nodep->latoffset));
  *longitude=latlong_to_radians(bin_to_latlong(nodes->file.lonzero+lonbin)+off_to_latlong(nodep->lonoffset));
 }
