@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -27,6 +27,7 @@
 #include <sys/types.h>
 
 #include "types.h"
+#include "cache.h"
 
 #include "files.h"
 #include "profiles.h"
@@ -85,7 +86,8 @@ struct _Nodes
  off_t     nodesoffset;         /*+ The offset of the nodes within the file. +*/
 
  Node      cached[6];           /*+ Some cached nodes read from the file in slim mode. +*/
- index_t   incache[6];          /*+ The indexes of the cached nodes. +*/
+
+ NodeCache *cache;              /*+ A RAM cache of nodes read from the file. +*/
 
 #endif
 };
@@ -95,6 +97,8 @@ struct _Nodes
 
 Nodes *LoadNodeList(const char *filename);
 
+void DestroyNodeList(Nodes *nodes);
+
 index_t FindClosestNode(Nodes *nodes,Segments *segments,Ways *ways,double latitude,double longitude,
                         distance_t distance,Profile *profile,distance_t *bestdist);
 
@@ -102,7 +106,7 @@ index_t FindClosestSegment(Nodes *nodes,Segments *segments,Ways *ways,double lat
                            distance_t distance,Profile *profile, distance_t *bestdist,
                            index_t *bestnode1,index_t *bestnode2,distance_t *bestdist1,distance_t *bestdist2);
 
-void GetLatLong(Nodes *nodes,index_t index,double *latitude,double *longitude);
+void GetLatLong(Nodes *nodes,index_t index,Node *nodep,double *latitude,double *longitude);
 
 
 /* Macros and inline functions */
@@ -127,7 +131,23 @@ void GetLatLong(Nodes *nodes,index_t index,double *latitude,double *longitude);
 
 #else
 
-static Node *LookupNode(Nodes *nodes,index_t index,int position);
+/* Prototypes */
+
+static inline Node *LookupNode(Nodes *nodes,index_t index,int position);
+
+CACHE_NEWCACHE_PROTO(Node)
+CACHE_DELETECACHE_PROTO(Node)
+CACHE_FETCHCACHE_PROTO(Node)
+CACHE_INVALIDATECACHE_PROTO(Node)
+
+
+/* Inline functions */
+
+CACHE_STRUCTURE(Node)
+CACHE_NEWCACHE(Node)
+CACHE_DELETECACHE(Node)
+CACHE_FETCHCACHE(Node)
+CACHE_INVALIDATECACHE(Node)
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -144,12 +164,7 @@ static Node *LookupNode(Nodes *nodes,index_t index,int position);
 
 static inline Node *LookupNode(Nodes *nodes,index_t index,int position)
 {
- if(nodes->incache[position-1]!=index)
-   {
-    SeekReadFile(nodes->fd,&nodes->cached[position-1],sizeof(Node),nodes->nodesoffset+(off_t)index*sizeof(Node));
-
-    nodes->incache[position-1]=index;
-   }
+ nodes->cached[position-1]=*FetchCachedNode(nodes->cache,index,nodes->fd,nodes->nodesoffset);
 
  return(&nodes->cached[position-1]);
 }
