@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -52,6 +52,7 @@ static latlong_t lat_min,lat_max,lon_min,lon_max;
 static int sort_by_id(NodeX *a,NodeX *b);
 static int deduplicate_and_index_by_id(NodeX *nodex,index_t index);
 
+static int update_id(NodeX *nodex,index_t index);
 static int sort_by_lat_long(NodeX *a,NodeX *b);
 static int index_by_lat_long(NodeX *nodex,index_t index);
 
@@ -93,12 +94,16 @@ NodesX *NewNodeList(int append,int readonly)
       }
 
  if(append)
-    nodesx->fd=OpenFileAppend(nodesx->filename_tmp);
+    nodesx->fd=OpenFileBufferedAppend(nodesx->filename_tmp);
  else if(!readonly)
-    nodesx->fd=OpenFileNew(nodesx->filename_tmp);
+    nodesx->fd=OpenFileBufferedNew(nodesx->filename_tmp);
  else
     nodesx->fd=-1;
 
+#if SLIM
+ nodesx->cache=NewNodeXCache();
+#endif
+
  return(nodesx);
 }
 
@@ -133,6 +138,10 @@ void FreeNodeList(NodesX *nodesx,int keep)
  if(nodesx->super)
     free(nodesx->super);
 
+#if SLIM
+ DeleteNodeXCache(nodesx->cache);
+#endif
+
  free(nodesx);
 }
 
@@ -163,7 +172,7 @@ void AppendNodeList(NodesX *nodesx,node_t id,double latitude,double longitude,tr
  nodex.allow=allow;
  nodex.flags=flags;
 
- WriteFile(nodesx->fd,&nodex,sizeof(NodeX));
+ WriteFileBuffered(nodesx->fd,&nodex,sizeof(NodeX));
 
  nodesx->number++;
 
@@ -180,7 +189,7 @@ void AppendNodeList(NodesX *nodesx,node_t id,double latitude,double longitude,tr
 void FinishNodeList(NodesX *nodesx)
 {
  if(nodesx->fd!=-1)
-    nodesx->fd=CloseFile(nodesx->fd);
+    nodesx->fd=CloseFileBuffered(nodesx->fd);
 }
 
 
@@ -260,11 +269,11 @@ void SortNodeList(NodesX *nodesx)
 
  /* Re-open the file read-only and a new file writeable */
 
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
+ nodesx->fd=ReOpenFileBuffered(nodesx->filename_tmp);
 
  DeleteFile(nodesx->filename_tmp);
 
- fd=OpenFileNew(nodesx->filename_tmp);
+ fd=OpenFileBufferedNew(nodesx->filename_tmp);
 
  /* Allocate the array of indexes */
 
@@ -282,10 +291,12 @@ void SortNodeList(NodesX *nodesx)
                                                            (int (*)(const void*,const void*))sort_by_id,
                                                            (int (*)(void*,index_t))deduplicate_and_index_by_id);
 
+ nodesx->knumber=nodesx->number;
+
  /* Close the files */
 
- nodesx->fd=CloseFile(nodesx->fd);
- CloseFile(fd);
+ nodesx->fd=CloseFileBuffered(nodesx->fd);
+ CloseFileBuffered(fd);
 
  /* Print the final message */
 
@@ -354,44 +365,97 @@ static int deduplicate_and_index_by_id(NodeX *nodex,index_t index)
 
   NodesX *nodesx The set of nodes to modify.
 
-  SegmentsX *segmentsx The set of segments to use.
+  WaysX *waysx The set of ways to use.
 
   int keep If set to 1 then keep the old data file otherwise delete it.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void RemoveNonHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,int keep)
+void RemoveNonHighwayNodes(NodesX *nodesx,WaysX *waysx,int keep)
 {
+ BitMask *usednode;
  NodeX nodex;
- index_t total=0,highway=0,nothighway=0;
+ index_t i,total=0,highway=0,nothighway=0;
  int fd;
 
  /* Print the start message */
 
- printf_first("Checking Nodes: Nodes=0");
+ printf_first("Checking Ways for unused Nodes: Ways=0");
+
+ /* Allocate the node usage bitmask */
+
+ usednode=AllocBitMask(nodesx->number);
+
+ logassert(usednode,"Failed to allocate memory (try using slim mode?)"); /* Check AllocBitMask() worked */
+
+ /* Re-open the file read-only */
+
+ waysx->fd=ReOpenFileBuffered(waysx->filename_tmp);
+
+ /* Loop through the ways and mark the used nodes */
+
+ for(i=0;i<waysx->number;i++)
+   {
+    WayX wayx;
+    FILESORT_VARINT waysize;
+    node_t node;
+
+    ReadFileBuffered(waysx->fd,&waysize,FILESORT_VARSIZE);
+
+    ReadFileBuffered(waysx->fd,&wayx,sizeof(WayX));
+
+    while(!ReadFileBuffered(waysx->fd,&node,sizeof(node_t)) && node!=NO_NODE_ID)
+      {
+       index_t index=IndexNodeX(nodesx,node);
+
+       waysize-=sizeof(node_t);
+
+       if(index!=NO_NODE)
+          SetBit(usednode,index);
+      }
+
+    waysize-=sizeof(node_t)+sizeof(WayX);
+
+    SkipFileBuffered(waysx->fd,waysize);
+
+    if(!((i+1)%1000))
+       printf_middle("Checking Ways for unused Nodes: Ways=%"Pindex_t,i+1);
+   }
+
+ /* Close the file */
+
+ waysx->fd=CloseFileBuffered(waysx->fd);
+
+ /* Print the final message */
+
+ printf_last("Checked Ways for unused Nodes: Ways=%"Pindex_t,waysx->number);
+
+
+ /* Print the start message */
+
+ printf_first("Removing unused Nodes: Nodes=0");
 
  /* Re-open the file read-only and a new file writeable */
 
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
+ nodesx->fd=ReOpenFileBuffered(nodesx->filename_tmp);
 
  if(keep)
     RenameFile(nodesx->filename_tmp,nodesx->filename);
  else
     DeleteFile(nodesx->filename_tmp);
 
- fd=OpenFileNew(nodesx->filename_tmp);
+ fd=OpenFileBufferedNew(nodesx->filename_tmp);
 
  /* Modify the on-disk image */
 
- while(!ReadFile(nodesx->fd,&nodex,sizeof(NodeX)))
+ while(!ReadFileBuffered(nodesx->fd,&nodex,sizeof(NodeX)))
    {
-    if(!IsBitSet(segmentsx->usednode,total))
+    if(!IsBitSet(usednode,total))
        nothighway++;
     else
       {
-       nodex.id=highway;
-       nodesx->idata[highway]=nodesx->idata[total];
+       nodesx->idata[highway]=nodex.id;
 
-       WriteFile(fd,&nodex,sizeof(NodeX));
+       WriteFileBuffered(fd,&nodex,sizeof(NodeX));
 
        highway++;
       }
@@ -399,24 +463,23 @@ void RemoveNonHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,int keep)
     total++;
 
     if(!(total%10000))
-       printf_middle("Checking Nodes: Nodes=%"Pindex_t" Highway=%"Pindex_t" not-Highway=%"Pindex_t,total,highway,nothighway);
+       printf_middle("Removing unused Nodes: Nodes=%"Pindex_t" Highway=%"Pindex_t" not-Highway=%"Pindex_t,total,highway,nothighway);
    }
 
  nodesx->number=highway;
 
  /* Close the files */
 
- nodesx->fd=CloseFile(nodesx->fd);
- CloseFile(fd);
+ nodesx->fd=CloseFileBuffered(nodesx->fd);
+ CloseFileBuffered(fd);
 
  /* Free the now-unneeded index */
 
- free(segmentsx->usednode);
- segmentsx->usednode=NULL;
+ free(usednode);
 
  /* Print the final message */
 
- printf_last("Checked Nodes: Nodes=%"Pindex_t" Highway=%"Pindex_t" not-Highway=%"Pindex_t,total,highway,nothighway);
+ printf_last("Removed unused Nodes: Nodes=%"Pindex_t" Highway=%"Pindex_t" not-Highway=%"Pindex_t,total,highway,nothighway);
 }
 
 
@@ -449,15 +512,15 @@ void RemovePrunedNodes(NodesX *nodesx,SegmentsX *segmentsx)
 
  /* Re-open the file read-only and a new file writeable */
 
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
+ nodesx->fd=ReOpenFileBuffered(nodesx->filename_tmp);
 
  DeleteFile(nodesx->filename_tmp);
 
- fd=OpenFileNew(nodesx->filename_tmp);
+ fd=OpenFileBufferedNew(nodesx->filename_tmp);
 
  /* Modify the on-disk image */
 
- while(!ReadFile(nodesx->fd,&nodex,sizeof(NodeX)))
+ while(!ReadFileBuffered(nodesx->fd,&nodex,sizeof(NodeX)))
    {
     if(segmentsx->firstnode[total]==NO_SEGMENT)
       {
@@ -467,10 +530,9 @@ void RemovePrunedNodes(NodesX *nodesx,SegmentsX *segmentsx)
       }
     else
       {
-       nodex.id=notpruned;
        nodesx->pdata[total]=notpruned;
 
-       WriteFile(fd,&nodex,sizeof(NodeX));
+       WriteFileBuffered(fd,&nodex,sizeof(NodeX));
 
        notpruned++;
       }
@@ -485,8 +547,8 @@ void RemovePrunedNodes(NodesX *nodesx,SegmentsX *segmentsx)
 
  /* Close the files */
 
- nodesx->fd=CloseFile(nodesx->fd);
- CloseFile(fd);
+ nodesx->fd=CloseFileBuffered(nodesx->fd);
+ CloseFileBuffered(fd);
 
  /* Print the final message */
 
@@ -508,17 +570,17 @@ void SortNodeListGeographically(NodesX *nodesx)
  if(nodesx->number==0)
     return;
 
- /* While we are here we can work out the range of data */
+ /* Print the start message */
+
+ printf_first("Sorting Nodes Geographically");
+
+ /* Work out the range of data */
 
  lat_min=radians_to_latlong( 2);
  lat_max=radians_to_latlong(-2);
  lon_min=radians_to_latlong( 4);
  lon_max=radians_to_latlong(-4);
 
- /* Print the start message */
-
- printf_first("Sorting Nodes Geographically");
-
  /* Allocate the memory for the geographical index array */
 
  nodesx->gdata=(index_t*)malloc(nodesx->number*sizeof(index_t));
@@ -527,24 +589,24 @@ void SortNodeListGeographically(NodesX *nodesx)
 
  /* Re-open the file read-only and a new file writeable */
 
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
+ nodesx->fd=ReOpenFileBuffered(nodesx->filename_tmp);
 
  DeleteFile(nodesx->filename_tmp);
 
- fd=OpenFileNew(nodesx->filename_tmp);
+ fd=OpenFileBufferedNew(nodesx->filename_tmp);
 
  /* Sort nodes geographically and index them */
 
  sortnodesx=nodesx;
 
- filesort_fixed(nodesx->fd,fd,sizeof(NodeX),NULL,
+ filesort_fixed(nodesx->fd,fd,sizeof(NodeX),(int (*)(void*,index_t))update_id,
                                             (int (*)(const void*,const void*))sort_by_lat_long,
                                             (int (*)(void*,index_t))index_by_lat_long);
 
  /* Close the files */
 
- nodesx->fd=CloseFile(nodesx->fd);
- CloseFile(fd);
+ nodesx->fd=CloseFileBuffered(nodesx->fd);
+ CloseFileBuffered(fd);
 
  /* Free the memory */
 
@@ -570,6 +632,27 @@ void SortNodeListGeographically(NodesX *nodesx)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Update the node ids.
+
+  int update_id Return 1 if the value is to be kept, otherwise 0.
+
+  NodeX *nodex The extended node.
+
+  index_t index The number of unsorted nodes that have been read from the input file.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int update_id(NodeX *nodex,index_t index)
+{
+ nodex->id=index;
+
+ if(IsBitSet(sortnodesx->super,index))
+    nodex->flags|=NODE_SUPER;
+
+ return(1);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Sort the nodes into latitude and longitude order (first by longitude bin
   number, then by latitude bin number and then by exact longitude and then by
@@ -634,9 +717,6 @@ static int index_by_lat_long(NodeX *nodex,index_t index)
 {
  sortnodesx->gdata[nodex->id]=index;
 
- if(IsBitSet(sortnodesx->super,nodex->id))
-    nodex->flags|=NODE_SUPER;
-
  if(nodex->latitude<lat_min)
     lat_min=nodex->latitude;
  if(nodex->latitude>lat_max)
@@ -683,13 +763,13 @@ void SaveNodeList(NodesX *nodesx,const char *filename,SegmentsX *segmentsx)
 
  /* Re-open the file */
 
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
+ nodesx->fd=ReOpenFileBuffered(nodesx->filename_tmp);
 
  /* Write out the nodes data */
 
- fd=OpenFileNew(filename);
+ fd=OpenFileBufferedNew(filename);
 
- SeekFile(fd,sizeof(NodesFile)+(nodesx->latbins*nodesx->lonbins+1)*sizeof(index_t));
+ SeekFileBuffered(fd,sizeof(NodesFile)+(nodesx->latbins*nodesx->lonbins+1)*sizeof(index_t));
 
  for(i=0;i<nodesx->number;i++)
    {
@@ -698,13 +778,13 @@ void SaveNodeList(NodesX *nodesx,const char *filename,SegmentsX *segmentsx)
     ll_bin_t latbin,lonbin;
     ll_bin2_t llbin;
 
-    ReadFile(nodesx->fd,&nodex,sizeof(NodeX));
+    ReadFileBuffered(nodesx->fd,&nodex,sizeof(NodeX));
 
     /* Create the Node */
 
     node.latoffset=latlong_to_off(nodex.latitude);
     node.lonoffset=latlong_to_off(nodex.longitude);
-    node.firstseg=segmentsx->firstnode[nodesx->gdata[nodex.id]];
+    node.firstseg=segmentsx->firstnode[i];
     node.allow=nodex.allow;
     node.flags=nodex.flags;
 
@@ -722,7 +802,7 @@ void SaveNodeList(NodesX *nodesx,const char *filename,SegmentsX *segmentsx)
 
     /* Write the data */
 
-    WriteFile(fd,&node,sizeof(Node));
+    WriteFileBuffered(fd,&node,sizeof(Node));
 
     if(!((i+1)%10000))
        printf_middle("Writing Nodes: Nodes=%"Pindex_t,i+1);
@@ -730,7 +810,7 @@ void SaveNodeList(NodesX *nodesx,const char *filename,SegmentsX *segmentsx)
 
  /* Close the file */
 
- nodesx->fd=CloseFile(nodesx->fd);
+ nodesx->fd=CloseFileBuffered(nodesx->fd);
 
  /* Finish off the offset indexing and write them out */
 
@@ -739,8 +819,8 @@ void SaveNodeList(NodesX *nodesx,const char *filename,SegmentsX *segmentsx)
  for(;latlonbin<=maxlatlonbins;latlonbin++)
     offsets[latlonbin]=nodesx->number;
 
- SeekFile(fd,sizeof(NodesFile));
- WriteFile(fd,offsets,(nodesx->latbins*nodesx->lonbins+1)*sizeof(index_t));
+ SeekFileBuffered(fd,sizeof(NodesFile));
+ WriteFileBuffered(fd,offsets,(nodesx->latbins*nodesx->lonbins+1)*sizeof(index_t));
 
  free(offsets);
 
@@ -755,10 +835,10 @@ void SaveNodeList(NodesX *nodesx,const char *filename,SegmentsX *segmentsx)
  nodesfile.latzero=nodesx->latzero;
  nodesfile.lonzero=nodesx->lonzero;
 
- SeekFile(fd,0);
- WriteFile(fd,&nodesfile,sizeof(NodesFile));
+ SeekFileBuffered(fd,0);
+ WriteFileBuffered(fd,&nodesfile,sizeof(NodesFile));
 
- CloseFile(fd);
+ CloseFileBuffered(fd);
 
  /* Print the final message */
 
