@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2014 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -32,6 +32,10 @@
 #include "results.h"
 
 
+/*+ To help when debugging +*/
+#define DEBUG 0
+
+
 /* Global variables */
 
 /*+ The option not to print any progress information. +*/
@@ -43,8 +47,8 @@ extern int option_quickest;
 
 /* Local functions */
 
-static index_t FindSuperSegment(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,index_t finish_node,index_t finish_segment);
-static Results *FindSuperRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,index_t start_node,index_t finish_node);
+static index_t FindSuperSegment(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,Profile *profile,index_t finish_node,index_t finish_segment);
+static Results *FindSuperRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,Profile *profile,index_t start_node,index_t finish_node);
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -79,6 +83,10 @@ Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
  Result  *result1,*result2;
  int     force_uturn=0;
 
+#if DEBUG
+ printf("  FindNormalRoute(...,start_node=%"Pindex_t" prev_segment=%"Pindex_t" finish_node=%"Pindex_t")\n",start_node,prev_segment,finish_node);
+#endif
+
  /* Set up the finish conditions */
 
  finish_score=INF_SCORE;
@@ -87,19 +95,19 @@ Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
  if(IsFakeNode(finish_node))
     GetFakeLatLong(finish_node,&finish_lat,&finish_lon);
  else
-    GetLatLong(nodes,finish_node,&finish_lat,&finish_lon);
+    GetLatLong(nodes,finish_node,NULL,&finish_lat,&finish_lon);
 
  /* Create the list of results and insert the first node into the queue */
 
- results=NewResultsList(64);
- queue=NewQueueList();
+ results=NewResultsList(8);
+ queue=NewQueueList(8);
 
  results->start_node=start_node;
  results->prev_segment=prev_segment;
 
  result1=InsertResult(results,results->start_node,results->prev_segment);
 
- InsertInQueue(queue,result1);
+ InsertInQueue(queue,result1,0);
 
  /* Check for barrier at start waypoint - must perform U-turn */
 
@@ -162,7 +170,15 @@ Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
 
        /* must obey one-way restrictions (unless profile allows) */
        if(profile->oneway && IsOnewayTo(segmentp,node1))
-          goto endloop;
+         {
+          if(profile->allow!=Transports_Bicycle)
+             goto endloop;
+
+          wayp=LookupWay(ways,segmentp->way,1);
+
+          if(!(wayp->props&Properties_CycleBothWays))
+             goto endloop;
+         }
 
        if(IsFakeNode(node1) || IsFakeNode(node2))
          {
@@ -220,13 +236,10 @@ Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
           goto endloop;
 
        for(i=1;i<Property_Count;i++)
-          if(ways->file.props & PROPERTIES(i))
-            {
-             if(wayp->props & PROPERTIES(i))
-                segment_pref*=profile->props_yes[i];
-             else
-                segment_pref*=profile->props_no[i];
-            }
+          if(wayp->props & PROPERTIES(i))
+             segment_pref*=profile->props_yes[i];
+          else
+             segment_pref*=profile->props_no[i];
 
        /* profile preferences must allow this highway */
        if(segment_pref==0)
@@ -254,35 +267,23 @@ Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
           result2=InsertResult(results,node2,seg2);
           result2->prev=result1;
           result2->score=cumulative_score;
-
-          if(node2==finish_node)
-            {
-             finish_score=cumulative_score;
-             finish_result=result2;
-            }
-          else
-            {
-             result2->sortby=result2->score;
-             InsertInQueue(queue,result2);
-            }
          }
        else if(cumulative_score<result2->score) /* New score for end node/segment combination is better */
          {
           result2->prev=result1;
           result2->score=cumulative_score;
           result2->segment=seg2;
-
-          if(node2==finish_node)
-            {
-             finish_score=cumulative_score;
-             finish_result=result2;
-            }
-          else
-            {
-             result2->sortby=result2->score;
-             InsertInQueue(queue,result2);
-            }
          }
+       else
+          goto endloop;
+
+       if(node2==finish_node)
+         {
+          finish_score=cumulative_score;
+          finish_result=result2;
+         }
+       else
+          InsertInQueue(queue,result2,result2->score);
 
       endloop:
 
@@ -306,12 +307,27 @@ Results *FindNormalRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
 
  if(!finish_result)
    {
+#if DEBUG
+    printf("    Failed\n");
+#endif
+
     FreeResultsList(results);
     return(NULL);
    }
 
  FixForwardRoute(results,finish_result);
 
+#if DEBUG
+ Result *r=FindResult(results,results->start_node,results->prev_segment);
+
+ while(r)
+   {
+    printf("    node=%"Pindex_t" segment=%"Pindex_t" score=%f\n",r->node,r->segment,r->score);
+
+    r=r->next;
+   }
+#endif
+
  return(results);
 }
 
@@ -346,8 +362,14 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
  Result  *result1,*result2,*result3,*result4;
  int     force_uturn=0;
 
+#if DEBUG
+ printf("  FindMiddleRoute(...,[begin has %d nodes],[end has %d nodes])\n",begin->number,end->number);
+#endif
+
+#if !DEBUG
  if(!option_quiet)
     printf_first("Routing: Super-Nodes checked = 0");
+#endif
 
  /* Set up the finish conditions */
 
@@ -357,12 +379,12 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
  if(IsFakeNode(end->finish_node))
     GetFakeLatLong(end->finish_node,&finish_lat,&finish_lon);
  else
-    GetLatLong(nodes,end->finish_node,&finish_lat,&finish_lon);
+    GetLatLong(nodes,end->finish_node,NULL,&finish_lat,&finish_lon);
 
  /* Create the list of results and insert the first node into the queue */
 
- results=NewResultsList(65536);
- queue=NewQueueList();
+ results=NewResultsList(20);
+ queue=NewQueueList(12);
 
  results->start_node=begin->start_node;
  results->prev_segment=begin->prev_segment;
@@ -373,7 +395,7 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
        results->prev_segment=NO_SEGMENT;
     else
       {
-       index_t superseg=FindSuperSegment(nodes,segments,ways,relations,begin->start_node,begin->prev_segment);
+       index_t superseg=FindSuperSegment(nodes,segments,ways,relations,profile,begin->start_node,begin->prev_segment);
 
        results->prev_segment=superseg;
       }
@@ -392,7 +414,7 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
        !IsFakeNode(result3->node) && IsSuperNode(LookupNode(nodes,result3->node,5)))
       {
        Result *result5=result1;
-       index_t superseg=FindSuperSegment(nodes,segments,ways,relations,result3->node,result3->segment);
+       index_t superseg=FindSuperSegment(nodes,segments,ways,relations,profile,result3->node,result3->segment);
 
        if(superseg!=result3->segment)
          {
@@ -407,9 +429,8 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
           result2->prev=result5;
 
           result2->score=result3->score;
-          result2->sortby=result3->score;
 
-          InsertInQueue(queue,result2);
+          InsertInQueue(queue,result2,result3->score);
 
           if((result4=FindResult(end,result2->node,result2->segment)))
             {
@@ -426,7 +447,7 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
    }
 
  if(begin->number==1)
-    InsertInQueue(queue,result1);
+    InsertInQueue(queue,result1,0);
 
  /* Check for barrier at start waypoint - must perform U-turn */
 
@@ -478,7 +499,15 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
 
        /* must obey one-way restrictions (unless profile allows) */
        if(profile->oneway && IsOnewayTo(segmentp,node1))
-          goto endloop;
+         {
+          if(profile->allow!=Transports_Bicycle)
+             goto endloop;
+
+          wayp=LookupWay(ways,segmentp->way,1);
+
+          if(!(wayp->props&Properties_CycleBothWays))
+             goto endloop;
+         }
 
        seg2=IndexSegment(segments,segmentp); /* segment cannot be a fake segment (must be a super-segment) */
 
@@ -520,13 +549,10 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
           goto endloop;
 
        for(i=1;i<Property_Count;i++)
-          if(ways->file.props & PROPERTIES(i))
-            {
-             if(wayp->props & PROPERTIES(i))
-                segment_pref*=profile->props_yes[i];
-             else
-                segment_pref*=profile->props_no[i];
-            }
+          if(wayp->props & PROPERTIES(i))
+             segment_pref*=profile->props_yes[i];
+          else
+             segment_pref*=profile->props_no[i];
 
        /* profile preferences must allow this highway */
        if(segment_pref==0)
@@ -558,67 +584,46 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
           result2=InsertResult(results,node2,seg2);
           result2->prev=result1;
           result2->score=cumulative_score;
-
-          if((result3=FindResult(end,node2,seg2)))
-            {
-             if((result2->score+result3->score)<finish_score)
-               {
-                finish_score=result2->score+result3->score;
-                finish_result=result2;
-               }
-            }
-          else
-            {
-             double lat,lon;
-             distance_t direct;
-
-             GetLatLong(nodes,node2,&lat,&lon); /* node2 cannot be a fake node (must be a super-node) */
-
-             direct=Distance(lat,lon,finish_lat,finish_lon);
-
-             if(option_quickest==0)
-                result2->sortby=result2->score+(score_t)direct/profile->max_pref;
-             else
-                result2->sortby=result2->score+(score_t)distance_speed_to_duration(direct,profile->max_speed)/profile->max_pref;
-
-             if(result2->sortby<finish_score)
-                InsertInQueue(queue,result2);
-            }
          }
        else if(cumulative_score<result2->score) /* New end node/segment pair is better */
          {
           result2->prev=result1;
           result2->score=cumulative_score;
+         }
+       else
+          goto endloop;
 
-          if((result3=FindResult(end,node2,seg2)))
+       if((result3=FindResult(end,node2,seg2)))
+         {
+          if((result2->score+result3->score)<finish_score)
             {
-             if((result2->score+result3->score)<finish_score)
-               {
-                finish_score=result2->score+result3->score;
-                finish_result=result2;
-               }
-            }
-          else if(result2->score<finish_score)
-            {
-             double lat,lon;
-             distance_t direct;
-
-             GetLatLong(nodes,node2,&lat,&lon); /* node2 cannot be a fake node (must be a super-node) */
-
-             direct=Distance(lat,lon,finish_lat,finish_lon);
-
-             if(option_quickest==0)
-                result2->sortby=result2->score+(score_t)direct/profile->max_pref;
-             else
-                result2->sortby=result2->score+(score_t)distance_speed_to_duration(direct,profile->max_speed)/profile->max_pref;
-
-             if(result2->sortby<finish_score)
-                InsertInQueue(queue,result2);
+             finish_score=result2->score+result3->score;
+             finish_result=result2;
             }
          }
+       else
+         {
+          double lat,lon;
+          distance_t direct;
+          score_t potential_score;
 
+          GetLatLong(nodes,node2,node2p,&lat,&lon); /* node2 cannot be a fake node (must be a super-node) */
+
+          direct=Distance(lat,lon,finish_lat,finish_lon);
+
+          if(option_quickest==0)
+             potential_score=result2->score+(score_t)direct/profile->max_pref;
+          else
+             potential_score=result2->score+(score_t)distance_speed_to_duration(direct,profile->max_speed)/profile->max_pref;
+
+          if(potential_score<finish_score)
+             InsertInQueue(queue,result2,potential_score);
+         }
+
+#if !DEBUG
        if(!option_quiet && !(results->number%1000))
           printf_middle("Routing: Super-Nodes checked = %d",results->number);
+#endif
 
       endloop:
 
@@ -626,8 +631,10 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
       }
    }
 
+#if !DEBUG
  if(!option_quiet)
     printf_last("Routing: Super-Nodes checked = %d",results->number);
+#endif
 
  FreeQueueList(queue);
 
@@ -635,6 +642,10 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
 
  if(!finish_result)
    {
+#if DEBUG
+    printf("    Failed\n");
+#endif
+
     FreeResultsList(results);
     return(NULL);
    }
@@ -653,6 +664,17 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
 
  FixForwardRoute(results,finish_result);
 
+#if DEBUG
+ Result *r=FindResult(results,results->start_node,results->prev_segment);
+
+ while(r)
+   {
+    printf("    node=%"Pindex_t" segment=%"Pindex_t" score=%f\n",r->node,r->segment,r->score);
+
+    r=r->next;
+   }
+#endif
+
  return(results);
 }
 
@@ -670,12 +692,14 @@ Results *FindMiddleRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
 
   Relations *relations The set of relations to use.
 
+  Profile *profile The profile containing the transport type, speeds and allowed highways.
+
   index_t finish_node The super-node that the route ends at.
 
   index_t finish_segment The segment that the route ends with.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static index_t FindSuperSegment(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,index_t finish_node,index_t finish_segment)
+static index_t FindSuperSegment(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,Profile *profile,index_t finish_node,index_t finish_segment)
 {
  Node *supernodep;
  Segment *supersegmentp;
@@ -703,7 +727,7 @@ static index_t FindSuperSegment(Nodes *nodes,Segments *segments,Ways *ways,Relat
 
        start_node=OtherNode(supersegmentp,finish_node);
 
-       results=FindSuperRoute(nodes,segments,ways,relations,start_node,finish_node);
+       results=FindSuperRoute(nodes,segments,ways,relations,profile,start_node,finish_node);
 
        if(!results)
           continue;
@@ -741,28 +765,34 @@ static index_t FindSuperSegment(Nodes *nodes,Segments *segments,Ways *ways,Relat
 
   Relations *relations The set of relations to use.
 
+  Profile *profile The profile containing the transport type, speeds and allowed highways.
+
   index_t start_node The start node.
 
   index_t finish_node The finish node.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static Results *FindSuperRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,index_t start_node,index_t finish_node)
+static Results *FindSuperRoute(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,Profile *profile,index_t start_node,index_t finish_node)
 {
  Results *results;
  Queue   *queue;
  Result  *result1,*result2;
 
+#if DEBUG
+ printf("    FindSuperRoute(...,start_node=%"Pindex_t" finish_node=%"Pindex_t")\n",start_node,finish_node);
+#endif
+
  /* Create the list of results and insert the first node into the queue */
 
- results=NewResultsList(64);
- queue=NewQueueList();
+ results=NewResultsList(8);
+ queue=NewQueueList(8);
 
  results->start_node=start_node;
  results->prev_segment=NO_SEGMENT;
 
  result1=InsertResult(results,results->start_node,results->prev_segment);
 
- InsertInQueue(queue,result1);
+ InsertInQueue(queue,result1,0);
 
  /* Loop across all nodes in the queue */
 
@@ -793,7 +823,17 @@ static Results *FindSuperRoute(Nodes *nodes,Segments *segments,Ways *ways,Relati
 
        /* must obey one-way restrictions */
        if(IsOnewayTo(segmentp,node1))
-          goto endloop;
+         {
+          Way *wayp;
+
+          if(profile->allow!=Transports_Bicycle)
+             goto endloop;
+
+          wayp=LookupWay(ways,segmentp->way,1);
+
+          if(!(wayp->props&Properties_CycleBothWays))
+             goto endloop;
+         }
 
        seg2=IndexSegment(segments,segmentp);
 
@@ -819,23 +859,18 @@ static Results *FindSuperRoute(Nodes *nodes,Segments *segments,Ways *ways,Relati
           result2=InsertResult(results,node2,seg2);
           result2->prev=result1;
           result2->score=cumulative_score;
-          result2->sortby=result2->score;
-
-          /* don't route beyond a super-node. */
-          if(!IsSuperNode(node2p))
-             InsertInQueue(queue,result2);
          }
        else if(cumulative_score<result2->score) /* New score for end node/segment combination is better */
          {
           result2->prev=result1;
           result2->segment=seg2;
           result2->score=cumulative_score;
-          result2->sortby=result2->score;
-
-          /* don't route beyond a super-node. */
-          if(!IsSuperNode(node2p))
-             InsertInQueue(queue,result2);
          }
+       else goto endloop;
+
+       /* don't route beyond a super-node. */
+       if(!IsSuperNode(node2p))
+          InsertInQueue(queue,result2,result2->score);
 
       endloop:
 
@@ -845,6 +880,17 @@ static Results *FindSuperRoute(Nodes *nodes,Segments *segments,Ways *ways,Relati
 
  FreeQueueList(queue);
 
+#if DEBUG
+ Result *r=FindResult(results,results->start_node,results->prev_segment);
+
+ while(r)
+   {
+    printf("      node=%"Pindex_t" segment=%"Pindex_t" score=%f\n",r->node,r->segment,r->score);
+
+    r=r->next;
+   }
+#endif
+
  return(results);
 }
 
@@ -869,28 +915,30 @@ static Results *FindSuperRoute(Nodes *nodes,Segments *segments,Ways *ways,Relati
   index_t prev_segment The previous segment before the start node.
 
   index_t finish_node The finish node.
-
-  int *nsuper Returns the number of super-nodes seen.
   ++++++++++++++++++++++++++++++++++++++*/
 
-Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,Profile *profile,index_t start_node,index_t prev_segment,index_t finish_node,int *nsuper)
+Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,Profile *profile,index_t start_node,index_t prev_segment,index_t finish_node)
 {
  Results *results;
  Queue   *queue;
  Result  *result1,*result2;
- int     found_finish=0,force_uturn=0;
+ int     nsuper=0,force_uturn=0;
+
+#if DEBUG
+ printf("  FindStartRoutes(...,start_node=%"Pindex_t" prev_segment=%"Pindex_t" finish_node=%"Pindex_t")\n",start_node,prev_segment,finish_node);
+#endif
 
  /* Create the list of results and insert the first node into the queue */
 
- results=NewResultsList(64);
- queue=NewQueueList();
+ results=NewResultsList(8);
+ queue=NewQueueList(8);
 
  results->start_node=start_node;
  results->prev_segment=prev_segment;
 
  result1=InsertResult(results,results->start_node,results->prev_segment);
 
- InsertInQueue(queue,result1);
+ InsertInQueue(queue,result1,0);
 
  /* Check for barrier at start waypoint - must perform U-turn */
 
@@ -949,7 +997,15 @@ Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
 
        /* must obey one-way restrictions (unless profile allows) */
        if(profile->oneway && IsOnewayTo(segmentp,node1))
-          goto endloop;
+         {
+          if(profile->allow!=Transports_Bicycle)
+             goto endloop;
+
+          wayp=LookupWay(ways,segmentp->way,1);
+
+          if(!(wayp->props&Properties_CycleBothWays))
+             goto endloop;
+         }
 
        if(IsFakeNode(node1) || IsFakeNode(node2))
          {
@@ -1000,13 +1056,10 @@ Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
           goto endloop;
 
        for(i=1;i<Property_Count;i++)
-          if(ways->file.props & PROPERTIES(i))
-            {
-             if(wayp->props & PROPERTIES(i))
-                segment_pref*=profile->props_yes[i];
-             else
-                segment_pref*=profile->props_no[i];
-            }
+          if(wayp->props & PROPERTIES(i))
+             segment_pref*=profile->props_yes[i];
+          else
+             segment_pref*=profile->props_no[i];
 
        /* profile preferences must allow this highway */
        if(segment_pref==0)
@@ -1035,28 +1088,33 @@ Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
           result2->score=cumulative_score;
 
           if(node2p && IsSuperNode(node2p))
-             (*nsuper)++;
-
-          if(node2p && !IsSuperNode(node2p))
-            {
-             result2->sortby=result2->score;
-             InsertInQueue(queue,result2);
-            }
+             nsuper++;
 
           if(node2==finish_node)
-             found_finish=1;
+             nsuper++;
+
+          if(node2==finish_node)
+            {
+             results->finish_node=finish_node;
+             results->last_segment=seg2;
+            }
          }
        else if(cumulative_score<result2->score) /* New score for end node/segment combination is better */
          {
           result2->prev=result1;
           result2->score=cumulative_score;
 
-          if(node2p && !IsSuperNode(node2p))
+          if(node2==finish_node)
             {
-             result2->sortby=result2->score;
-             InsertInQueue(queue,result2);
+             results->finish_node=finish_node;
+             results->last_segment=seg2;
             }
          }
+       else
+          goto endloop;
+
+       if(node2p && !IsSuperNode(node2p))
+          InsertInQueue(queue,result2,result2->score);
 
       endloop:
 
@@ -1078,12 +1136,304 @@ Results *FindStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *r
 
  /* Check it worked */
 
- if(results->number==1 || (*nsuper==0 && found_finish==0))
+ if(results->number==1 || nsuper==0)
    {
+#if DEBUG
+    printf("    Failed (%d results, %d super)\n",results->number,nsuper);
+#endif
+
     FreeResultsList(results);
     return(NULL);
    }
 
+#if DEBUG
+ Result *s=FirstResult(results);
+
+ while(s)
+   {
+    if(s->node==finish_node)
+      {
+       Result *r=FindResult(results,s->node,s->segment);
+
+       printf("    -------- route to finish node\n");
+
+       while(r)
+         {
+          printf("    node=%"Pindex_t" segment=%"Pindex_t" score=%f\n",r->node,r->segment,r->score);
+
+          r=r->prev;
+         }
+      }
+
+    if(!IsFakeNode(s->node))
+      {
+       Node *n=LookupNode(nodes,s->node,1);
+
+       if(IsSuperNode(n))
+         {
+          Result *r=FindResult(results,s->node,s->segment);
+
+          printf("    -------- route to super node\n");
+
+          while(r)
+            {
+             printf("    node=%"Pindex_t" segment=%"Pindex_t" score=%f\n",r->node,r->segment,r->score);
+
+             r=r->prev;
+            }
+         }
+      }
+
+    s=NextResult(results,s);
+   }
+#endif
+
+ return(results);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Continue finding routes from a set of super-nodes to a finish point.
+
+  Results *ExtendStartRoutes Returns the set of results that were passed in.
+
+  Nodes *nodes The set of nodes to use.
+
+  Segments *segments The set of segments to use.
+
+  Ways *ways The set of ways to use.
+
+  Relations *relations The set of relations to use.
+
+  Profile *profile The profile containing the transport type, speeds and allowed highways.
+
+  Results *begin The partial set of routes already computed.
+
+  index_t finish_node The finish node.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+Results *ExtendStartRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,Profile *profile,Results *begin,index_t finish_node)
+{
+ Results *results=begin;
+ Queue   *queue;
+ Result  *result1,*result2,*result3;
+ Result  *finish_result=NULL;
+ score_t finish_score=INF_SCORE;
+
+#if DEBUG
+ printf("  ExtendStartRoutes(...,[begin has %d nodes],finish_node=%"Pindex_t")\n",begin->number,finish_node);
+#endif
+
+ /* Check the list of results and insert the super nodes into the queue */
+
+ queue=NewQueueList(8);
+
+ result3=FirstResult(begin);
+
+ while(result3)
+   {
+    if(result3->node==finish_node)
+       if(result3->score<finish_score)
+         {
+          finish_score=result3->score;
+          finish_result=result3;
+         }
+
+    if(!IsFakeNode(result3->node))
+       if(IsSuperNode(LookupNode(nodes,result3->node,5)))
+          InsertInQueue(queue,result3,result3->score);
+
+    result3=NextResult(results,result3);
+   }
+
+ /* Loop across all nodes in the queue */
+
+ while((result1=PopFromQueue(queue)))
+   {
+    Node *node1p=NULL;
+    Segment *segmentp;
+    index_t node1,seg1,seg1r;
+    index_t turnrelation=NO_RELATION;
+
+    /* score must be better than current best score */
+    if(result1->score>=finish_score)
+       continue;
+
+    node1=result1->node;
+    seg1=result1->segment;
+
+    if(IsFakeSegment(seg1))
+       seg1r=IndexRealSegment(seg1);
+    else
+       seg1r=seg1;
+
+    if(!IsFakeNode(node1))
+       node1p=LookupNode(nodes,node1,1);
+
+    /* lookup if a turn restriction applies */
+    if(profile->turns && node1p && IsTurnRestrictedNode(node1p))
+       turnrelation=FindFirstTurnRelation2(relations,node1,seg1r);
+
+    /* Loop across all segments */
+
+    if(IsFakeNode(node1))
+       segmentp=FirstFakeSegment(node1);
+    else
+       segmentp=FirstSegment(segments,node1p,1);
+
+    while(segmentp)
+      {
+       Node *node2p=NULL;
+       Way *wayp;
+       index_t node2,seg2,seg2r;
+       score_t segment_pref,segment_score,cumulative_score;
+       int i;
+
+       node2=OtherNode(segmentp,node1); /* need this here because we use node2 at the end of the loop */
+
+       /* must be a normal segment */
+       if(!IsNormalSegment(segmentp))
+          goto endloop;
+
+       /* must obey one-way restrictions (unless profile allows) */
+       if(profile->oneway && IsOnewayTo(segmentp,node1))
+         {
+          if(profile->allow!=Transports_Bicycle)
+             goto endloop;
+
+          wayp=LookupWay(ways,segmentp->way,1);
+
+          if(!(wayp->props&Properties_CycleBothWays))
+             goto endloop;
+         }
+
+       if(IsFakeNode(node1) || IsFakeNode(node2))
+         {
+          seg2 =IndexFakeSegment(segmentp);
+          seg2r=IndexRealSegment(seg2);
+         }
+       else
+         {
+          seg2 =IndexSegment(segments,segmentp);
+          seg2r=seg2;
+         }
+
+       /* must not perform U-turn (unless profile allows) */
+       if(profile->turns && (seg1==seg2 || seg1==seg2r || seg1r==seg2 || (seg1r==seg2r && IsFakeUTurn(seg1,seg2))))
+          goto endloop;
+
+       /* must obey turn relations */
+       if(turnrelation!=NO_RELATION && !IsTurnAllowed(relations,turnrelation,node1,seg1r,seg2r,profile->allow))
+          goto endloop;
+
+       wayp=LookupWay(ways,segmentp->way,1);
+
+       /* mode of transport must be allowed on the highway */
+       if(!(wayp->allow&profile->allow))
+          goto endloop;
+
+       /* must obey weight restriction (if exists) */
+       if(wayp->weight && wayp->weight<profile->weight)
+          goto endloop;
+
+       /* must obey height/width/length restriction (if exists) */
+       if((wayp->height && wayp->height<profile->height) ||
+          (wayp->width  && wayp->width <profile->width ) ||
+          (wayp->length && wayp->length<profile->length))
+          goto endloop;
+
+       segment_pref=profile->highway[HIGHWAY(wayp->type)];
+
+       /* highway preferences must allow this highway */
+       if(segment_pref==0)
+          goto endloop;
+
+       for(i=1;i<Property_Count;i++)
+          if(wayp->props & PROPERTIES(i))
+             segment_pref*=profile->props_yes[i];
+          else
+             segment_pref*=profile->props_no[i];
+
+       /* profile preferences must allow this highway */
+       if(segment_pref==0)
+          goto endloop;
+
+       if(!IsFakeNode(node2))
+          node2p=LookupNode(nodes,node2,2);
+
+       /* mode of transport must be allowed through node2 unless it is the final node */
+       if(node2p && node2!=finish_node && !(node2p->allow&profile->allow))
+          goto endloop;
+
+       if(option_quickest==0)
+          segment_score=(score_t)DISTANCE(segmentp->distance)/segment_pref;
+       else
+          segment_score=(score_t)Duration(segmentp,wayp,profile)/segment_pref;
+
+       cumulative_score=result1->score+segment_score;
+
+       /* score must be better than current best score */
+       if(cumulative_score>=finish_score)
+          goto endloop;
+
+       result2=FindResult(results,node2,seg2);
+
+       if(!result2) /* New end node/segment combination */
+         {
+          result2=InsertResult(results,node2,seg2);
+          result2->prev=result1;
+          result2->score=cumulative_score;
+         }
+       else if(cumulative_score<result2->score) /* New score for end node/segment combination is better */
+         {
+          result2->prev=result1;
+          result2->score=cumulative_score;
+         }
+       else
+          goto endloop;
+
+       if(node2==finish_node)
+         {
+          if(cumulative_score<finish_score)
+            {
+             finish_score=cumulative_score;
+             finish_result=result2;
+            }
+         }
+       else
+          InsertInQueue(queue,result2,result2->score);
+
+      endloop:
+
+       if(IsFakeNode(node1))
+          segmentp=NextFakeSegment(segmentp,node1);
+       else if(IsFakeNode(node2))
+          segmentp=NULL; /* cannot call NextSegment() with a fake segment */
+       else
+         {
+          segmentp=NextSegment(segments,segmentp,node1);
+
+          if(!segmentp && IsFakeNode(finish_node))
+             segmentp=ExtraFakeSegment(node1,finish_node);
+         }
+      }
+   }
+
+ FreeQueueList(queue);
+
+ FixForwardRoute(results,finish_result);
+
+#if DEBUG
+ Result *r=FindResult(results,results->start_node,results->prev_segment);
+
+ while(r)
+   {
+    printf("    node=%"Pindex_t" segment=%"Pindex_t" score=%f\n",r->node,r->segment,r->score);
+
+    r=r->next;
+   }
+#endif
+
  return(results);
 }
 
@@ -1112,16 +1462,20 @@ Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *
  Queue   *queue;
  Result  *result1,*result2,*result3;
 
+#if DEBUG
+ printf("  FindFinishRoutes(...,finish_node=%"Pindex_t")\n",finish_node);
+#endif
+
  /* Create the results and insert the finish node into the queue */
 
- results=NewResultsList(64);
- queue=NewQueueList();
+ results=NewResultsList(8);
+ queue=NewQueueList(8);
 
  results->finish_node=finish_node;
 
  result1=InsertResult(results,finish_node,NO_SEGMENT);
 
- InsertInQueue(queue,result1);
+ InsertInQueue(queue,result1,0);
 
  /* Loop across all nodes in the queue */
 
@@ -1172,7 +1526,15 @@ Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *
 
        /* must obey one-way restrictions (unless profile allows) */
        if(profile->oneway && IsOnewayFrom(segmentp,node1)) /* working backwards => disallow oneway *from* node1 */
-          goto endloop;
+         {
+          if(profile->allow!=Transports_Bicycle)
+             goto endloop;
+
+          wayp=LookupWay(ways,segmentp->way,1);
+
+          if(!(wayp->props&Properties_CycleBothWays))
+             goto endloop;
+         }
 
        node2=OtherNode(segmentp,node1);
 
@@ -1223,13 +1585,10 @@ Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *
           goto endloop;
 
        for(i=1;i<Property_Count;i++)
-          if(ways->file.props & PROPERTIES(i))
-            {
-             if(wayp->props & PROPERTIES(i))
-                segment_pref*=profile->props_yes[i];
-             else
-                segment_pref*=profile->props_no[i];
-            }
+          if(wayp->props & PROPERTIES(i))
+             segment_pref*=profile->props_yes[i];
+          else
+             segment_pref*=profile->props_no[i];
 
        /* profile preferences must allow this highway */
        if(segment_pref==0)
@@ -1256,24 +1615,17 @@ Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *
           result2=InsertResult(results,node2,seg2);
           result2->next=result1;   /* working backwards */
           result2->score=cumulative_score;
-
-          if(IsFakeNode(node1) || !IsSuperNode(node1p))
-            {
-             result2->sortby=result2->score;
-             InsertInQueue(queue,result2);
-            }
          }
        else if(cumulative_score<result2->score) /* New end node is better */
          {
           result2->next=result1; /* working backwards */
           result2->score=cumulative_score;
-
-          if(IsFakeNode(node1) || !IsSuperNode(node1p))
-            {
-             result2->sortby=result2->score;
-             InsertInQueue(queue,result2);
-            }
          }
+       else
+          goto endloop;
+
+       if(IsFakeNode(node1) || !IsSuperNode(node1p))
+          InsertInQueue(queue,result2,result2->score);
 
       endloop:
 
@@ -1290,13 +1642,17 @@ Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *
 
  if(results->number==1)
    {
+#if DEBUG
+    printf("    Failed\n");
+#endif
+
     FreeResultsList(results);
     return(NULL);
    }
 
  /* Create a results structure with the node at the end of the segment opposite the start */
 
- results2=NewResultsList(64);
+ results2=NewResultsList(8);
 
  results2->finish_node=results->finish_node;
 
@@ -1333,6 +1689,34 @@ Results *FindFinishRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *
 
  FreeResultsList(results);
 
+#if DEBUG
+ Result *s=FirstResult(results2);
+
+ while(s)
+   {
+    if(!IsFakeNode(s->node))
+      {
+       Node *n=LookupNode(nodes,s->node,1);
+
+       if(IsSuperNode(n))
+         {
+          Result *r=FindResult(results2,s->node,s->segment);
+
+          printf("    --------\n");
+
+          while(r)
+            {
+             printf("    node=%"Pindex_t" segment=%"Pindex_t" score=%f\n",r->node,r->segment,r->score);
+
+             r=r->next;
+            }
+         }
+      }
+
+    s=NextResult(results2,s);
+   }
+#endif
+
  return(results2);
 }
 
@@ -1362,7 +1746,11 @@ Results *CombineRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *rel
  Result *midres,*comres1;
  Results *combined;
 
- combined=NewResultsList(256);
+#if DEBUG
+ printf("  CombineRoutes(...,[begin has %d nodes],[middle has %d nodes])\n",begin->number,middle->number);
+#endif
+
+ combined=NewResultsList(10);
 
  combined->start_node=begin->start_node;
  combined->prev_segment=begin->prev_segment;
@@ -1398,7 +1786,7 @@ Results *CombineRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *rel
 
        comres2=InsertResult(combined,begres->node,begres->segment);
 
-       comres2->score=begres->score+comres1->score;
+       comres2->score=begres->score;
        comres2->prev=comres1;
 
        begres=begres->next;
@@ -1445,7 +1833,7 @@ Results *CombineRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *rel
 
           comres2=InsertResult(combined,result->node,result->segment);
 
-          comres2->score=result->score+comres1->score;
+          comres2->score=midres->score+result->score;
           comres2->prev=comres1;
 
           result=result->next;
@@ -1463,6 +1851,17 @@ Results *CombineRoutes(Nodes *nodes,Segments *segments,Ways *ways,Relations *rel
 
  FixForwardRoute(combined,comres1);
 
+#if DEBUG
+ Result *r=FindResult(combined,combined->start_node,combined->prev_segment);
+
+ while(r)
+   {
+    printf("    node=%"Pindex_t" segment=%"Pindex_t" score=%f\n",r->node,r->segment,r->score);
+
+    r=r->next;
+   }
+#endif
+
  return(combined);
 }
 
