@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2014 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -61,7 +61,7 @@
 extern int option_quickest;
 
 /*+ The options to select the format of the output. +*/
-extern int option_html,option_gpx_track,option_gpx_route,option_text,option_text_all;
+extern int option_html,option_gpx_track,option_gpx_route,option_text,option_text_all,option_stdout;
 
 
 /* Local variables */
@@ -88,7 +88,7 @@ static char junction_other_way[Highway_Count][Highway_Count]=
 /*++++++++++++++++++++++++++++++++++++++
   Print the optimum route between two nodes.
 
-  Results **results The set of results to print (some may be NULL - ignore them).
+  Results **results The set of results to print (consecutive in array even if not consecutive waypoints).
 
   int nresults The number of results in the list.
 
@@ -109,84 +109,100 @@ void PrintRoute(Results **results,int nresults,Nodes *nodes,Segments *segments,W
  distance_t cum_distance=0;
  duration_t cum_duration=0;
 
- int point=1;
+ int point=0;
  int segment_count=0,route_count=0;
  int point_count=0;
  int roundabout=0;
 
  /* Open the files */
 
- if(option_quickest==0)
+ if(option_stdout)
    {
-    /* Print the result for the shortest route */
-
     if(option_html)
-       htmlfile    =fopen("shortest.html","w");
+       htmlfile    =stdout;
     if(option_gpx_track)
-       gpxtrackfile=fopen("shortest-track.gpx","w");
+       gpxtrackfile=stdout;
     if(option_gpx_route)
-       gpxroutefile=fopen("shortest-route.gpx","w");
+       gpxroutefile=stdout;
     if(option_text)
-       textfile    =fopen("shortest.txt","w");
+       textfile    =stdout;
     if(option_text_all)
-       textallfile =fopen("shortest-all.txt","w");
-
-    if(option_html && !htmlfile)
-       fprintf(stderr,"Warning: Cannot open file 'shortest.html' for writing [%s].\n",strerror(errno));
-    if(option_gpx_track && !gpxtrackfile)
-       fprintf(stderr,"Warning: Cannot open file 'shortest-track.gpx' for writing [%s].\n",strerror(errno));
-    if(option_gpx_route && !gpxroutefile)
-       fprintf(stderr,"Warning: Cannot open file 'shortest-route.gpx' for writing [%s].\n",strerror(errno));
-    if(option_text && !textfile)
-       fprintf(stderr,"Warning: Cannot open file 'shortest.txt' for writing [%s].\n",strerror(errno));
-    if(option_text_all && !textallfile)
-       fprintf(stderr,"Warning: Cannot open file 'shortest-all.txt' for writing [%s].\n",strerror(errno));
+       textallfile =stdout;
    }
  else
    {
-    /* Print the result for the quickest route */
+    if(option_quickest==0)
+      {
+       /* Print the result for the shortest route */
 
-    if(option_html)
-       htmlfile    =fopen("quickest.html","w");
-    if(option_gpx_track)
-       gpxtrackfile=fopen("quickest-track.gpx","w");
-    if(option_gpx_route)
-       gpxroutefile=fopen("quickest-route.gpx","w");
-    if(option_text)
-       textfile    =fopen("quickest.txt","w");
-    if(option_text_all)
-       textallfile =fopen("quickest-all.txt","w");
+       if(option_html)
+          htmlfile    =fopen("shortest.html","w");
+       if(option_gpx_track)
+          gpxtrackfile=fopen("shortest-track.gpx","w");
+       if(option_gpx_route)
+          gpxroutefile=fopen("shortest-route.gpx","w");
+       if(option_text)
+          textfile    =fopen("shortest.txt","w");
+       if(option_text_all)
+          textallfile =fopen("shortest-all.txt","w");
 
-    if(option_html && !htmlfile)
-       fprintf(stderr,"Warning: Cannot open file 'quickest.html' for writing [%s].\n",strerror(errno));
-    if(option_gpx_track && !gpxtrackfile)
-       fprintf(stderr,"Warning: Cannot open file 'quickest-track.gpx' for writing [%s].\n",strerror(errno));
-    if(option_gpx_route && !gpxroutefile)
-       fprintf(stderr,"Warning: Cannot open file 'quickest-route.gpx' for writing [%s].\n",strerror(errno));
-    if(option_text && !textfile)
-       fprintf(stderr,"Warning: Cannot open file 'quickest.txt' for writing [%s].\n",strerror(errno));
-    if(option_text_all && !textallfile)
-       fprintf(stderr,"Warning: Cannot open file 'quickest-all.txt' for writing [%s].\n",strerror(errno));
+       if(option_html && !htmlfile)
+          fprintf(stderr,"Warning: Cannot open file 'shortest.html' for writing [%s].\n",strerror(errno));
+       if(option_gpx_track && !gpxtrackfile)
+          fprintf(stderr,"Warning: Cannot open file 'shortest-track.gpx' for writing [%s].\n",strerror(errno));
+       if(option_gpx_route && !gpxroutefile)
+          fprintf(stderr,"Warning: Cannot open file 'shortest-route.gpx' for writing [%s].\n",strerror(errno));
+       if(option_text && !textfile)
+          fprintf(stderr,"Warning: Cannot open file 'shortest.txt' for writing [%s].\n",strerror(errno));
+       if(option_text_all && !textallfile)
+          fprintf(stderr,"Warning: Cannot open file 'shortest-all.txt' for writing [%s].\n",strerror(errno));
+      }
+    else
+      {
+       /* Print the result for the quickest route */
+
+       if(option_html)
+          htmlfile    =fopen("quickest.html","w");
+       if(option_gpx_track)
+          gpxtrackfile=fopen("quickest-track.gpx","w");
+       if(option_gpx_route)
+          gpxroutefile=fopen("quickest-route.gpx","w");
+       if(option_text)
+          textfile    =fopen("quickest.txt","w");
+       if(option_text_all)
+          textallfile =fopen("quickest-all.txt","w");
+
+       if(option_html && !htmlfile)
+          fprintf(stderr,"Warning: Cannot open file 'quickest.html' for writing [%s].\n",strerror(errno));
+       if(option_gpx_track && !gpxtrackfile)
+          fprintf(stderr,"Warning: Cannot open file 'quickest-track.gpx' for writing [%s].\n",strerror(errno));
+       if(option_gpx_route && !gpxroutefile)
+          fprintf(stderr,"Warning: Cannot open file 'quickest-route.gpx' for writing [%s].\n",strerror(errno));
+       if(option_text && !textfile)
+          fprintf(stderr,"Warning: Cannot open file 'quickest.txt' for writing [%s].\n",strerror(errno));
+       if(option_text_all && !textallfile)
+          fprintf(stderr,"Warning: Cannot open file 'quickest-all.txt' for writing [%s].\n",strerror(errno));
+      }
    }
 
  /* Print the head of the files */
 
  if(htmlfile)
    {
-    fprintf(htmlfile,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">\n");
-    fprintf(htmlfile,"<HTML>\n");
+    fprintf(htmlfile,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n");
+    fprintf(htmlfile,"<html>\n");
     if(translate_xml_copyright_creator[0] && translate_xml_copyright_creator[1])
        fprintf(htmlfile,"<!-- %s : %s -->\n",translate_xml_copyright_creator[0],translate_xml_copyright_creator[1]);
     if(translate_xml_copyright_source[0] && translate_xml_copyright_source[1])
        fprintf(htmlfile,"<!-- %s : %s -->\n",translate_xml_copyright_source[0],translate_xml_copyright_source[1]);
     if(translate_xml_copyright_license[0] && translate_xml_copyright_license[1])
        fprintf(htmlfile,"<!-- %s : %s -->\n",translate_xml_copyright_license[0],translate_xml_copyright_license[1]);
-    fprintf(htmlfile,"<HEAD>\n");
-    fprintf(htmlfile,"<TITLE>");
+    fprintf(htmlfile,"<head>\n");
+    fprintf(htmlfile,"<title>");
     fprintf(htmlfile,translate_html_title,option_quickest?translate_xml_route_quickest:translate_xml_route_shortest);
-    fprintf(htmlfile,"</TITLE>\n");
-    fprintf(htmlfile,"<META http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n");
-    fprintf(htmlfile,"<STYLE type=\"text/css\">\n");
+    fprintf(htmlfile,"</title>\n");
+    fprintf(htmlfile,"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n");
+    fprintf(htmlfile,"<style type=\"text/css\">\n");
     fprintf(htmlfile,"<!--\n");
     fprintf(htmlfile,"   table   {table-layout: fixed; border: none; border-collapse: collapse;}\n");
     fprintf(htmlfile,"   table.c {color: grey; font-size: x-small;} /* copyright */\n");
@@ -204,12 +220,12 @@ void PrintRoute(Results **results,int nresults,Nodes *nodes,Segments *segments,W
     fprintf(htmlfile,"   span.t  {font-variant: small-caps;} /* turn */\n");
     fprintf(htmlfile,"   span.b  {font-variant: small-caps;} /* bearing */\n");
     fprintf(htmlfile,"-->\n");
-    fprintf(htmlfile,"</STYLE>\n");
-    fprintf(htmlfile,"</HEAD>\n");
-    fprintf(htmlfile,"<BODY>\n");
-    fprintf(htmlfile,"<H1>");
+    fprintf(htmlfile,"</style>\n");
+    fprintf(htmlfile,"</head>\n");
+    fprintf(htmlfile,"<body>\n");
+    fprintf(htmlfile,"<h1>");
     fprintf(htmlfile,translate_html_title,option_quickest?translate_xml_route_quickest:translate_xml_route_shortest);
-    fprintf(htmlfile,"</H1>\n");
+    fprintf(htmlfile,"</h1>\n");
     fprintf(htmlfile,"<table>\n");
    }
 
@@ -305,10 +321,7 @@ void PrintRoute(Results **results,int nresults,Nodes *nodes,Segments *segments,W
 
  /* Loop through all the sections of the route and print them */
 
- while(!results[point])
-    point++;
-
- while(point<=nresults)
+ do
    {
     int next_point=point;
     distance_t junc_distance=0;
@@ -346,23 +359,26 @@ void PrintRoute(Results **results,int nresults,Nodes *nodes,Segments *segments,W
        if(IsFakeNode(result->node))
           GetFakeLatLong(result->node,&latitude,&longitude);
        else
-          GetLatLong(nodes,result->node,&latitude,&longitude);
-
-       if(!IsFakeNode(result->node))
+         {
           resultnodep=LookupNode(nodes,result->node,6);
 
+          GetLatLong(nodes,result->node,resultnodep,&latitude,&longitude);
+         }
+
        /* Calculate the next result */
 
        next_result=result->next;
 
        if(!next_result)
-          for(next_point=point+1;next_point<=nresults;next_point++)
-             if(results[next_point])
-               {
-                next_result=FindResult(results[next_point],results[next_point]->start_node,results[next_point]->prev_segment);
-                next_result=next_result->next;
-                break;
-               }
+         {
+          next_point++;
+
+          if(next_point<nresults)
+            {
+             next_result=FindResult(results[next_point],results[next_point]->start_node,results[next_point]->prev_segment);
+             next_result=next_result->next;
+            }
+         }
 
        /* Calculate the information about this segment */
 
@@ -793,6 +809,7 @@ void PrintRoute(Results **results,int nresults,Nodes *nodes,Segments *segments,W
 
     point=next_point;
    }
+ while(point<nresults);
 
  /* Print the tail of the files */
 
@@ -815,8 +832,8 @@ void PrintRoute(Results **results,int nresults,Nodes *nodes,Segments *segments,W
        fprintf(htmlfile,"</table>\n");
       }
 
-    fprintf(htmlfile,"</BODY>\n");
-    fprintf(htmlfile,"</HTML>\n");
+    fprintf(htmlfile,"</body>\n");
+    fprintf(htmlfile,"</html>\n");
    }
 
  if(gpxtrackfile)
@@ -833,14 +850,17 @@ void PrintRoute(Results **results,int nresults,Nodes *nodes,Segments *segments,W
 
  /* Close the files */
 
- if(htmlfile)
-    fclose(htmlfile);
- if(gpxtrackfile)
-    fclose(gpxtrackfile);
- if(gpxroutefile)
-    fclose(gpxroutefile);
- if(textfile)
-    fclose(textfile);
- if(textallfile)
-    fclose(textallfile);
+ if(!option_stdout)
+   {
+    if(htmlfile)
+       fclose(htmlfile);
+    if(gpxtrackfile)
+       fclose(gpxtrackfile);
+    if(gpxroutefile)
+       fclose(gpxroutefile);
+    if(textfile)
+       fclose(textfile);
+    if(textallfile)
+       fclose(textallfile);
+   }
 }
