@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2014 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -22,9 +22,9 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <unistd.h>
 #include <string.h>
 #include <errno.h>
-#include <sys/time.h>
 
 #include "types.h"
 #include "ways.h"
@@ -39,9 +39,11 @@
 
 #include "files.h"
 #include "logging.h"
+#include "errorlogx.h"
 #include "functions.h"
 #include "osmparser.h"
 #include "tagging.h"
+#include "uncompress.h"
 
 
 /* Global variables */
@@ -67,11 +69,10 @@ static void print_usage(int detail,const char *argerr,const char *err);
 
 int main(int argc,char** argv)
 {
- struct timeval start_time;
- NodesX     *Nodes;
- SegmentsX  *Segments,*SuperSegments=NULL,*MergedSegments=NULL;
- WaysX      *Ways;
- RelationsX *Relations;
+ NodesX     *OSMNodes;
+ SegmentsX  *OSMSegments,*SuperSegments=NULL,*MergedSegments=NULL;
+ WaysX      *OSMWays;
+ RelationsX *OSMRelations;
  int         iteration=0,quit=0;
  int         max_iterations=5;
  char       *dirname=NULL,*prefix=NULL,*tagging=NULL,*errorlog=NULL;
@@ -81,7 +82,7 @@ int main(int argc,char** argv)
  int         option_prune_isolated=500,option_prune_short=5,option_prune_straight=3;
  int         arg;
 
- gettimeofday(&start_time,NULL);
+ printf_program_start();
 
  /* Parse the command line arguments */
 
@@ -89,18 +90,28 @@ int main(int argc,char** argv)
    {
     if(!strcmp(argv[arg],"--help"))
        print_usage(1,NULL,NULL);
+    else if(!strncmp(argv[arg],"--dir=",6))
+       dirname=&argv[arg][6];
+    else if(!strncmp(argv[arg],"--prefix=",9))
+       prefix=&argv[arg][9];
     else if(!strncmp(argv[arg],"--sort-ram-size=",16))
        option_filesort_ramsize=atoi(&argv[arg][16]);
 #if defined(USE_PTHREADS) && USE_PTHREADS
     else if(!strncmp(argv[arg],"--sort-threads=",15))
        option_filesort_threads=atoi(&argv[arg][15]);
 #endif
-    else if(!strncmp(argv[arg],"--dir=",6))
-       dirname=&argv[arg][6];
     else if(!strncmp(argv[arg],"--tmpdir=",9))
        option_tmpdirname=&argv[arg][9];
-    else if(!strncmp(argv[arg],"--prefix=",9))
-       prefix=&argv[arg][9];
+    else if(!strncmp(argv[arg],"--tagging=",10))
+       tagging=&argv[arg][10];
+    else if(!strcmp(argv[arg],"--loggable"))
+       option_loggable=1;
+    else if(!strcmp(argv[arg],"--logtime"))
+       option_logtime=1;
+    else if(!strcmp(argv[arg],"--errorlog"))
+       errorlog="error.log";
+    else if(!strncmp(argv[arg],"--errorlog=",11))
+       errorlog=&argv[arg][11];
     else if(!strcmp(argv[arg],"--parse-only"))
        option_parse_only=1;
     else if(!strcmp(argv[arg],"--process-only"))
@@ -111,18 +122,8 @@ int main(int argc,char** argv)
        option_keep=1;
     else if(!strcmp(argv[arg],"--changes"))
        option_changes=1;
-    else if(!strcmp(argv[arg],"--loggable"))
-       option_loggable=1;
-    else if(!strcmp(argv[arg],"--logtime"))
-       option_logtime=1;
-    else if(!strcmp(argv[arg],"--errorlog"))
-       errorlog="error.log";
-    else if(!strncmp(argv[arg],"--errorlog=",11))
-       errorlog=&argv[arg][11];
     else if(!strncmp(argv[arg],"--max-iterations=",17))
        max_iterations=atoi(&argv[arg][17]);
-    else if(!strncmp(argv[arg],"--tagging=",10))
-       tagging=&argv[arg][10];
     else if(!strncmp(argv[arg],"--prune",7))
       {
        if(!strcmp(&argv[arg][7],"-none"))
@@ -153,6 +154,9 @@ int main(int argc,char** argv)
  if(option_filenames && option_process_only)
     print_usage(0,NULL,"Cannot use '--process-only' and filenames at the same time.");
 
+ if(!option_filenames && !option_process_only)
+    print_usage(0,NULL,"File names must be specified unless using '--process-only'");
+
  if(!option_filesort_ramsize)
    {
 #if SLIM
@@ -179,7 +183,7 @@ int main(int argc,char** argv)
        if(!ExistsFile(tagging))
          {
           fprintf(stderr,"Error: The '--tagging' option specifies a file that does not exist.\n");
-          return(1);
+          exit(EXIT_FAILURE);
          }
       }
     else
@@ -191,107 +195,123 @@ int main(int argc,char** argv)
        else
          {
           fprintf(stderr,"Error: The '--tagging' option was not used and the default 'tagging.xml' does not exist.\n");
-          return(1);
+          exit(EXIT_FAILURE);
          }
       }
 
     if(ParseXMLTaggingRules(tagging))
       {
        fprintf(stderr,"Error: Cannot read the tagging rules in the file '%s'.\n",tagging);
-       return(1);
+       exit(EXIT_FAILURE);
       }
    }
 
  /* Create new node, segment, way and relation variables */
 
- Nodes=NewNodeList(option_append||option_changes,option_process_only);
+ OSMNodes=NewNodeList(option_append||option_changes,option_process_only);
 
- Segments=NewSegmentList(option_append||option_changes,option_process_only);
+ OSMWays=NewWayList(option_append||option_changes,option_process_only);
 
- Ways=NewWayList(option_append||option_changes,option_process_only);
-
- Relations=NewRelationList(option_append||option_changes,option_process_only);
+ OSMRelations=NewRelationList(option_append||option_changes,option_process_only);
 
  /* Create the error log file */
 
  if(errorlog)
-    open_errorlog(FileName(dirname,prefix,errorlog),option_append||option_changes||option_process_only);
+    open_errorlog(FileName(dirname,prefix,errorlog),option_append||option_changes||option_process_only,option_keep);
 
  /* Parse the file */
 
 if(!option_process_only)
   {
-   if(option_filenames)
+   for(arg=1;arg<argc;arg++)
      {
-      for(arg=1;arg<argc;arg++)
+      int fd;
+      char *filename,*p;
+
+      if(argv[arg][0]=='-' && argv[arg][1]=='-')
+         continue;
+
+      filename=strcpy(malloc(strlen(argv[arg])+1),argv[arg]);
+
+      fd=OpenFile(filename);
+
+      if((p=strstr(filename,".bz2")) && !strcmp(p,".bz2"))
         {
-         FILE *file;
+         fd=Uncompress_Bzip2(fd);
+         *p=0;
+        }
 
-         if(argv[arg][0]=='-' && argv[arg][1]=='-')
-            continue;
+      if((p=strstr(filename,".gz")) && !strcmp(p,".gz"))
+        {
+         fd=Uncompress_Gzip(fd);
+         *p=0;
+        }
 
-         file=fopen(argv[arg],"rb");
+      if((p=strstr(filename,".xz")) && !strcmp(p,".xz"))
+        {
+         fd=Uncompress_Xz(fd);
+         *p=0;
+        }
 
-         if(!file)
+      if(option_changes)
+        {
+         printf("\nParse OSC Data [%s]\n==============\n\n",filename);
+         fflush(stdout);
+
+         if((p=strstr(filename,".pbf")) && !strcmp(p,".pbf"))
            {
-            fprintf(stderr,"Cannot open file '%s' for reading [%s].\n",argv[arg],strerror(errno));
-            exit(EXIT_FAILURE);
+            logassert(0,"Unable to read a PBF file to apply changes (format does not permit this)");
            }
-
-         if(option_changes)
+         else if((p=strstr(filename,".o5c")) && !strcmp(p,".o5c"))
            {
-            printf("\nParse OSC Data [%s]\n==============\n\n",argv[arg]);
-            fflush(stdout);
-
-            if(ParseOSC(file,Nodes,Segments,Ways,Relations))
+            if(ParseO5CFile(fd,OSMNodes,OSMWays,OSMRelations))
                exit(EXIT_FAILURE);
            }
          else
            {
-            printf("\nParse OSM Data [%s]\n==============\n\n",argv[arg]);
-            fflush(stdout);
-
-            if(ParseOSM(file,Nodes,Segments,Ways,Relations))
+            if(ParseOSCFile(fd,OSMNodes,OSMWays,OSMRelations))
                exit(EXIT_FAILURE);
            }
-
-         fclose(file);
-        }
-     }
-   else
-     {
-      if(option_changes)
-        {
-         printf("\nParse OSC Data\n==============\n\n");
-         fflush(stdout);
-
-         if(ParseOSC(stdin,Nodes,Segments,Ways,Relations))
-            exit(EXIT_FAILURE);
         }
       else
         {
-         printf("\nParse OSM Data\n==============\n\n");
+         printf("\nParse OSM Data [%s]\n==============\n\n",filename);
          fflush(stdout);
 
-         if(ParseOSM(stdin,Nodes,Segments,Ways,Relations))
-            exit(EXIT_FAILURE);
+         if((p=strstr(filename,".pbf")) && !strcmp(p,".pbf"))
+           {
+            if(ParsePBFFile(fd,OSMNodes,OSMWays,OSMRelations))
+               exit(EXIT_FAILURE);
+           }
+         else if((p=strstr(filename,".o5m")) && !strcmp(p,".o5m"))
+           {
+            if(ParseO5MFile(fd,OSMNodes,OSMWays,OSMRelations))
+               exit(EXIT_FAILURE);
+           }
+         else
+           {
+            if(ParseOSMFile(fd,OSMNodes,OSMWays,OSMRelations))
+               exit(EXIT_FAILURE);
+           }
         }
+
+      CloseFile(fd);
+
+      free(filename);
      }
 
    DeleteXMLTaggingRules();
   }
 
- FinishNodeList(Nodes);
- FinishSegmentList(Segments);
- FinishWayList(Ways);
- FinishRelationList(Relations);
+ FinishNodeList(OSMNodes);
+ FinishWayList(OSMWays);
+ FinishRelationList(OSMRelations);
 
  if(option_parse_only)
    {
-    FreeNodeList(Nodes,1);
-    FreeSegmentList(Segments,1);
-    FreeWayList(Ways,1);
-    FreeRelationList(Relations,1);
+    FreeNodeList(OSMNodes,1);
+    FreeWayList(OSMWays,1);
+    FreeRelationList(OSMRelations,1);
 
     return(0);
    }
@@ -302,61 +322,50 @@ if(!option_process_only)
  printf("\nSort OSM Data\n=============\n\n");
  fflush(stdout);
 
- /* Sort the nodes, segments, ways and relations */
+ /* Sort the nodes, ways and relations */
 
- SortNodeList(Nodes);
+ SortNodeList(OSMNodes);
 
- if(option_changes)
-    ApplySegmentChanges(Segments);
+ SortWayList(OSMWays);
 
- SortSegmentList(Segments);
-
- SortWayList(Ways);
-
- SortRelationList(Relations);
+ SortRelationList(OSMRelations);
 
  /* Process the data */
 
  printf("\nProcess OSM Data\n================\n\n");
  fflush(stdout);
 
- /* Extract the way names (must be before using the ways) */
+ /* Remove non-highway nodes by looking through the ways */
 
- ExtractWayNames(Ways,option_keep||option_changes);
+ RemoveNonHighwayNodes(OSMNodes,OSMWays,option_keep||option_changes);
 
- /* Remove bad segments (must be after sorting the nodes, segments and ways) */
+ /* Separate the segments and way names and sort them. */
 
- RemoveBadSegments(Segments,Nodes,Ways,option_keep||option_changes);
+ OSMSegments=SplitWays(OSMWays,OSMNodes,option_keep||option_changes);
 
- /* Remove non-highway nodes (must be after removing the bad segments) */
+ SortWayNames(OSMWays);
 
- RemoveNonHighwayNodes(Nodes,Segments,option_keep||option_changes);
+ SortSegmentList(OSMSegments);
 
- /* Process the route relations and first part of turn relations (must be before compacting the ways) */
+ /* Process the segments and index them */
 
- ProcessRouteRelations(Relations,Ways,option_keep||option_changes);
+ ProcessSegments(OSMSegments,OSMNodes,OSMWays);
 
- ProcessTurnRelations1(Relations,Nodes,Ways,option_keep||option_changes);
+ IndexSegments(OSMSegments,OSMNodes,OSMWays);
 
- /* Measure the segments and replace node/way id with index (must be after removing non-highway nodes) */
+ /* Process the route relations and turn relations (must be before compacting the ways) */
 
- MeasureSegments(Segments,Nodes,Ways);
+ ProcessRouteRelations(OSMRelations,OSMWays,option_keep||option_changes);
+
+ ProcessTurnRelations(OSMRelations,OSMNodes,OSMSegments,OSMWays,option_keep||option_changes);
+
+ /* Compact the ways (must be after processing turn relations) */
+
+ CompactWayList(OSMWays,OSMSegments);
 
  /* Index the segments */
 
- IndexSegments(Segments,Nodes,Ways);
-
- /* Convert the turn relations from ways into nodes */
-
- ProcessTurnRelations2(Relations,Nodes,Segments,Ways);
-
- /* Compact the ways (must be after turn relations 2) */
-
- CompactWayList(Ways,Segments);
-
- /* Index the segments */
-
- IndexSegments(Segments,Nodes,Ways);
+ IndexSegments(OSMSegments,OSMNodes,OSMWays);
 
  /* Prune unwanted nodes/segments. */
 
@@ -365,33 +374,34 @@ if(!option_process_only)
     printf("\nPrune Unneeded Data\n===================\n\n");
     fflush(stdout);
 
-    StartPruning(Nodes,Segments,Ways);
+    StartPruning(OSMNodes,OSMSegments,OSMWays);
 
     if(option_prune_straight)
-       PruneStraightHighwayNodes(Nodes,Segments,Ways,option_prune_straight);
+       PruneStraightHighwayNodes(OSMNodes,OSMSegments,OSMWays,option_prune_straight);
 
     if(option_prune_isolated)
-       PruneIsolatedRegions(Nodes,Segments,Ways,option_prune_isolated);
+       PruneIsolatedRegions(OSMNodes,OSMSegments,OSMWays,option_prune_isolated);
 
     if(option_prune_short)
-       PruneShortSegments(Nodes,Segments,Ways,option_prune_short);
+       PruneShortSegments(OSMNodes,OSMSegments,OSMWays,option_prune_short);
 
-    FinishPruning(Nodes,Segments,Ways);
+    FinishPruning(OSMNodes,OSMSegments,OSMWays);
 
-    /* Remove the pruned nodes and segments and update the indexes */
+    /* Remove the pruned nodes, segments, ways and relations and update the indexes */
 
-    RemovePrunedNodes(Nodes,Segments);
-    RemovePrunedSegments(Segments,Ways);
-    CompactWayList(Ways,Segments);
-    RemovePrunedTurnRelations(Relations,Nodes);
-    IndexSegments(Segments,Nodes,Ways);
+    RemovePrunedNodes(OSMNodes,OSMSegments);
+    RemovePrunedSegments(OSMSegments,OSMWays);
+    CompactWayList(OSMWays,OSMSegments);
+    RemovePrunedTurnRelations(OSMRelations,OSMNodes);
+
+    IndexSegments(OSMSegments,OSMNodes,OSMWays);
    }
 
  /* Repeated iteration on Super-Nodes and Super-Segments */
 
  do
    {
-    int nsuper;
+    index_t nsuper;
 
     printf("\nProcess Super-Data (iteration %d)\n================================%s\n\n",iteration,iteration>9?"=":"");
     fflush(stdout);
@@ -400,13 +410,13 @@ if(!option_process_only)
       {
        /* Select the super-nodes */
 
-       ChooseSuperNodes(Nodes,Segments,Ways);
+       ChooseSuperNodes(OSMNodes,OSMSegments,OSMWays);
 
        /* Select the super-segments */
 
-       SuperSegments=CreateSuperSegments(Nodes,Segments,Ways);
+       SuperSegments=CreateSuperSegments(OSMNodes,OSMSegments,OSMWays);
 
-       nsuper=Segments->number;
+       nsuper=OSMSegments->number;
       }
     else
       {
@@ -414,26 +424,26 @@ if(!option_process_only)
 
        /* Select the super-nodes */
 
-       ChooseSuperNodes(Nodes,SuperSegments,Ways);
+       ChooseSuperNodes(OSMNodes,SuperSegments,OSMWays);
 
        /* Select the super-segments */
 
-       SuperSegments2=CreateSuperSegments(Nodes,SuperSegments,Ways);
+       SuperSegments2=CreateSuperSegments(OSMNodes,SuperSegments,OSMWays);
 
        nsuper=SuperSegments->number;
 
-       FreeSegmentList(SuperSegments,0);
+       FreeSegmentList(SuperSegments);
 
        SuperSegments=SuperSegments2;
       }
 
     /* Sort the super-segments and remove duplicates */
 
-    DeduplicateSuperSegments(SuperSegments,Ways);
+    DeduplicateSuperSegments(SuperSegments,OSMWays);
 
     /* Index the segments */
 
-    IndexSegments(SuperSegments,Nodes,Ways);
+    IndexSegments(SuperSegments,OSMNodes,OSMWays);
 
     /* Check for end condition */
 
@@ -454,17 +464,13 @@ if(!option_process_only)
 
  /* Merge the super-segments */
 
- MergedSegments=MergeSuperSegments(Segments,SuperSegments);
+ MergedSegments=MergeSuperSegments(OSMSegments,SuperSegments);
 
- FreeSegmentList(Segments,0);
+ FreeSegmentList(OSMSegments);
 
- FreeSegmentList(SuperSegments,0);
+ FreeSegmentList(SuperSegments);
 
- Segments=MergedSegments;
-
- /* Re-index the merged segments */
-
- IndexSegments(Segments,Nodes,Ways);
+ OSMSegments=MergedSegments;
 
  /* Cross reference the nodes and segments */
 
@@ -473,17 +479,17 @@ if(!option_process_only)
 
  /* Sort the nodes and segments geographically */
 
- SortNodeListGeographically(Nodes);
+ SortNodeListGeographically(OSMNodes);
 
- SortSegmentListGeographically(Segments,Nodes);
+ SortSegmentListGeographically(OSMSegments,OSMNodes);
 
  /* Re-index the segments */
 
- IndexSegments(Segments,Nodes,Ways);
+ IndexSegments(OSMSegments,OSMNodes,OSMWays);
 
  /* Sort the turn relations geographically */
 
- SortTurnRelationListGeographically(Relations,Nodes,Segments);
+ SortTurnRelationListGeographically(OSMRelations,OSMNodes,OSMSegments);
 
  /* Output the results */
 
@@ -492,43 +498,55 @@ if(!option_process_only)
 
  /* Write out the nodes */
 
- SaveNodeList(Nodes,FileName(dirname,prefix,"nodes.mem"),Segments);
-
- FreeNodeList(Nodes,0);
+ SaveNodeList(OSMNodes,FileName(dirname,prefix,"nodes.mem"),OSMSegments);
 
  /* Write out the segments */
 
- SaveSegmentList(Segments,FileName(dirname,prefix,"segments.mem"));
-
- FreeSegmentList(Segments,0);
+ SaveSegmentList(OSMSegments,FileName(dirname,prefix,"segments.mem"));
 
  /* Write out the ways */
 
- SaveWayList(Ways,FileName(dirname,prefix,"ways.mem"));
-
- FreeWayList(Ways,0);
+ SaveWayList(OSMWays,FileName(dirname,prefix,"ways.mem"));
 
  /* Write out the relations */
 
- SaveRelationList(Relations,FileName(dirname,prefix,"relations.mem"));
+ SaveRelationList(OSMRelations,FileName(dirname,prefix,"relations.mem"));
 
- FreeRelationList(Relations,0);
-
- /* Close the error log file */
+ /* Close the error log file and process the data */
 
  if(errorlog)
+   {
     close_errorlog();
 
- /* Print the total time */
+    if(option_keep)
+      {
+       ErrorLogsX *OSMErrorLogs;
 
- if(option_logtime)
-   {
-    printf("\n");
-    fprintf_elapsed_time(stdout,&start_time);
-    printf("Complete\n");
-    fflush(stdout);
+       printf("\nCreate Error Log\n================\n\n");
+       fflush(stdout);
+
+       OSMErrorLogs=NewErrorLogList();
+
+       ProcessErrorLogs(OSMErrorLogs,OSMNodes,OSMWays,OSMRelations);
+
+       SortErrorLogsGeographically(OSMErrorLogs);
+
+       SaveErrorLogs(OSMErrorLogs,FileName(dirname,prefix,"errorlogs.mem"));
+
+       FreeErrorLogList(OSMErrorLogs);
+      }
    }
 
+ /* Free the memory (delete the temporary files) */
+
+ FreeNodeList(OSMNodes,0);
+ FreeWayList(OSMWays,0);
+ FreeRelationList(OSMRelations,0);
+
+ FreeSegmentList(OSMSegments);
+
+ printf_program_end();
+
  return(0);
 }
 
@@ -564,7 +582,19 @@ static void print_usage(int detail,const char *argerr,const char *err)
          "                      [--prune-isolated=<len>]\n"
          "                      [--prune-short=<len>]\n"
          "                      [--prune-straight=<len>]\n"
-         "                      [<filename.osm> ...]\n");
+         "                      [<filename.osm> ... | <filename.osc> ...\n"
+         "                       | <filename.pbf> ...\n"
+         "                       | <filename.o5m> ... | <filename.o5c> ..."
+#if defined(USE_BZIP2) && USE_BZIP2
+         "\n                       | <filename.(osm|osc|o5m|o5c).bz2> ..."
+#endif
+#if defined(USE_GZIP) && USE_GZIP
+         "\n                       | <filename.(osm|osc|o5m|o5c).gz> ..."
+#endif
+#if defined(USE_XZ) && USE_XZ
+         "\n                       | <filename.(osm|osc|o5m|o5c).xz> ..."
+#endif
+         "]\n");
 
  if(argerr)
     fprintf(stderr,
@@ -625,8 +655,19 @@ static void print_usage(int detail,const char *argerr,const char *err)
             "--prune-straight=<len>    Remove nodes in almost straight highways (defaults to\n"
             "                          removing nodes up to 3m offset from a straight line).\n"
             "\n"
-            "<filename.osm> ...        The name(s) of the file(s) to process (by default\n"
-            "                          data is read from standard input).\n"
+            "<filename.osm>, <filename.osc>, <filename.pbf>, <filename.o5m>, <filename.o5c>\n"
+            "                          The name(s) of the file(s) to read and parse.\n"
+            "                          Filenames ending '.pbf' read as PBF, filenames ending\n"
+            "                          '.o5m' or '.o5c' read as O5M/O5C, others as XML.\n"
+#if defined(USE_BZIP2) && USE_BZIP2
+            "                          Filenames ending '.bz2' will be bzip2 uncompressed.\n"
+#endif
+#if defined(USE_GZIP) && USE_GZIP
+            "                          Filenames ending '.gz' will be gzip uncompressed.\n"
+#endif
+#if defined(USE_XZ) && USE_XZ
+            "                          Filenames ending '.xz' will be xz uncompressed.\n"
+#endif
             "\n"
             "<transport> defaults to all but can be set to:\n"
             "%s"
