@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2011-2012 Andrew M. Bishop
+ This file Copyright 2011-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -82,11 +82,11 @@ void StartPruning(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
 
  /* Open the file read-only */
 
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
+ segmentsx->fd=ReOpenFileBuffered(segmentsx->filename_tmp);
 
  /* Read the on-disk image */
 
- while(!ReadFile(segmentsx->fd,&segmentx,sizeof(SegmentX)))
+ while(!ReadFileBuffered(segmentsx->fd,&segmentx,sizeof(SegmentX)))
    {
     index_t node1=segmentx.node1;
 
@@ -108,7 +108,7 @@ void StartPruning(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
 
  /* Close the file */
 
- segmentsx->fd=CloseFile(segmentsx->fd);
+ segmentsx->fd=CloseFileBuffered(segmentsx->fd);
 
  /* Print the final message */
 
@@ -150,12 +150,12 @@ void FinishPruning(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
 
 void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distance_t minimum)
 {
+ WaysX *newwaysx;
+ WayX tmpwayx;
  transport_t transport;
  BitMask *connected,*region;
  index_t *regionsegments,*othersegments;
- int nallocregionsegments,nallocothersegments;
- index_t nnewways=0;
- int fd;
+ index_t nallocregionsegments,nallocothersegments;
 
  if(nodesx->number==0 || segmentsx->number==0)
     return;
@@ -167,12 +167,19 @@ void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,dista
  segmentsx->data=MapFileWriteable(segmentsx->filename_tmp);
  waysx->data=MapFile(waysx->filename_tmp);
 #else
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
- segmentsx->fd=ReOpenFileWriteable(segmentsx->filename_tmp);
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+ nodesx->fd=SlimMapFile(nodesx->filename_tmp);
+ segmentsx->fd=SlimMapFileWriteable(segmentsx->filename_tmp);
+ waysx->fd=SlimMapFile(waysx->filename_tmp);
+
+ InvalidateNodeXCache(nodesx->cache);
+ InvalidateSegmentXCache(segmentsx->cache);
+ InvalidateWayXCache(waysx->cache);
 #endif
 
- fd=ReOpenFileWriteable(waysx->filename_tmp);
+ newwaysx=NewWayList(0,0);
+ CloseFileBuffered(newwaysx->fd);
+
+ newwaysx->fd=SlimMapFileWriteable(newwaysx->filename_tmp);
 
  connected=AllocBitMask(segmentsx->number);
  region   =AllocBitMask(segmentsx->number);
@@ -195,6 +202,9 @@ void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,dista
     const char *transport_str=TransportName(transport);
     transports_t transports=TRANSPORTS(transport);
 
+    if(!(waysx->allow&transports))
+       continue;
+
     /* Print the start message */
 
     printf_first("Pruning Isolated Regions (%s): Segments=0 Adjusted=0 Pruned=0",transport_str);
@@ -206,10 +216,10 @@ void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,dista
 
     for(i=0;i<segmentsx->number;i++)
       {
-       int nregionsegments=0,nothersegments=0;
+       index_t nregionsegments=0,nothersegments=0;
        distance_t total=0;
        SegmentX *segmentx;
-       WayX *wayx,tmpwayx;
+       WayX *wayx;
 
        if(IsBitSet(connected,i))
           goto endloop;
@@ -222,7 +232,7 @@ void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,dista
        if(segmentx->way<waysx->number)
           wayx=LookupWayX(waysx,segmentx->way,1);
        else
-          SeekReadFile(fd,(wayx=&tmpwayx),sizeof(WayX),segmentx->way*sizeof(WayX));
+          SlimFetch(newwaysx->fd,(wayx=&tmpwayx),sizeof(WayX),(segmentx->way-waysx->number)*sizeof(WayX));
 
        if(!(wayx->way.allow&transports))
           goto endloop;
@@ -265,7 +275,7 @@ void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,dista
                    if(segmentx->way<waysx->number)
                       wayx=LookupWayX(waysx,segmentx->way,1);
                    else
-                      SeekReadFile(fd,(wayx=&tmpwayx),sizeof(WayX),segmentx->way*sizeof(WayX));
+                      SlimFetch(newwaysx->fd,(wayx=&tmpwayx),sizeof(WayX),(segmentx->way-waysx->number)*sizeof(WayX));
 
                    if(wayx->way.allow&transports)
                      {
@@ -317,7 +327,7 @@ void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,dista
              if(segmentx->way<waysx->number)
                 wayx=LookupWayX(waysx,segmentx->way,1);
              else
-                SeekReadFile(fd,(wayx=&tmpwayx),sizeof(WayX),segmentx->way*sizeof(WayX));
+                SlimFetch(newwaysx->fd,(wayx=&tmpwayx),sizeof(WayX),(segmentx->way-waysx->number)*sizeof(WayX));
 
              if(wayx->way.allow==transports)
                {
@@ -333,11 +343,11 @@ void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,dista
 
                    tmpwayx.way.allow&=~transports;
 
-                   segmentx->way=waysx->number+nnewways;
+                   segmentx->way=waysx->number+newwaysx->number;
 
-                   SeekWriteFile(fd,&tmpwayx,sizeof(WayX),segmentx->way*sizeof(WayX));
+                   SlimReplace(newwaysx->fd,&tmpwayx,sizeof(WayX),(segmentx->way-waysx->number)*sizeof(WayX));
 
-                   nnewways++;
+                   newwaysx->number++;
 
                    PutBackSegmentX(segmentsx,segmentx);
                   }
@@ -345,7 +355,7 @@ void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,dista
                   {
                    tmpwayx.way.allow&=~transports;
 
-                   SeekWriteFile(fd,&tmpwayx,sizeof(WayX),segmentx->way*sizeof(WayX));
+                   SlimReplace(newwaysx->fd,&tmpwayx,sizeof(WayX),(segmentx->way-waysx->number)*sizeof(WayX));
                   }
 
                 nadjusted++;
@@ -391,14 +401,26 @@ void PruneIsolatedRegions(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,dista
  segmentsx->data=UnmapFile(segmentsx->data);
  waysx->data=UnmapFile(waysx->data);
 #else
- nodesx->fd=CloseFile(nodesx->fd);
- segmentsx->fd=CloseFile(segmentsx->fd);
- waysx->fd=CloseFile(waysx->fd);
+ nodesx->fd=SlimUnmapFile(nodesx->fd);
+ segmentsx->fd=SlimUnmapFile(segmentsx->fd);
+ waysx->fd=SlimUnmapFile(waysx->fd);
 #endif
 
- CloseFile(fd);
+ SlimUnmapFile(newwaysx->fd);
 
- waysx->number+=nnewways;
+ waysx->number+=newwaysx->number;
+
+ waysx->fd=OpenFileBufferedAppend(waysx->filename_tmp);
+
+ newwaysx->fd=ReOpenFileBuffered(newwaysx->filename_tmp);
+
+ while(!ReadFileBuffered(newwaysx->fd,&tmpwayx,sizeof(WayX)))
+    WriteFileBuffered(waysx->fd,&tmpwayx,sizeof(WayX));
+
+ CloseFileBuffered(waysx->fd);
+ CloseFileBuffered(newwaysx->fd);
+
+ FreeWayList(newwaysx,0);
 }
 
 
@@ -433,9 +455,13 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
  segmentsx->data=MapFileWriteable(segmentsx->filename_tmp);
  waysx->data=MapFile(waysx->filename_tmp);
 #else
- nodesx->fd=ReOpenFileWriteable(nodesx->filename_tmp);
- segmentsx->fd=ReOpenFileWriteable(segmentsx->filename_tmp);
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+ nodesx->fd=SlimMapFileWriteable(nodesx->filename_tmp);
+ segmentsx->fd=SlimMapFileWriteable(segmentsx->filename_tmp);
+ waysx->fd=SlimMapFile(waysx->filename_tmp);
+
+ InvalidateNodeXCache(nodesx->cache);
+ InvalidateSegmentXCache(segmentsx->cache);
+ InvalidateWayXCache(waysx->cache);
 #endif
 
  /* Loop through the segments and find the short ones for possible modification */
@@ -453,8 +479,8 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
                        :     S2
 
                        :
-      Final state:   ..N3        N2 --
-                       :            S2
+      Final state:   ..N3
+                       :
 
       = OR =
 
@@ -463,14 +489,15 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
                        :     S1        S2        S3    :
 
                        :                               :
-      Final state:   ..N1 ------------ N3 ------------ N4..   N2 --
-                       :       S1               S3     :         S2
+      Final state:   ..N1 ------------ N3 ------------ N4..
+                       :       S1               S3     :
 
       Not if N1 is the same as N4.
-      Not if S2 has different one-way properties from S1 and S3.
+      Must not delete N2 (or N3) if S2 (or S3) has different one-way properties from S1.
+      Must not delete N2 (or N3) if S2 (or S3) has different highway properties from S1.
       Must combine N2, S2 and N3 disallowed transports into new N3.
-      Must not delete N2 or N3 if they are mini-roundabouts.
-      Must not delete N2 or N3 if they are part of turn relations.
+      Must not delete N2 (or N3) if it is a mini-roundabout.
+      Must not delete N2 (or N3) if it is involved in a turn restriction.
 
       = OR =
 
@@ -479,14 +506,14 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
                        :     S1        S2  :
 
                        :               :
-      Final state:   ..N1 ------------ N3..   N2 --
-                       :       S1      :         S2
+      Final state:   ..N1 ------------ N3..
+                       :       S1      :
 
       Not if N1 is the same as N3.
       Not if S1 has different one-way properties from S2.
       Not if S1 has different highway properties from S2.
       Not if N2 disallows transports allowed on S1 and S2.
-      Not if N2 is a mini-roundabouts.
+      Not if N2 is a mini-roundabout.
       Not if N2 is involved in a turn restriction.
      */
 
@@ -502,7 +529,7 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
        node2=segmentx2->node1;
        node3=segmentx2->node2;
 
-       /* Count the segments connected to N2 and N3 */
+       /* Count the segments connected to N2 */
 
        segmentx=FirstSegmentX(segmentsx,node2,4);
 
@@ -526,6 +553,8 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
           segmentx=NextSegmentX(segmentsx,segmentx,node2);
          }
 
+       /* Count the segments connected to N3 */
+
        segmentx=FirstSegmentX(segmentsx,node3,4);
 
        while(segmentx)
@@ -605,6 +634,21 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
           if(!join12 && !join23)
              goto endloop;
 
+          /* Check if allowed due to highway properties */
+
+          wayx1=LookupWayX(waysx,segmentx1->way,1);
+          wayx2=LookupWayX(waysx,segmentx2->way,2);
+          wayx3=LookupWayX(waysx,segmentx3->way,3);
+
+          if(WaysCompare(&wayx1->way,&wayx2->way))
+             join12=0;
+
+          if(WaysCompare(&wayx3->way,&wayx2->way))
+             join23=0;
+
+          if(!join12 && !join23)
+             goto endloop;
+
           /* Check if allowed due to mini-roundabout and turn restriction */
 
           nodex2=LookupNodeX(nodesx,node2,2);
@@ -616,6 +660,9 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
           if(nodex3->flags&NODE_MINIRNDBT)
              join23=0;
 
+          if(!join12 && !join23)
+             goto endloop;
+
           if(nodex2->flags&NODE_TURNRSTRCT2 || nodex2->flags&NODE_TURNRSTRCT)
              join12=0;
 
@@ -638,10 +685,6 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
              newnodex=nodex2;
             }
 
-          wayx1=LookupWayX(waysx,segmentx1->way,1);
-          wayx2=LookupWayX(waysx,segmentx2->way,2);
-          wayx3=LookupWayX(waysx,segmentx3->way,3);
-
           newnodex->allow=nodex2->allow&nodex3->allow; /* combine the restrictions of the two nodes */
           newnodex->allow&=~((~wayx2->way.allow)&wayx3->way.allow); /* disallow anything blocked by segment2 */
           newnodex->allow&=~((~wayx2->way.allow)&wayx1->way.allow); /* disallow anything blocked by segment2 */
@@ -783,9 +826,9 @@ void PruneShortSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,distanc
  segmentsx->data=UnmapFile(segmentsx->data);
  waysx->data=UnmapFile(waysx->data);
 #else
- nodesx->fd=CloseFile(nodesx->fd);
- segmentsx->fd=CloseFile(segmentsx->fd);
- waysx->fd=CloseFile(waysx->fd);
+ nodesx->fd=SlimUnmapFile(nodesx->fd);
+ segmentsx->fd=SlimUnmapFile(segmentsx->fd);
+ waysx->fd=SlimUnmapFile(waysx->fd);
 #endif
 
  /* Print the final message */
@@ -810,8 +853,8 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
 {
  index_t i;
  index_t npruned=0;
+ index_t nalloc;
  BitMask *checked;
- int nalloc;
  index_t *nodes,*segments;
  double *lats,*lons;
  double maximumf;
@@ -819,8 +862,6 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
  if(nodesx->number==0 || segmentsx->number==0 || waysx->number==0)
     return;
 
- maximumf=distance_to_km(maximum);
-
  /* Print the start message */
 
  printf_first("Pruning Straight Highway Nodes: Nodes=0 Pruned=0");
@@ -832,9 +873,13 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
  segmentsx->data=MapFileWriteable(segmentsx->filename_tmp);
  waysx->data=MapFile(waysx->filename_tmp);
 #else
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
- segmentsx->fd=ReOpenFileWriteable(segmentsx->filename_tmp);
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+ nodesx->fd=SlimMapFile(nodesx->filename_tmp);
+ segmentsx->fd=SlimMapFileWriteable(segmentsx->filename_tmp);
+ waysx->fd=SlimMapFile(waysx->filename_tmp);
+
+ InvalidateNodeXCache(nodesx->cache);
+ InvalidateSegmentXCache(segmentsx->cache);
+ InvalidateWayXCache(waysx->cache);
 #endif
 
  checked=AllocBitMask(nodesx->number);
@@ -853,7 +898,9 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
  logassert(lats,"Failed to allocate memory (try using slim mode?)");    /* Check malloc() worked */
  logassert(lons,"Failed to allocate memory (try using slim mode?)");    /* Check malloc() worked */
 
- /* Loop through the nodes and find stretchs of simple highway for possible modification */
+ /* Loop through the nodes and find stretches of simple highway for possible modification */
+
+ maximumf=distance_to_km(maximum);
 
  for(i=0;i<nodesx->number;i++)
    {
@@ -899,7 +946,7 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
 
           /* Count the segments connected to the node */
 
-          segmentx=FirstSegmentX(segmentsx,nodes[current],1);
+          segmentx=FirstSegmentX(segmentsx,nodes[current],3);
 
           while(segmentx)
             {
@@ -973,6 +1020,8 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
 
        if(segcount==2)
          {
+          /* Make space in the lists */
+
           if(upper==(nalloc-1))
             {
              nodes   =(index_t*)realloc(nodes   ,(nalloc+=1024)*sizeof(index_t));
@@ -1006,6 +1055,7 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
 
              nodes[upper]=node2;
              segments[upper-1]=segment2;
+             segments[upper]=NO_SEGMENT;
 
              current--;
             }
@@ -1041,16 +1091,29 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
                 segments[upper-1]=segment2;
                }
 
+             segments[upper]=NO_SEGMENT;
+
              current++;
             }
 
           if(nodes[upper]==nodes[lower])
             {
+             if(!lowerbounded && !upperbounded)
+               {
+                nodex=LookupNodeX(nodesx,nodes[lower],1);
+
+                lats[lower]=latlong_to_radians(nodex->latitude);
+                lons[lower]=latlong_to_radians(nodex->longitude);
+               }
+
+             lats[upper]=lats[lower];
+             lons[upper]=lons[lower];
+
              lowerbounded=1;
              upperbounded=1;
             }
          }
-       else
+       else /* if(segment!=2) */
          {
           if(current==upper)
              upperbounded=1;
@@ -1071,13 +1134,13 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
 
     /* Check for straight highway */
 
-    while((upper-lower)>=2)
+    for(;lower<(upper-1);lower++)
       {
-       index_t bestc=lower;
-
-       for(current=lower+2;current<=upper;current++)
+       for(current=upper;current>(lower+1);current--)
          {
-          double dist1,dist2,dist3,dist3a,dist3b,distp;
+          SegmentX *segmentx;
+          distance_t dist=0;
+          double dist1,dist2,dist3,distp;
           index_t c;
 
           dist3=distance(lats[lower],lons[lower],lats[current],lons[current]);
@@ -1089,75 +1152,63 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
 
              /* Use law of cosines (assume flat Earth) */
 
-             dist3a=(dist1*dist1-dist2*dist2+dist3*dist3)/(2.0*dist3);
-             dist3b=dist3-dist3a;
-
-             if((dist1+dist2)<dist3)
+             if(dist3==0)
+                distp=dist1; /* == dist2 */
+             else if((dist1+dist2)<dist3)
                 distp=0;
-             else if(dist3a>=0 && dist3b>=0)
-                distp=sqrt(dist1*dist1-dist3a*dist3a);
-             else if(dist3a>0)
-                distp=dist2;
-             else /* if(dist3b>0) */
-                distp=dist1;
+             else
+               {
+                double dist3a=(dist1*dist1-dist2*dist2+dist3*dist3)/(2.0*dist3);
+                double dist3b=dist3-dist3a;
+
+                if(dist3a>=0 && dist3b>=0)
+                   distp=sqrt(dist1*dist1-dist3a*dist3a);
+                else if(dist3a>0)
+                   distp=dist2;
+                else /* if(dist3b>0) */
+                   distp=dist1;
+               }
 
              if(distp>maximumf) /* gone too far */
                 break;
             }
 
-          if(c>bestc)
-             bestc=c;
-
-          if(bestc>c)
-             c=bestc;
-
-          if(c==current && current!=upper) /* Can replace at least this far (not finished yet) */
+          if(c<current) /* not finished */
              continue;
 
-          if((c-lower)<2)       /* first three points are not straight */
+          /* Delete some segments and shift along */
+
+          for(c=lower+1;c<current;c++)
             {
-             lower=c;
-             break;
+             segmentx=LookupSegmentX(segmentsx,segments[c],1);
+
+             dist+=DISTANCE(segmentx->distance);
+
+             prune_segment(segmentsx,segmentx);
+
+             npruned++;
             }
-          else                  /* delete some segments and shift along */
+
+          segmentx=LookupSegmentX(segmentsx,segments[lower],1);
+
+          if(nodes[lower]==nodes[current]) /* loop; all within maximum distance */
             {
-             SegmentX *segmentx;
-             distance_t distance=0;
+             prune_segment(segmentsx,segmentx);
 
-             current=c;
-
-             for(c=lower+1;c<current;c++)
-               {
-                segmentx=LookupSegmentX(segmentsx,segments[c],1);
-
-                distance+=DISTANCE(segmentx->distance);
-
-                prune_segment(segmentsx,segmentx);
-
-                npruned++;
-               }
-
-             segmentx=LookupSegmentX(segmentsx,segments[lower],1);
-
-             if(nodes[lower]==nodes[current]) /* loop; all within maximum distance */
-               {
-                prune_segment(segmentsx,segmentx);
-
-                npruned++;
-               }
-             else
-               {
-                segmentx->distance+=distance;
-
-                if(segmentx->node1==nodes[lower])
-                   modify_segment(segmentsx,segmentx,nodes[lower],nodes[current]);
-                else /* if(segmentx->node2==nodes[lower]) */
-                   modify_segment(segmentsx,segmentx,nodes[current],nodes[lower]);
-               }
-
-             lower=current;
-             break;
+             npruned++;
             }
+          else
+            {
+             segmentx->distance+=dist;
+
+             if(segmentx->node1==nodes[lower])
+                modify_segment(segmentsx,segmentx,nodes[lower],nodes[current]);
+             else /* if(segmentx->node2==nodes[lower]) */
+                modify_segment(segmentsx,segmentx,nodes[current],nodes[lower]);
+            }
+
+          lower=current-1;
+          break;
          }
       }
 
@@ -1182,9 +1233,9 @@ void PruneStraightHighwayNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,
  segmentsx->data=UnmapFile(segmentsx->data);
  waysx->data=UnmapFile(waysx->data);
 #else
- nodesx->fd=CloseFile(nodesx->fd);
- segmentsx->fd=CloseFile(segmentsx->fd);
- waysx->fd=CloseFile(waysx->fd);
+ nodesx->fd=SlimUnmapFile(nodesx->fd);
+ segmentsx->fd=SlimUnmapFile(segmentsx->fd);
+ waysx->fd=SlimUnmapFile(waysx->fd);
 #endif
 
  /* Print the final message */
