@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -52,11 +52,7 @@ static WaysX *sortwaysx;
 
 /* Local functions */
 
-static int sort_by_way_id(SegmentX *a,SegmentX *b);
-static int apply_changes(SegmentX *segmentx,index_t index);
-
 static int sort_by_id(SegmentX *a,SegmentX *b);
-static int deduplicate(SegmentX *segmentx,index_t index);
 
 static int delete_pruned(SegmentX *segmentx,index_t index);
 
@@ -71,13 +67,9 @@ static distance_t DistanceX(NodeX *nodex1,NodeX *nodex2);
   Allocate a new segment list (create a new file or open an existing one).
 
   SegmentsX *NewSegmentList Returns the segment list.
-
-  int append Set to 1 if the file is to be opened for appending.
-
-  int readonly Set to 1 if the file is to be opened for reading.
   ++++++++++++++++++++++++++++++++++++++*/
 
-SegmentsX *NewSegmentList(int append,int readonly)
+SegmentsX *NewSegmentList(void)
 {
  SegmentsX *segmentsx;
 
@@ -85,30 +77,15 @@ SegmentsX *NewSegmentList(int append,int readonly)
 
  logassert(segmentsx,"Failed to allocate memory (try using slim mode?)"); /* Check calloc() worked */
 
- segmentsx->filename    =(char*)malloc(strlen(option_tmpdirname)+32);
  segmentsx->filename_tmp=(char*)malloc(strlen(option_tmpdirname)+32);
 
- sprintf(segmentsx->filename    ,"%s/segmentsx.parsed.mem",option_tmpdirname);
- sprintf(segmentsx->filename_tmp,"%s/segmentsx.%p.tmp"    ,option_tmpdirname,(void*)segmentsx);
+ sprintf(segmentsx->filename_tmp,"%s/segmentsx.%p.tmp",option_tmpdirname,(void*)segmentsx);
 
- if(append || readonly)
-    if(ExistsFile(segmentsx->filename))
-      {
-       off_t size;
+ segmentsx->fd=OpenFileBufferedNew(segmentsx->filename_tmp);
 
-       size=SizeFile(segmentsx->filename);
-
-       segmentsx->number=size/sizeof(SegmentX);
-
-       RenameFile(segmentsx->filename,segmentsx->filename_tmp);
-      }
-
- if(append)
-    segmentsx->fd=OpenFileAppend(segmentsx->filename_tmp);
- else if(!readonly)
-    segmentsx->fd=OpenFileNew(segmentsx->filename_tmp);
- else
-    segmentsx->fd=-1;
+#if SLIM
+ segmentsx->cache=NewSegmentXCache();
+#endif
 
  return(segmentsx);
 }
@@ -118,18 +95,12 @@ SegmentsX *NewSegmentList(int append,int readonly)
   Free a segment list.
 
   SegmentsX *segmentsx The set of segments to be freed.
-
-  int keep If set then the results file is to be kept.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void FreeSegmentList(SegmentsX *segmentsx,int keep)
+void FreeSegmentList(SegmentsX *segmentsx)
 {
- if(keep)
-    RenameFile(segmentsx->filename_tmp,segmentsx->filename);
- else
-    DeleteFile(segmentsx->filename_tmp);
+ DeleteFile(segmentsx->filename_tmp);
 
- free(segmentsx->filename);
  free(segmentsx->filename_tmp);
 
  if(segmentsx->firstnode)
@@ -138,8 +109,9 @@ void FreeSegmentList(SegmentsX *segmentsx,int keep)
  if(segmentsx->next1)
     free(segmentsx->next1);
 
- if(segmentsx->usednode)
-    free(segmentsx->usednode);
+#if SLIM
+ DeleteSegmentXCache(segmentsx->cache);
+#endif
 
  free(segmentsx);
 }
@@ -150,22 +122,22 @@ void FreeSegmentList(SegmentsX *segmentsx,int keep)
 
   SegmentsX *segmentsx The set of segments to modify.
 
-  way_t way The way that the segment belongs to.
+  index_t way The index of the way that the segment belongs to.
 
-  node_t node1 The first node in the segment.
+  index_t node1 The index of the first node in the segment.
 
-  node_t node2 The second node in the segment.
+  index_t node2 The index of the second node in the segment.
 
   distance_t distance The distance between the nodes (or just the flags).
   ++++++++++++++++++++++++++++++++++++++*/
 
-void AppendSegmentList(SegmentsX *segmentsx,way_t way,node_t node1,node_t node2,distance_t distance)
+void AppendSegmentList(SegmentsX *segmentsx,index_t way,index_t node1,index_t node2,distance_t distance)
 {
  SegmentX segmentx;
 
  if(node1>node2)
    {
-    node_t temp;
+    index_t temp;
 
     temp=node1;
     node1=node2;
@@ -181,7 +153,7 @@ void AppendSegmentList(SegmentsX *segmentsx,way_t way,node_t node1,node_t node2,
  segmentx.way=way;
  segmentx.distance=distance;
 
- WriteFile(segmentsx->fd,&segmentx,sizeof(SegmentX));
+ WriteFileBuffered(segmentsx->fd,&segmentx,sizeof(SegmentX));
 
  segmentsx->number++;
 
@@ -198,7 +170,7 @@ void AppendSegmentList(SegmentsX *segmentsx,way_t way,node_t node1,node_t node2,
 void FinishSegmentList(SegmentsX *segmentsx)
 {
  if(segmentsx->fd!=-1)
-    segmentsx->fd=CloseFile(segmentsx->fd);
+    segmentsx->fd=CloseFileBuffered(segmentsx->fd);
 }
 
 
@@ -293,113 +265,14 @@ SegmentX *NextSegmentX(SegmentsX *segmentsx,SegmentX *segmentx,index_t nodeindex
  
  
 /*++++++++++++++++++++++++++++++++++++++
-  Apply the changes to the segments (no unique id to use).
+  Sort the segment list.
 
-  SegmentsX *segmentsx The set of segments to sort and modify.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-void ApplySegmentChanges(SegmentsX *segmentsx)
-{
- int fd;
- index_t xnumber;
-
- /* Print the start message */
-
- printf_first("Applying Segment Changes");
-
- /* Re-open the file read-only and a new file writeable */
-
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
-
- DeleteFile(segmentsx->filename_tmp);
-
- fd=OpenFileNew(segmentsx->filename_tmp);
-
- /* Sort by node indexes */
-
- xnumber=segmentsx->number;
-
- segmentsx->number=filesort_fixed(segmentsx->fd,fd,sizeof(SegmentX),NULL,
-                                                                    (int (*)(const void*,const void*))sort_by_way_id,
-                                                                    (int (*)(void*,index_t))apply_changes);
-
- /* Close the files */
-
- segmentsx->fd=CloseFile(segmentsx->fd);
- CloseFile(fd);
-
- /* Print the final message */
-
- printf_last("Applying Segment Changes: Segments=%"Pindex_t" Changed=%"Pindex_t,xnumber,xnumber-segmentsx->number);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  Sort the segments into way id order.
-
-  int sort_by_way_id Returns the comparison of the way fields.
-
-  SegmentX *a The first segment.
-
-  SegmentX *b The second segment.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int sort_by_way_id(SegmentX *a,SegmentX *b)
-{
- way_t a_id=a->way;
- way_t b_id=b->way;
-
- if(a_id<b_id)
-    return(-1);
- else if(a_id>b_id)
-    return(1);
- else /* if(a_id==b_id) */
-    return(-FILESORT_PRESERVE_ORDER(a,b)); /* latest version first */
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  Apply the changes to the segments.
-
-  int apply_changes Return 1 if the value is to be kept, otherwise 0.
-
-  SegmentX *segmentx The extended segment.
-
-  index_t index The number of sorted segments that have already been written to the output file.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int apply_changes(SegmentX *segmentx,index_t index)
-{
- static way_t prevway=NO_WAY_ID;
- static int deleted=0;
-
- if(prevway!=segmentx->way)
-   {
-    prevway=segmentx->way;
-    deleted=0;
-   }
-
- if(!deleted)
-    if(segmentx->node1==NO_NODE_ID)
-       deleted=1;
-
- if(deleted)
-    return(0);
- else
-    return(1);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  Sort the segment list and deduplicate it.
-
-  SegmentsX *segmentsx The set of segments to sort and modify.
+  SegmentsX *segmentsx The set of segments to sort.
   ++++++++++++++++++++++++++++++++++++++*/
 
 void SortSegmentList(SegmentsX *segmentsx)
 {
  int fd;
- index_t xnumber;
 
  /* Print the start message */
 
@@ -407,28 +280,26 @@ void SortSegmentList(SegmentsX *segmentsx)
 
  /* Re-open the file read-only and a new file writeable */
 
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
+ segmentsx->fd=ReOpenFileBuffered(segmentsx->filename_tmp);
 
  DeleteFile(segmentsx->filename_tmp);
 
- fd=OpenFileNew(segmentsx->filename_tmp);
+ fd=OpenFileBufferedNew(segmentsx->filename_tmp);
 
  /* Sort by node indexes */
 
- xnumber=segmentsx->number;
-
  segmentsx->number=filesort_fixed(segmentsx->fd,fd,sizeof(SegmentX),NULL,
                                                                     (int (*)(const void*,const void*))sort_by_id,
-                                                                    (int (*)(void*,index_t))deduplicate);
+                                                                    NULL);
 
  /* Close the files */
 
- segmentsx->fd=CloseFile(segmentsx->fd);
- CloseFile(fd);
+ segmentsx->fd=CloseFileBuffered(segmentsx->fd);
+ CloseFileBuffered(fd);
 
  /* Print the final message */
 
- printf_last("Sorted Segments: Segments=%"Pindex_t" Duplicates=%"Pindex_t,xnumber,xnumber-segmentsx->number);
+ printf_last("Sorted Segments: Segments=%"Pindex_t,segmentsx->number);
 }
 
 
@@ -444,8 +315,8 @@ void SortSegmentList(SegmentsX *segmentsx)
 
 static int sort_by_id(SegmentX *a,SegmentX *b)
 {
- node_t a_id1=a->node1;
- node_t b_id1=b->node1;
+ index_t a_id1=a->node1;
+ index_t b_id1=b->node1;
 
  if(a_id1<b_id1)
     return(-1);
@@ -453,8 +324,8 @@ static int sort_by_id(SegmentX *a,SegmentX *b)
     return(1);
  else /* if(a_id1==b_id1) */
    {
-    node_t a_id2=a->node2;
-    node_t b_id2=b->node2;
+    index_t a_id2=a->node2;
+    index_t b_id2=b->node2;
 
     if(a_id2<b_id2)
        return(-1);
@@ -487,194 +358,39 @@ static int sort_by_id(SegmentX *a,SegmentX *b)
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  Discard duplicate segments.
-
-  int deduplicate Return 1 if the value is to be kept, otherwise 0.
-
-  SegmentX *segmentx The extended segment.
-
-  index_t index The number of sorted segments that have already been written to the output file.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int deduplicate(SegmentX *segmentx,index_t index)
-{
- static node_t prevnode1=NO_NODE_ID,prevnode2=NO_NODE_ID;
- static way_t prevway=NO_WAY_ID;
- static distance_t prevdist=0;
-
- if(prevnode1!=segmentx->node1 || prevnode2!=segmentx->node2 || prevway!=segmentx->way || prevdist!=segmentx->distance)
-   {
-    prevnode1=segmentx->node1;
-    prevnode2=segmentx->node2;
-    prevway=segmentx->way;
-    prevdist=segmentx->distance;
-
-    return(1);
-   }
- else
-    return(0);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  Remove bad segments (duplicated, zero length or with missing nodes).
+  Process segments (non-trivial duplicates).
 
   SegmentsX *segmentsx The set of segments to modify.
 
   NodesX *nodesx The set of nodes to use.
 
   WaysX *waysx The set of ways to use.
-
-  int keep If set to 1 then keep the old data file otherwise delete it.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void RemoveBadSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx,int keep)
+void ProcessSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx)
 {
- index_t noway=0,loop=0,nonode=0,duplicate=0,good=0,total=0;
- node_t prevnode1=NO_NODE_ID,prevnode2=NO_NODE_ID;
- way_t prevway=NO_WAY_ID;
+ index_t duplicate=0,good=0,total=0;
+ index_t prevnode1=NO_NODE,prevnode2=NO_NODE;
+ index_t prevway=NO_WAY;
  distance_t prevdist=0;
  SegmentX segmentx;
  int fd;
 
  /* Print the start message */
 
- printf_first("Checking Segments: Segments=0 Loop=0 No-Way=0 No-Node=0 Duplicate=0");
-
- /* Allocate the node usage bitmask */
-
- segmentsx->usednode=AllocBitMask(nodesx->number);
-
- logassert(segmentsx->usednode,"Failed to allocate memory (try using slim mode?)"); /* Check AllocBitMask() worked */
-
- /* Re-open the file read-only and a new file writeable */
-
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
-
- if(keep)
-    RenameFile(segmentsx->filename_tmp,segmentsx->filename);
- else
-    DeleteFile(segmentsx->filename_tmp);
-
- fd=OpenFileNew(segmentsx->filename_tmp);
-
- /* Modify the on-disk image */
-
- while(!ReadFile(segmentsx->fd,&segmentx,sizeof(SegmentX)))
-   {
-    index_t index1=IndexNodeX(nodesx,segmentx.node1);
-    index_t index2=IndexNodeX(nodesx,segmentx.node2);
-    index_t indexw=IndexWayX(waysx,segmentx.way);
-
-    if(indexw==NO_WAY)
-      {
-       logerror("Segment belongs to way %"Pway_t" but it doesn't exist.\n",segmentx.way);
-
-       noway++;
-      }
-    else if(segmentx.node1==segmentx.node2)
-      {
-       logerror("Segment connects node %"Pnode_t" to itself.\n",segmentx.node1);
-
-       loop++;
-      }
-    else if(index1==NO_NODE || index2==NO_NODE)
-      {
-       if(index1==NO_NODE && index2==NO_NODE)
-          logerror("Segment connects nodes %"Pnode_t" and %"Pnode_t" but neither exist.\n",segmentx.node1,segmentx.node2);
-
-       if(index1==NO_NODE && index2!=NO_NODE)
-          logerror("Segment connects nodes %"Pnode_t" and %"Pnode_t" but the first one does not exist.\n",segmentx.node1,segmentx.node2);
-
-       if(index1!=NO_NODE && index2==NO_NODE)
-          logerror("Segment connects nodes %"Pnode_t" and %"Pnode_t" but the second one does not exist.\n",segmentx.node1,segmentx.node2);
-
-       nonode++;
-      }
-    else if(prevnode1==segmentx.node1 && prevnode2==segmentx.node2)
-      {
-       if(prevway==segmentx.way)
-          ; /* already logged an error - only possible to get here for oneway opposite direction segments */
-       else
-         {
-          if(!(prevdist&SEGMENT_AREA) && !(segmentx.distance&SEGMENT_AREA))
-             logerror("Segment connecting nodes %"Pnode_t" and %"Pnode_t" is duplicated.\n",segmentx.node1,segmentx.node2);
-
-          if(!(prevdist&SEGMENT_AREA) && (segmentx.distance&SEGMENT_AREA))
-             logerror("Segment connecting nodes %"Pnode_t" and %"Pnode_t" is duplicated (discarded the area).\n",segmentx.node1,segmentx.node2);
-
-          if((prevdist&SEGMENT_AREA) && !(segmentx.distance&SEGMENT_AREA))
-             logerror("Segment connecting nodes %"Pnode_t" and %"Pnode_t" is duplicated (discarded the non-area).\n",segmentx.node1,segmentx.node2);
-
-          if((prevdist&SEGMENT_AREA) && (segmentx.distance&SEGMENT_AREA))
-             logerror("Segment connecting nodes %"Pnode_t" and %"Pnode_t" is duplicated (both are areas).\n",segmentx.node1,segmentx.node2);
-         }
-
-       duplicate++;
-      }
-    else
-      {
-       WriteFile(fd,&segmentx,sizeof(SegmentX));
-
-       SetBit(segmentsx->usednode,index1);
-       SetBit(segmentsx->usednode,index2);
-
-       prevnode1=segmentx.node1;
-       prevnode2=segmentx.node2;
-       prevway=segmentx.way;
-       prevdist=DISTANCE(segmentx.distance);
-
-       good++;
-      }
-
-    total++;
-
-    if(!(total%10000))
-       printf_middle("Checking Segments: Segments=%"Pindex_t" Loop=%"Pindex_t" No-Way=%"Pindex_t" No-Node=%"Pindex_t" Duplicate=%"Pindex_t,total,loop,noway,nonode,duplicate);
-   }
-
- segmentsx->number=good;
-
- /* Close the files */
-
- segmentsx->fd=CloseFile(segmentsx->fd);
- CloseFile(fd);
-
- /* Print the final message */
-
- printf_last("Checked Segments: Segments=%"Pindex_t" Loop=%"Pindex_t" No-Way=%"Pindex_t" No-Node=%"Pindex_t" Duplicate=%"Pindex_t,total,loop,noway,nonode,duplicate);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  Measure the segments and replace node/way ids with indexes.
-
-  SegmentsX *segmentsx The set of segments to process.
-
-  NodesX *nodesx The set of nodes to use.
-
-  WaysX *waysx The set of ways to use.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-void MeasureSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx)
-{
- index_t index=0;
- int fd;
- SegmentX segmentx;
-
- if(segmentsx->number==0)
-    return;
-
- /* Print the start message */
-
- printf_first("Measuring Segments: Segments=0");
+ printf_first("Processing Segments: Segments=0 Duplicates=0");
 
  /* Map into memory /  open the file */
 
 #if !SLIM
  nodesx->data=MapFile(nodesx->filename_tmp);
+ waysx->data=MapFile(waysx->filename_tmp);
 #else
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
+ nodesx->fd=SlimMapFile(nodesx->filename_tmp);
+ waysx->fd=SlimMapFile(waysx->filename_tmp);
+
+ InvalidateNodeXCache(nodesx->cache);
+ InvalidateWayXCache(waysx->cache);
 #endif
 
  /* Allocate the way usage bitmask */
@@ -685,70 +401,93 @@ void MeasureSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx)
 
  /* Re-open the file read-only and a new file writeable */
 
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
+ segmentsx->fd=ReOpenFileBuffered(segmentsx->filename_tmp);
 
  DeleteFile(segmentsx->filename_tmp);
 
- fd=OpenFileNew(segmentsx->filename_tmp);
+ fd=OpenFileBufferedNew(segmentsx->filename_tmp);
 
  /* Modify the on-disk image */
 
- while(!ReadFile(segmentsx->fd,&segmentx,sizeof(SegmentX)))
+ while(!ReadFileBuffered(segmentsx->fd,&segmentx,sizeof(SegmentX)))
    {
-    index_t node1=IndexNodeX(nodesx,segmentx.node1);
-    index_t node2=IndexNodeX(nodesx,segmentx.node2);
-    index_t way  =IndexWayX (waysx ,segmentx.way);
+    NodeX *nodex1=LookupNodeX(nodesx,segmentx.node1,1);
+    NodeX *nodex2=LookupNodeX(nodesx,segmentx.node2,2);
 
-    NodeX *nodex1=LookupNodeX(nodesx,node1,1);
-    NodeX *nodex2=LookupNodeX(nodesx,node2,2);
+    if(prevnode1==segmentx.node1 && prevnode2==segmentx.node2)
+      {
+       if(prevway==segmentx.way)
+         {
+          WayX *wayx=LookupWayX(waysx,segmentx.way,1);
 
-    /* Replace the node and way ids with their indexes */
+          logerror("Segment connecting nodes %"Pnode_t" and %"Pnode_t" in way %"Pway_t" is duplicated.\n",logerror_node(nodex1->id),logerror_node(nodex2->id),logerror_way(wayx->id));
+         }
+       else
+         {
+          if(!(prevdist&SEGMENT_AREA) && !(segmentx.distance&SEGMENT_AREA))
+             logerror("Segment connecting nodes %"Pnode_t" and %"Pnode_t" is duplicated.\n",logerror_node(nodex1->id),logerror_node(nodex2->id));
 
-    segmentx.node1=node1;
-    segmentx.node2=node2;
-    segmentx.way  =way;
+          if(!(prevdist&SEGMENT_AREA) && (segmentx.distance&SEGMENT_AREA))
+             logerror("Segment connecting nodes %"Pnode_t" and %"Pnode_t" is duplicated (discarded the area).\n",logerror_node(nodex1->id),logerror_node(nodex2->id));
 
-    SetBit(segmentsx->usedway,segmentx.way);
+          if((prevdist&SEGMENT_AREA) && !(segmentx.distance&SEGMENT_AREA))
+             logerror("Segment connecting nodes %"Pnode_t" and %"Pnode_t" is duplicated (discarded the non-area).\n",logerror_node(nodex1->id),logerror_node(nodex2->id));
 
-    /* Set the distance but keep the other flags except for area */
+          if((prevdist&SEGMENT_AREA) && (segmentx.distance&SEGMENT_AREA))
+             logerror("Segment connecting nodes %"Pnode_t" and %"Pnode_t" is duplicated (both are areas).\n",logerror_node(nodex1->id),logerror_node(nodex2->id));
+         }
 
-    segmentx.distance=DISTANCE(DistanceX(nodex1,nodex2))|DISTFLAG(segmentx.distance);
-    segmentx.distance&=~SEGMENT_AREA;
+       duplicate++;
+      }
+    else
+      {
+       prevnode1=segmentx.node1;
+       prevnode2=segmentx.node2;
+       prevway=segmentx.way;
+       prevdist=DISTANCE(segmentx.distance);
 
-    /* Write the modified segment */
+       /* Mark the ways which are used */
 
-    WriteFile(fd,&segmentx,sizeof(SegmentX));
+       SetBit(segmentsx->usedway,segmentx.way);
 
-    index++;
+       /* Set the distance but keep the other flags except for area */
 
-    if(!(index%10000))
-       printf_middle("Measuring Segments: Segments=%"Pindex_t,index);
+       segmentx.distance=DISTANCE(DistanceX(nodex1,nodex2))|DISTFLAG(segmentx.distance);
+       segmentx.distance&=~SEGMENT_AREA;
+
+       /* Write the modified segment */
+
+       WriteFileBuffered(fd,&segmentx,sizeof(SegmentX));
+
+       good++;
+      }
+
+    total++;
+
+    if(!(total%10000))
+       printf_middle("Processing Segments: Segments=%"Pindex_t" Duplicates=%"Pindex_t,total,duplicate);
    }
 
+ segmentsx->number=good;
+
  /* Close the files */
 
- segmentsx->fd=CloseFile(segmentsx->fd);
- CloseFile(fd);
-
- /* Free the other now-unneeded indexes */
-
- free(nodesx->idata);
- nodesx->idata=NULL;
-
- free(waysx->idata);
- waysx->idata=NULL;
+ segmentsx->fd=CloseFileBuffered(segmentsx->fd);
+ CloseFileBuffered(fd);
 
  /* Unmap from memory / close the file */
 
 #if !SLIM
  nodesx->data=UnmapFile(nodesx->data);
+ waysx->data=UnmapFile(waysx->data);
 #else
- nodesx->fd=CloseFile(nodesx->fd);
+ nodesx->fd=SlimUnmapFile(nodesx->fd);
+ waysx->fd=SlimUnmapFile(waysx->fd);
 #endif
 
  /* Print the final message */
 
- printf_last("Measured Segments: Segments=%"Pindex_t,segmentsx->number);
+ printf_last("Processed Segments: Segments=%"Pindex_t" Duplicates=%"Pindex_t,total,duplicate);
 }
 
 
@@ -764,7 +503,11 @@ void MeasureSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx)
 
 void IndexSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx)
 {
- index_t index,i;
+ index_t index,start=0,i;
+ SegmentX *segmentx_list=NULL;
+#if SLIM
+ index_t length=0;
+#endif
 
  if(segmentsx->number==0)
     return;
@@ -790,14 +533,31 @@ void IndexSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx)
 #if !SLIM
  segmentsx->data=MapFileWriteable(segmentsx->filename_tmp);
 #else
- segmentsx->fd=ReOpenFileWriteable(segmentsx->filename_tmp);
+ segmentsx->fd=SlimMapFileWriteable(segmentsx->filename_tmp);
+
+ segmentx_list=(SegmentX*)malloc(1024*sizeof(SegmentX));
 #endif
 
- /* Read through the segments in reverse order */
+ /* Read through the segments in reverse order (in chunks to help slim mode) */
 
  for(index=segmentsx->number-1;index!=NO_SEGMENT;index--)
    {
-    SegmentX *segmentx=LookupSegmentX(segmentsx,index,1);
+    SegmentX *segmentx;
+
+    if((index%1024)==1023 || index==(segmentsx->number-1))
+      {
+       start=1024*(index/1024);
+
+#if !SLIM
+       segmentx_list=LookupSegmentX(segmentsx,start,1);
+#else
+       length=index-start+1;
+
+       SlimFetch(segmentsx->fd,segmentx_list,length*sizeof(SegmentX),start*sizeof(SegmentX));
+#endif
+      }
+
+    segmentx=segmentx_list+(index-start);
 
     if(nodesx->pdata)
       {
@@ -810,13 +570,16 @@ void IndexSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx)
 
     segmentx->next2=segmentsx->firstnode[segmentx->node2];
 
-    PutBackSegmentX(segmentsx,segmentx);
-
     segmentsx->firstnode[segmentx->node1]=index;
     segmentsx->firstnode[segmentx->node2]=index;
 
     if(!(index%10000))
        printf_middle("Indexing Segments: Segments=%"Pindex_t,segmentsx->number-index);
+
+#if SLIM
+    if(index==start)
+       SlimReplace(segmentsx->fd,segmentx_list,length*sizeof(SegmentX),start*sizeof(SegmentX));
+#endif
    }
 
  /* Unmap from memory / close the files */
@@ -824,7 +587,9 @@ void IndexSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx)
 #if !SLIM
  segmentsx->data=UnmapFile(segmentsx->data);
 #else
- segmentsx->fd=CloseFile(segmentsx->fd);
+ segmentsx->fd=SlimUnmapFile(segmentsx->fd);
+
+ free(segmentx_list);
 #endif
 
  /* Free the memory */
@@ -875,11 +640,11 @@ void RemovePrunedSegments(SegmentsX *segmentsx,WaysX *waysx)
 
  /* Re-open the file read-only and a new file writeable */
 
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
+ segmentsx->fd=ReOpenFileBuffered(segmentsx->filename_tmp);
 
  DeleteFile(segmentsx->filename_tmp);
 
- fd=OpenFileNew(segmentsx->filename_tmp);
+ fd=OpenFileBufferedNew(segmentsx->filename_tmp);
 
  /* Sort by node indexes */
 
@@ -893,8 +658,8 @@ void RemovePrunedSegments(SegmentsX *segmentsx,WaysX *waysx)
 
  /* Close the files */
 
- segmentsx->fd=CloseFile(segmentsx->fd);
- CloseFile(fd);
+ segmentsx->fd=CloseFileBuffered(segmentsx->fd);
+ CloseFileBuffered(fd);
 
  /* Print the final message */
 
@@ -948,16 +713,18 @@ void DeduplicateSuperSegments(SegmentsX *segmentsx,WaysX *waysx)
 #if !SLIM
  waysx->data=MapFile(waysx->filename_tmp);
 #else
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+ waysx->fd=SlimMapFile(waysx->filename_tmp);
+
+ InvalidateWayXCache(waysx->cache);
 #endif
 
  /* Re-open the file read-only and a new file writeable */
 
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
+ segmentsx->fd=ReOpenFileBuffered(segmentsx->filename_tmp);
 
  DeleteFile(segmentsx->filename_tmp);
 
- fd=OpenFileNew(segmentsx->filename_tmp);
+ fd=OpenFileBufferedNew(segmentsx->filename_tmp);
 
  /* Sort by node indexes */
 
@@ -972,15 +739,15 @@ void DeduplicateSuperSegments(SegmentsX *segmentsx,WaysX *waysx)
 
  /* Close the files */
 
- segmentsx->fd=CloseFile(segmentsx->fd);
- CloseFile(fd);
+ segmentsx->fd=CloseFileBuffered(segmentsx->fd);
+ CloseFileBuffered(fd);
 
  /* Unmap from memory / close the file */
 
 #if !SLIM
  waysx->data=UnmapFile(waysx->data);
 #else
- waysx->fd=CloseFile(waysx->fd);
+ waysx->fd=SlimUnmapFile(waysx->fd);
 #endif
 
  /* Print the final message */
@@ -1077,11 +844,11 @@ void SortSegmentListGeographically(SegmentsX *segmentsx,NodesX *nodesx)
 
  /* Re-open the file read-only and a new file writeable */
 
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
+ segmentsx->fd=ReOpenFileBuffered(segmentsx->filename_tmp);
 
  DeleteFile(segmentsx->filename_tmp);
 
- fd=OpenFileNew(segmentsx->filename_tmp);
+ fd=OpenFileBufferedNew(segmentsx->filename_tmp);
 
  /* Update the segments with geographically sorted node indexes and sort them */
 
@@ -1092,8 +859,8 @@ void SortSegmentListGeographically(SegmentsX *segmentsx,NodesX *nodesx)
                                                   NULL);
  /* Close the files */
 
- segmentsx->fd=CloseFile(segmentsx->fd);
- CloseFile(fd);
+ segmentsx->fd=CloseFileBuffered(segmentsx->fd);
+ CloseFileBuffered(fd);
 
  /* Print the final message */
 
@@ -1153,20 +920,20 @@ void SaveSegmentList(SegmentsX *segmentsx,const char *filename)
 
  /* Re-open the file */
 
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
+ segmentsx->fd=ReOpenFileBuffered(segmentsx->filename_tmp);
 
  /* Write out the segments data */
 
- fd=OpenFileNew(filename);
+ fd=OpenFileBufferedNew(filename);
 
- SeekFile(fd,sizeof(SegmentsFile));
+ SeekFileBuffered(fd,sizeof(SegmentsFile));
 
  for(i=0;i<segmentsx->number;i++)
    {
     SegmentX segmentx;
     Segment  segment={0};
 
-    ReadFile(segmentsx->fd,&segmentx,sizeof(SegmentX));
+    ReadFileBuffered(segmentsx->fd,&segmentx,sizeof(SegmentX));
 
     segment.node1   =segmentx.node1;
     segment.node2   =segmentx.node2;
@@ -1179,7 +946,7 @@ void SaveSegmentList(SegmentsX *segmentsx,const char *filename)
     if(IsNormalSegment(&segment))
        normal_number++;
 
-    WriteFile(fd,&segment,sizeof(Segment));
+    WriteFileBuffered(fd,&segment,sizeof(Segment));
 
     if(!((i+1)%10000))
        printf_middle("Writing Segments: Segments=%"Pindex_t,i+1);
@@ -1191,14 +958,14 @@ void SaveSegmentList(SegmentsX *segmentsx,const char *filename)
  segmentsfile.snumber=super_number;
  segmentsfile.nnumber=normal_number;
 
- SeekFile(fd,0);
- WriteFile(fd,&segmentsfile,sizeof(SegmentsFile));
+ SeekFileBuffered(fd,0);
+ WriteFileBuffered(fd,&segmentsfile,sizeof(SegmentsFile));
 
- CloseFile(fd);
+ CloseFileBuffered(fd);
 
  /* Close the file */
 
- segmentsx->fd=CloseFile(segmentsx->fd);
+ segmentsx->fd=CloseFileBuffered(segmentsx->fd);
 
  /* Print the final message */
 
