@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -29,6 +29,7 @@
 
 #include "typesx.h"
 
+#include "cache.h"
 #include "files.h"
 
 
@@ -38,12 +39,12 @@
 /*+ An extended structure used for processing. +*/
 struct _SegmentX
 {
- node_t     node1;              /*+ The id of the starting node; initially the OSM value, later the NodeX index. +*/
- node_t     node2;              /*+ The id of the finishing node; initially the OSM value, later the NodeX index. +*/
+ index_t    node1;              /*+ The NodeX index of the starting node. +*/
+ index_t    node2;              /*+ The NodeX index of the finishing node. +*/
 
  index_t    next2;              /*+ The index of the next segment with the same node2. +*/
 
- way_t      way;                /*+ The id of the way; initially the OSM value, later the WayX index. +*/
+ index_t    way;                /*+ The WayX index of the way. +*/
 
  distance_t distance;           /*+ The distance between the nodes. +*/
 };
@@ -52,7 +53,6 @@ struct _SegmentX
 /*+ A structure containing a set of segments (memory format). +*/
 struct _SegmentsX
 {
- char      *filename;           /*+ The name of the intermediate file (for the SegmentsX). +*/
  char      *filename_tmp;       /*+ The name of the temporary file (for the SegmentsX). +*/
 
  int        fd;                 /*+ The file descriptor of the open file (for the SegmentsX). +*/
@@ -68,38 +68,34 @@ struct _SegmentsX
  SegmentX   cached[4];          /*+ Four cached extended segments read from the file in slim mode. +*/
  index_t    incache[4];         /*+ The indexes of the cached extended segments. +*/
 
+ SegmentXCache *cache;          /*+ A RAM cache of extended segments read from the file. +*/
+
 #endif
 
  index_t   *firstnode;          /*+ The first segment index for each node. +*/
 
  index_t   *next1;              /*+ The index of the next segment with the same node1 (used while pruning). +*/
 
- BitMask   *usednode;           /*+ A flag to indicate if a node is used (used for removing bad segments). +*/
-
  BitMask   *usedway;            /*+ A flag to indicate if a way is used (used for removing pruned ways). +*/
 };
 
 
 /* Functions in segmentsx.c */
 
-SegmentsX *NewSegmentList(int append,int readonly);
-void FreeSegmentList(SegmentsX *segmentsx,int keep);
+SegmentsX *NewSegmentList(void);
+void FreeSegmentList(SegmentsX *segmentsx);
 
-void AppendSegmentList(SegmentsX *segmentsx,way_t way,node_t node1,node_t node2,distance_t distance);
+void AppendSegmentList(SegmentsX *segmentsx,index_t way,index_t node1,index_t node2,distance_t distance);
 void FinishSegmentList(SegmentsX *segmentsx);
 
 SegmentX *FirstSegmentX(SegmentsX *segmentsx,index_t nodeindex,int position);
 SegmentX *NextSegmentX(SegmentsX *segmentsx,SegmentX *segmentx,index_t nodeindex);
 
-void ApplySegmentChanges(SegmentsX *segmentsx);
-
 void SortSegmentList(SegmentsX *segmentsx);
 
 void IndexSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx);
 
-void RemoveBadSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx,int keep);
-
-void MeasureSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx);
+void ProcessSegments(SegmentsX *segmentsx,NodesX *nodesx,WaysX *waysx);
 
 void RemovePrunedSegments(SegmentsX *segmentsx,WaysX *waysx);
 
@@ -122,19 +118,37 @@ void SaveSegmentList(SegmentsX *segmentsx,const char *filename);
 
 #define IndexSegmentX(segmentsx,segmentx)                (index_t)((segmentx)-&(segmentsx)->data[0])
 
-#define PutBackSegmentX(segmentsx,segmentx)              /* nop */
+#define PutBackSegmentX(segmentsx,segmentx)              while(0) { /* nop */ }
 
-#define ReLookupSegmentX(segmentsx,segmentx)             /* nop */
+#define ReLookupSegmentX(segmentsx,segmentx)             while(0) { /* nop */ }
   
 #else
 
-static SegmentX *LookupSegmentX(SegmentsX *segmentsx,index_t index,int position);
+/* Prototypes */
 
-static index_t IndexSegmentX(SegmentsX *segmentsx,SegmentX *segmentx);
+static inline SegmentX *LookupSegmentX(SegmentsX *segmentsx,index_t index,int position);
 
-static void PutBackSegmentX(SegmentsX *segmentsx,SegmentX *segmentx);
+static inline index_t IndexSegmentX(SegmentsX *segmentsx,SegmentX *segmentx);
 
-static void ReLookupSegmentX(SegmentsX *segmentsx,SegmentX *segmentx);
+static inline void PutBackSegmentX(SegmentsX *segmentsx,SegmentX *segmentx);
+
+static inline void ReLookupSegmentX(SegmentsX *segmentsx,SegmentX *segmentx);
+
+CACHE_NEWCACHE_PROTO(SegmentX)
+CACHE_DELETECACHE_PROTO(SegmentX)
+CACHE_FETCHCACHE_PROTO(SegmentX)
+CACHE_REPLACECACHE_PROTO(SegmentX)
+CACHE_INVALIDATECACHE_PROTO(SegmentX)
+
+
+/* Inline functions */
+
+CACHE_STRUCTURE(SegmentX)
+CACHE_NEWCACHE(SegmentX)
+CACHE_DELETECACHE(SegmentX)
+CACHE_FETCHCACHE(SegmentX)
+CACHE_REPLACECACHE(SegmentX)
+CACHE_INVALIDATECACHE(SegmentX)
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -151,7 +165,7 @@ static void ReLookupSegmentX(SegmentsX *segmentsx,SegmentX *segmentx);
 
 static inline SegmentX *LookupSegmentX(SegmentsX *segmentsx,index_t index,int position)
 {
- SeekReadFile(segmentsx->fd,&segmentsx->cached[position-1],sizeof(SegmentX),(off_t)index*sizeof(SegmentX));
+ segmentsx->cached[position-1]=*FetchCachedSegmentX(segmentsx->cache,index,segmentsx->fd,0);
 
  segmentsx->incache[position-1]=index;
 
@@ -189,7 +203,7 @@ static inline void PutBackSegmentX(SegmentsX *segmentsx,SegmentX *segmentx)
 {
  int position1=segmentx-&segmentsx->cached[0];
 
- SeekWriteFile(segmentsx->fd,&segmentsx->cached[position1],sizeof(SegmentX),(off_t)segmentsx->incache[position1]*sizeof(SegmentX));
+ ReplaceCachedSegmentX(segmentsx->cache,segmentx,segmentsx->incache[position1],segmentsx->fd,0);
 }
 
 
@@ -205,7 +219,7 @@ static inline void ReLookupSegmentX(SegmentsX *segmentsx,SegmentX *segmentx)
 {
  int position1=segmentx-&segmentsx->cached[0];
 
- SeekReadFile(segmentsx->fd,&segmentsx->cached[position1],sizeof(SegmentX),(off_t)segmentsx->incache[position1]*sizeof(SegmentX));
+ segmentsx->cached[position1]=*FetchCachedSegmentX(segmentsx->cache,segmentsx->incache[position1],segmentsx->fd,0);
 }
 
 #endif /* SLIM */
