@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -77,26 +77,32 @@ void ChooseSuperNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
 
  /* Map into memory / open the files */
 
+ nodesx->fd=ReOpenFileBuffered(nodesx->filename_tmp);
+
 #if !SLIM
- nodesx->data=MapFile(nodesx->filename_tmp);
  segmentsx->data=MapFile(segmentsx->filename_tmp);
  waysx->data=MapFile(waysx->filename_tmp);
 #else
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+ segmentsx->fd=SlimMapFile(segmentsx->filename_tmp);
+ waysx->fd=SlimMapFile(waysx->filename_tmp);
+
+ InvalidateSegmentXCache(segmentsx->cache);
+ InvalidateWayXCache(waysx->cache);
 #endif
 
  /* Find super-nodes */
 
  for(i=0;i<nodesx->number;i++)
    {
+    NodeX nodex;
+
+    ReadFileBuffered(nodesx->fd,&nodex,sizeof(NodeX));
+
     if(IsBitSet(nodesx->super,i))
       {
        int issuper=0;
-       NodeX *nodex=LookupNodeX(nodesx,i,1);
 
-       if(nodex->flags&(NODE_TURNRSTRCT|NODE_TURNRSTRCT2))
+       if(nodex.flags&(NODE_TURNRSTRCT|NODE_TURNRSTRCT2))
           issuper=1;
        else
          {
@@ -123,7 +129,7 @@ void ChooseSuperNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
 
              /* If the node allows less traffic types than any connecting way then it is super if it allows anything */
 
-             if((wayx->way.allow&nodex->allow)!=wayx->way.allow && nodex->allow!=Transports_None)
+             if((wayx->way.allow&nodex.allow)!=wayx->way.allow && nodex.allow!=Transports_None)
                {
                 issuper=1;
                 break;
@@ -176,15 +182,15 @@ void ChooseSuperNodes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
  /* Unmap from memory / close the files */
 
 #if !SLIM
- nodesx->data=UnmapFile(nodesx->data);
  segmentsx->data=UnmapFile(segmentsx->data);
  waysx->data=UnmapFile(waysx->data);
 #else
- nodesx->fd=CloseFile(nodesx->fd);
- segmentsx->fd=CloseFile(segmentsx->fd);
- waysx->fd=CloseFile(waysx->fd);
+ segmentsx->fd=SlimUnmapFile(segmentsx->fd);
+ waysx->fd=SlimUnmapFile(waysx->fd);
 #endif
 
+ nodesx->fd=CloseFileBuffered(nodesx->fd);
+
  /* Print the final message */
 
  printf_last("Found Super-Nodes: Nodes=%"Pindex_t" Super-Nodes=%"Pindex_t,nodesx->number,nnodes);
@@ -209,7 +215,7 @@ SegmentsX *CreateSuperSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
  SegmentsX *supersegmentsx;
  index_t sn=0,ss=0;
 
- supersegmentsx=NewSegmentList(0,0);
+ supersegmentsx=NewSegmentList();
 
  if(segmentsx->number==0 || waysx->number==0)
    {
@@ -229,9 +235,13 @@ SegmentsX *CreateSuperSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
  segmentsx->data=MapFile(segmentsx->filename_tmp);
  waysx->data=MapFile(waysx->filename_tmp);
 #else
- nodesx->fd=ReOpenFile(nodesx->filename_tmp);
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+ nodesx->fd=SlimMapFile(nodesx->filename_tmp);
+ segmentsx->fd=SlimMapFile(segmentsx->filename_tmp);
+ waysx->fd=SlimMapFile(waysx->filename_tmp);
+
+ InvalidateNodeXCache(nodesx->cache);
+ InvalidateSegmentXCache(segmentsx->cache);
+ InvalidateWayXCache(waysx->cache);
 #endif
 
  /* Create super-segments for each super-node. */
@@ -291,8 +301,6 @@ SegmentsX *CreateSuperSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
 
                 result=NextResult(results,result);
                }
-
-             FreeResultsList(results);
             }
 
           segmentx=NextSegmentX(segmentsx,segmentx,i);
@@ -305,6 +313,8 @@ SegmentsX *CreateSuperSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
       }
    }
 
+ FinishSegmentList(supersegmentsx);
+
  /* Unmap from memory / close the files */
 
 #if !SLIM
@@ -312,17 +322,15 @@ SegmentsX *CreateSuperSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
  segmentsx->data=UnmapFile(segmentsx->data);
  waysx->data=UnmapFile(waysx->data);
 #else
- nodesx->fd=CloseFile(nodesx->fd);
- segmentsx->fd=CloseFile(segmentsx->fd);
- waysx->fd=CloseFile(waysx->fd);
+ nodesx->fd=SlimUnmapFile(nodesx->fd);
+ segmentsx->fd=SlimUnmapFile(segmentsx->fd);
+ waysx->fd=SlimUnmapFile(waysx->fd);
 #endif
 
  /* Print the final message */
 
  printf_last("Created Super-Segments: Super-Nodes=%"Pindex_t" Super-Segments=%"Pindex_t,sn,ss);
 
- FinishSegmentList(supersegmentsx);
-
  return(supersegmentsx);
 }
 
@@ -339,11 +347,11 @@ SegmentsX *CreateSuperSegments(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx)
 
 SegmentsX *MergeSuperSegments(SegmentsX *segmentsx,SegmentsX *supersegmentsx)
 {
- index_t i,j;
+ index_t i,j,lastj;
  index_t merged=0,added=0;
  SegmentsX *mergedsegmentsx;
 
- mergedsegmentsx=NewSegmentList(0,0);
+ mergedsegmentsx=NewSegmentList();
 
  if(segmentsx->number==0)
    {
@@ -356,32 +364,37 @@ SegmentsX *MergeSuperSegments(SegmentsX *segmentsx,SegmentsX *supersegmentsx)
 
  printf_first("Merging Segments: Segments=0 Super=0 Merged=0 Added=0");
 
- /* Map into memory / open the files */
+ /* Open the files */
 
-#if !SLIM
- segmentsx->data=MapFile(segmentsx->filename_tmp);
+ segmentsx->fd=ReOpenFileBuffered(segmentsx->filename_tmp);
  if(supersegmentsx->number>0)
-    supersegmentsx->data=MapFile(supersegmentsx->filename_tmp);
-#else
- segmentsx->fd=ReOpenFile(segmentsx->filename_tmp);
- if(supersegmentsx->number>0)
-    supersegmentsx->fd=ReOpenFile(supersegmentsx->filename_tmp);
-#endif
+    supersegmentsx->fd=ReOpenFileBuffered(supersegmentsx->filename_tmp);
 
  /* Loop through and create a new list of combined segments */
 
- for(i=0,j=0;i<segmentsx->number;i++)
+ lastj=-1;
+ j=0;
+
+ for(i=0;i<segmentsx->number;i++)
    {
     int super=0;
-    SegmentX *segmentx=LookupSegmentX(segmentsx,i,1);
+    SegmentX segmentx;
+
+    ReadFileBuffered(segmentsx->fd,&segmentx,sizeof(SegmentX));
 
     while(j<supersegmentsx->number)
       {
-       SegmentX *supersegmentx=LookupSegmentX(supersegmentsx,j,1);
+       SegmentX supersegmentx;
 
-       if(segmentx->node1   ==supersegmentx->node1 &&
-          segmentx->node2   ==supersegmentx->node2 &&
-          segmentx->distance==supersegmentx->distance)
+       if(j!=lastj)
+         {
+          ReadFileBuffered(supersegmentsx->fd,&supersegmentx,sizeof(SegmentX));
+          lastj=j;
+         }
+
+       if(segmentx.node1   ==supersegmentx.node1 &&
+          segmentx.node2   ==supersegmentx.node2 &&
+          segmentx.distance==supersegmentx.distance)
          {
           merged++;
           j++;
@@ -389,14 +402,14 @@ SegmentsX *MergeSuperSegments(SegmentsX *segmentsx,SegmentsX *supersegmentsx)
           super=1;
           break;
          }
-       else if((segmentx->node1==supersegmentx->node1 &&
-                segmentx->node2==supersegmentx->node2) ||
-               (segmentx->node1==supersegmentx->node1 &&
-                segmentx->node2>supersegmentx->node2) ||
-               (segmentx->node1>supersegmentx->node1))
+       else if((segmentx.node1==supersegmentx.node1 &&
+                segmentx.node2==supersegmentx.node2) ||
+               (segmentx.node1==supersegmentx.node1 &&
+                segmentx.node2>supersegmentx.node2) ||
+               (segmentx.node1>supersegmentx.node1))
          {
           /* mark as super-segment */
-          AppendSegmentList(mergedsegmentsx,supersegmentx->way,supersegmentx->node1,supersegmentx->node2,supersegmentx->distance|SEGMENT_SUPER);
+          AppendSegmentList(mergedsegmentsx,supersegmentx.way,supersegmentx.node1,supersegmentx.node2,supersegmentx.distance|SEGMENT_SUPER);
           added++;
           j++;
          }
@@ -408,32 +421,37 @@ SegmentsX *MergeSuperSegments(SegmentsX *segmentsx,SegmentsX *supersegmentsx)
       }
 
     if(super)
-       AppendSegmentList(mergedsegmentsx,segmentx->way,segmentx->node1,segmentx->node2,segmentx->distance|SEGMENT_SUPER|SEGMENT_NORMAL);
+       AppendSegmentList(mergedsegmentsx,segmentx.way,segmentx.node1,segmentx.node2,segmentx.distance|SEGMENT_SUPER|SEGMENT_NORMAL);
     else
-       AppendSegmentList(mergedsegmentsx,segmentx->way,segmentx->node1,segmentx->node2,segmentx->distance|SEGMENT_NORMAL);
+       AppendSegmentList(mergedsegmentsx,segmentx.way,segmentx.node1,segmentx.node2,segmentx.distance|SEGMENT_NORMAL);
 
     if(!((i+1)%10000))
        printf_middle("Merging Segments: Segments=%"Pindex_t" Super=%"Pindex_t" Merged=%"Pindex_t" Added=%"Pindex_t,i+1,j,merged,added);
    }
 
- /* Unmap from memory / close the files */
+ while(j<supersegmentsx->number)
+   {
+    SegmentX supersegmentx;
 
-#if !SLIM
- segmentsx->data=UnmapFile(segmentsx->data);
+    ReadFileBuffered(supersegmentsx->fd,&supersegmentx,sizeof(SegmentX));
+
+    AppendSegmentList(mergedsegmentsx,supersegmentx.way,supersegmentx.node1,supersegmentx.node2,supersegmentx.distance|SEGMENT_SUPER);
+    added++;
+    j++;
+   }
+
+ FinishSegmentList(mergedsegmentsx);
+
+ /* Close the files */
+
+ segmentsx->fd=CloseFileBuffered(segmentsx->fd);
  if(supersegmentsx->number>0)
-    supersegmentsx->data=UnmapFile(supersegmentsx->data);
-#else
- segmentsx->fd=CloseFile(segmentsx->fd);
- if(supersegmentsx->number>0)
-    supersegmentsx->fd=CloseFile(supersegmentsx->fd);
-#endif
+    supersegmentsx->fd=CloseFileBuffered(supersegmentsx->fd);
 
  /* Print the final message */
 
  printf_last("Merged Segments: Segments=%"Pindex_t" Super=%"Pindex_t" Merged=%"Pindex_t" Added=%"Pindex_t,segmentsx->number,supersegmentsx->number,merged,added);
 
- FinishSegmentList(mergedsegmentsx);
-
  return(mergedsegmentsx);
 }
 
@@ -456,20 +474,26 @@ SegmentsX *MergeSuperSegments(SegmentsX *segmentsx,SegmentsX *supersegmentsx)
 
 static Results *FindSuperRoutes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx,node_t start,Way *match)
 {
- Results *results;
- Queue *queue;
+ static Results *results=NULL;
+ static Queue *queue=NULL;
  Result *result1,*result2;
  WayX *wayx;
 
  /* Insert the first node into the queue */
 
- results=NewResultsList(64);
+ if(!results)
+    results=NewResultsList(8);
+ else
+    ResetResultsList(results);
 
- queue=NewQueueList();
+ if(!queue)
+    queue=NewQueueList(8);
+ else
+    ResetQueueList(queue);
 
  result1=InsertResult(results,start,NO_SEGMENT);
 
- InsertInQueue(queue,result1);
+ InsertInQueue(queue,result1,0);
 
  /* Loop across all nodes in the queue */
 
@@ -521,21 +545,19 @@ static Results *FindSuperRoutes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx
           result2=InsertResult(results,node2,seg2);
           result2->prev=result1;
           result2->score=cumulative_distance;
-          result2->sortby=cumulative_distance;
 
           /* don't route beyond a super-node. */
           if(!IsBitSet(nodesx->super,node2))
-             InsertInQueue(queue,result2);
+             InsertInQueue(queue,result2,cumulative_distance);
          }
        else if(cumulative_distance<result2->score)
          {
           result2->prev=result1;
           result2->score=cumulative_distance;
-          result2->sortby=cumulative_distance;
 
           /* don't route beyond a super-node. */
           if(!IsBitSet(nodesx->super,node2))
-             InsertInQueue(queue,result2);
+             InsertInQueue(queue,result2,cumulative_distance);
          }
 
       endloop:
@@ -544,7 +566,5 @@ static Results *FindSuperRoutes(NodesX *nodesx,SegmentsX *segmentsx,WaysX *waysx
       }
    }
 
- FreeQueueList(queue);
-
  return(results);
 }
