@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2010-2012 Andrew M. Bishop
+ This file Copyright 2010-2014 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -23,6 +23,8 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
+#include <inttypes.h>
+#include <stdint.h>
 
 #include "files.h"
 #include "tagging.h"
@@ -30,39 +32,126 @@
 #include "logging.h"
 
 
-/* Global variables */
+/* Constants */
 
-TaggingRuleList NodeRules={NULL,0};
-TaggingRuleList WayRules={NULL,0};
-TaggingRuleList RelationRules={NULL,0};
+#define TAGACTION_IF       1
+#define TAGACTION_IFNOT    2
+
+#define TAGACTION_INHERIT  3    /* Not a real action, just a marker */
+
+#define TAGACTION_SET      4
+#define TAGACTION_UNSET    5
+#define TAGACTION_OUTPUT   6
+#define TAGACTION_LOGERROR 7
 
 
 /* Local variables */
 
-TaggingRuleList *current_list=NULL;
-TaggingRule     *current_rule=NULL;
+static TaggingRuleList NodeRules={NULL,0};
+static TaggingRuleList WayRules={NULL,0};
+static TaggingRuleList RelationRules={NULL,0};
+
+static int current_list_stack_depth=0;
+static TaggingRuleList **current_list_stack=NULL;
+static TaggingRuleList *current_list=NULL;
+
+static int64_t current_id;
+
+static char *default_logerror_message="ignoring it";
 
 
 /* Local functions */
 
-static void apply_actions(TaggingRuleList *rules,TaggingRule *rule,int match,TagList *input,TagList *output,node_t id);
+static TaggingRuleList *AppendTaggingRule(TaggingRuleList *rules,const char *k,const char *v,int action);
+static void AppendTaggingAction(TaggingRuleList *rules,const char *k,const char *v,int action,const char *message);
+static void DeleteTaggingRuleList(TaggingRuleList *rules);
+
+static void ApplyRules(TaggingRuleList *rules,TagList *input,TagList *output,const char *match_k,const char *match_v);
 
 
 /* The XML tag processing function prototypes */
 
 //static int xmlDeclaration_function(const char *_tag_,int _type_,const char *version,const char *encoding);
 //static int RoutinoTaggingType_function(const char *_tag_,int _type_);
-static int RelationType_function(const char *_tag_,int _type_);
-static int WayType_function(const char *_tag_,int _type_);
 static int NodeType_function(const char *_tag_,int _type_);
+static int WayType_function(const char *_tag_,int _type_);
+static int RelationType_function(const char *_tag_,int _type_);
 static int IfType_function(const char *_tag_,int _type_,const char *k,const char *v);
-static int LogErrorType_function(const char *_tag_,int _type_,const char *k,const char *v);
-static int OutputType_function(const char *_tag_,int _type_,const char *k,const char *v);
-static int UnsetType_function(const char *_tag_,int _type_,const char *k);
+static int IfNotType_function(const char *_tag_,int _type_,const char *k,const char *v);
 static int SetType_function(const char *_tag_,int _type_,const char *k,const char *v);
+static int UnsetType_function(const char *_tag_,int _type_,const char *k);
+static int OutputType_function(const char *_tag_,int _type_,const char *k,const char *v);
+static int LogErrorType_function(const char *_tag_,int _type_,const char *k,const char *v,const char *message);
 
 
-/* The XML tag definitions */
+/* The XML tag definitions (forward declarations) */
+
+static xmltag xmlDeclaration_tag;
+static xmltag RoutinoTaggingType_tag;
+static xmltag NodeType_tag;
+static xmltag WayType_tag;
+static xmltag RelationType_tag;
+static xmltag IfType_tag;
+static xmltag IfNotType_tag;
+static xmltag SetType_tag;
+static xmltag UnsetType_tag;
+static xmltag OutputType_tag;
+static xmltag LogErrorType_tag;
+
+
+/* The XML tag definition values */
+
+/*+ The complete set of tags at the top level. +*/
+static xmltag *xml_toplevel_tags[]={&xmlDeclaration_tag,&RoutinoTaggingType_tag,NULL};
+
+/*+ The xmlDeclaration type tag. +*/
+static xmltag xmlDeclaration_tag=
+              {"xml",
+               2, {"version","encoding"},
+               NULL,
+               {NULL}};
+
+/*+ The RoutinoTaggingType type tag. +*/
+static xmltag RoutinoTaggingType_tag=
+              {"routino-tagging",
+               0, {NULL},
+               NULL,
+               {&NodeType_tag,&WayType_tag,&RelationType_tag,NULL}};
+
+/*+ The NodeType type tag. +*/
+static xmltag NodeType_tag=
+              {"node",
+               0, {NULL},
+               NodeType_function,
+               {&IfType_tag,&IfNotType_tag,NULL}};
+
+/*+ The WayType type tag. +*/
+static xmltag WayType_tag=
+              {"way",
+               0, {NULL},
+               WayType_function,
+               {&IfType_tag,&IfNotType_tag,NULL}};
+
+/*+ The RelationType type tag. +*/
+static xmltag RelationType_tag=
+              {"relation",
+               0, {NULL},
+               RelationType_function,
+               {&IfType_tag,&IfNotType_tag,NULL}};
+
+/*+ The IfType type tag. +*/
+static xmltag IfType_tag=
+              {"if",
+               2, {"k","v"},
+               IfType_function,
+               {&IfType_tag,&IfNotType_tag,&SetType_tag,&UnsetType_tag,&OutputType_tag,&LogErrorType_tag,NULL}};
+
+/*+ The IfNotType type tag. +*/
+static xmltag IfNotType_tag=
+              {"ifnot",
+               2, {"k","v"},
+               IfNotType_function,
+               {&IfType_tag,&IfNotType_tag,&SetType_tag,&UnsetType_tag,&OutputType_tag,&LogErrorType_tag,NULL}};
 
 /*+ The SetType type tag. +*/
 static xmltag SetType_tag=
@@ -88,248 +177,14 @@ static xmltag OutputType_tag=
 /*+ The LogErrorType type tag. +*/
 static xmltag LogErrorType_tag=
               {"logerror",
-               2, {"k","v"},
+               3, {"k","v","message"},
                LogErrorType_function,
                {NULL}};
 
-/*+ The IfType type tag. +*/
-static xmltag IfType_tag=
-              {"if",
-               2, {"k","v"},
-               IfType_function,
-               {&SetType_tag,&UnsetType_tag,&OutputType_tag,&LogErrorType_tag,NULL}};
-
-/*+ The NodeType type tag. +*/
-static xmltag NodeType_tag=
-              {"node",
-               0, {NULL},
-               NodeType_function,
-               {&IfType_tag,NULL}};
-
-/*+ The WayType type tag. +*/
-static xmltag WayType_tag=
-              {"way",
-               0, {NULL},
-               WayType_function,
-               {&IfType_tag,NULL}};
-
-/*+ The RelationType type tag. +*/
-static xmltag RelationType_tag=
-              {"relation",
-               0, {NULL},
-               RelationType_function,
-               {&IfType_tag,NULL}};
-
-/*+ The RoutinoTaggingType type tag. +*/
-static xmltag RoutinoTaggingType_tag=
-              {"routino-tagging",
-               0, {NULL},
-               NULL,
-               {&NodeType_tag,&WayType_tag,&RelationType_tag,NULL}};
-
-/*+ The xmlDeclaration type tag. +*/
-static xmltag xmlDeclaration_tag=
-              {"xml",
-               2, {"version","encoding"},
-               NULL,
-               {NULL}};
-
-
-/*+ The complete set of tags at the top level. +*/
-static xmltag *xml_toplevel_tags[]={&xmlDeclaration_tag,&RoutinoTaggingType_tag,NULL};
-
 
 /* The XML tag processing functions */
 
 
-/*++++++++++++++++++++++++++++++++++++++
-  The function that is called when the SetType XSD type is seen
-
-  int SetType_function Returns 0 if no error occured or something else otherwise.
-
-  const char *_tag_ Set to the name of the element tag that triggered this function call.
-
-  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
-
-  const char *k The contents of the 'k' attribute (or NULL if not defined).
-
-  const char *v The contents of the 'v' attribute (or NULL if not defined).
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int SetType_function(const char *_tag_,int _type_,const char *k,const char *v)
-{
- if(_type_&XMLPARSE_TAG_START)
-    AppendTaggingAction(current_rule,k,v,TAGACTION_SET);
-
- return(0);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  The function that is called when the UnsetType XSD type is seen
-
-  int UnsetType_function Returns 0 if no error occured or something else otherwise.
-
-  const char *_tag_ Set to the name of the element tag that triggered this function call.
-
-  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
-
-  const char *k The contents of the 'k' attribute (or NULL if not defined).
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int UnsetType_function(const char *_tag_,int _type_,const char *k)
-{
- if(_type_&XMLPARSE_TAG_START)
-    AppendTaggingAction(current_rule,k,NULL,TAGACTION_UNSET);
-
- return(0);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  The function that is called when the OutputType XSD type is seen
-
-  int OutputType_function Returns 0 if no error occured or something else otherwise.
-
-  const char *_tag_ Set to the name of the element tag that triggered this function call.
-
-  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
-
-  const char *k The contents of the 'k' attribute (or NULL if not defined).
-
-  const char *v The contents of the 'v' attribute (or NULL if not defined).
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int OutputType_function(const char *_tag_,int _type_,const char *k,const char *v)
-{
- if(_type_&XMLPARSE_TAG_START)
-    AppendTaggingAction(current_rule,k,v,TAGACTION_OUTPUT);
-
- return(0);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  The function that is called when the LogErrorType XSD type is seen
-
-  int LogErrorType_function Returns 0 if no error occured or something else otherwise.
-
-  const char *_tag_ Set to the name of the element tag that triggered this function call.
-
-  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
-
-  const char *k The contents of the 'k' attribute (or NULL if not defined).
-
-  const char *v The contents of the 'v' attribute (or NULL if not defined).
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int LogErrorType_function(const char *_tag_,int _type_,const char *k,const char *v)
-{
- if(_type_&XMLPARSE_TAG_START)
-    AppendTaggingAction(current_rule,k,v,TAGACTION_LOGERROR);
-
- return(0);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  The function that is called when the IfType XSD type is seen
-
-  int IfType_function Returns 0 if no error occured or something else otherwise.
-
-  const char *_tag_ Set to the name of the element tag that triggered this function call.
-
-  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
-
-  const char *k The contents of the 'k' attribute (or NULL if not defined).
-
-  const char *v The contents of the 'v' attribute (or NULL if not defined).
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int IfType_function(const char *_tag_,int _type_,const char *k,const char *v)
-{
- if(_type_&XMLPARSE_TAG_START)
-   {
-    current_rule=AppendTaggingRule(current_list,k,v);
-   }
-
- return(0);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  The function that is called when the NodeType XSD type is seen
-
-  int NodeType_function Returns 0 if no error occured or something else otherwise.
-
-  const char *_tag_ Set to the name of the element tag that triggered this function call.
-
-  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int NodeType_function(const char *_tag_,int _type_)
-{
- if(_type_&XMLPARSE_TAG_START)
-    current_list=&NodeRules;
-
- return(0);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  The function that is called when the WayType XSD type is seen
-
-  int WayType_function Returns 0 if no error occured or something else otherwise.
-
-  const char *_tag_ Set to the name of the element tag that triggered this function call.
-
-  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int WayType_function(const char *_tag_,int _type_)
-{
- if(_type_&XMLPARSE_TAG_START)
-    current_list=&WayRules;
-
- return(0);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  The function that is called when the RelationType XSD type is seen
-
-  int RelationType_function Returns 0 if no error occured or something else otherwise.
-
-  const char *_tag_ Set to the name of the element tag that triggered this function call.
-
-  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int RelationType_function(const char *_tag_,int _type_)
-{
- if(_type_&XMLPARSE_TAG_START)
-    current_list=&RelationRules;
-
- return(0);
-}
-
-
-/*++++++++++++++++++++++++++++++++++++++
-  The function that is called when the RoutinoTaggingType XSD type is seen
-
-  int RoutinoTaggingType_function Returns 0 if no error occured or something else otherwise.
-
-  const char *_tag_ Set to the name of the element tag that triggered this function call.
-
-  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-//static int RoutinoTaggingType_function(const char *_tag_,int _type_)
-//{
-// return(0);
-//}
-
-
 /*++++++++++++++++++++++++++++++++++++++
   The function that is called when the XML declaration is seen
 
@@ -350,6 +205,246 @@ static int RelationType_function(const char *_tag_,int _type_)
 //}
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the RoutinoTaggingType XSD type is seen
+
+  int RoutinoTaggingType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+//static int RoutinoTaggingType_function(const char *_tag_,int _type_)
+//{
+// return(0);
+//}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the NodeType XSD type is seen
+
+  int NodeType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int NodeType_function(const char *_tag_,int _type_)
+{
+ if(_type_&XMLPARSE_TAG_START)
+   {
+    current_list_stack_depth=0;
+    current_list=&NodeRules;
+   }
+
+ return(0);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the WayType XSD type is seen
+
+  int WayType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int WayType_function(const char *_tag_,int _type_)
+{
+ if(_type_&XMLPARSE_TAG_START)
+   {
+    current_list_stack_depth=0;
+    current_list=&WayRules;
+   }
+
+ return(0);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the RelationType XSD type is seen
+
+  int RelationType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int RelationType_function(const char *_tag_,int _type_)
+{
+ if(_type_&XMLPARSE_TAG_START)
+   {
+    current_list_stack_depth=0;
+    current_list=&RelationRules;
+   }
+
+ return(0);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the IfType XSD type is seen
+
+  int IfType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+
+  const char *k The contents of the 'k' attribute (or NULL if not defined).
+
+  const char *v The contents of the 'v' attribute (or NULL if not defined).
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int IfType_function(const char *_tag_,int _type_,const char *k,const char *v)
+{
+ if(_type_&XMLPARSE_TAG_START)
+   {
+    if(!current_list_stack || (current_list_stack_depth%8)==7)
+       current_list_stack=(TaggingRuleList**)realloc((void*)current_list_stack,(current_list_stack_depth+8)*sizeof(TaggingRuleList*));
+
+    current_list_stack[current_list_stack_depth++]=current_list;
+
+    current_list=AppendTaggingRule(current_list,k,v,TAGACTION_IF);
+   }
+
+ if(_type_&XMLPARSE_TAG_END)
+    current_list=current_list_stack[--current_list_stack_depth];
+
+ return(0);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the IfNotType XSD type is seen
+
+  int IfNotType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+
+  const char *k The contents of the 'k' attribute (or NULL if not defined).
+
+  const char *v The contents of the 'v' attribute (or NULL if not defined).
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int IfNotType_function(const char *_tag_,int _type_,const char *k,const char *v)
+{
+ if(_type_&XMLPARSE_TAG_START)
+   {
+    if(!current_list_stack || (current_list_stack_depth%8)==7)
+       current_list_stack=(TaggingRuleList**)realloc((void*)current_list_stack,(current_list_stack_depth+8)*sizeof(TaggingRuleList*));
+
+    current_list_stack[current_list_stack_depth++]=current_list;
+
+    current_list=AppendTaggingRule(current_list,k,v,TAGACTION_IFNOT);
+   }
+
+ if(_type_&XMLPARSE_TAG_END)
+    current_list=current_list_stack[--current_list_stack_depth];
+
+ return(0);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the SetType XSD type is seen
+
+  int SetType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+
+  const char *k The contents of the 'k' attribute (or NULL if not defined).
+
+  const char *v The contents of the 'v' attribute (or NULL if not defined).
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int SetType_function(const char *_tag_,int _type_,const char *k,const char *v)
+{
+ if(_type_&XMLPARSE_TAG_START)
+    AppendTaggingAction(current_list,k,v,TAGACTION_SET,NULL);
+
+ return(0);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the UnsetType XSD type is seen
+
+  int UnsetType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+
+  const char *k The contents of the 'k' attribute (or NULL if not defined).
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int UnsetType_function(const char *_tag_,int _type_,const char *k)
+{
+ if(_type_&XMLPARSE_TAG_START)
+    AppendTaggingAction(current_list,k,NULL,TAGACTION_UNSET,NULL);
+
+ return(0);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the OutputType XSD type is seen
+
+  int OutputType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+
+  const char *k The contents of the 'k' attribute (or NULL if not defined).
+
+  const char *v The contents of the 'v' attribute (or NULL if not defined).
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int OutputType_function(const char *_tag_,int _type_,const char *k,const char *v)
+{
+ if(_type_&XMLPARSE_TAG_START)
+    AppendTaggingAction(current_list,k,v,TAGACTION_OUTPUT,NULL);
+
+ return(0);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The function that is called when the LogErrorType XSD type is seen
+
+  int LogErrorType_function Returns 0 if no error occured or something else otherwise.
+
+  const char *_tag_ Set to the name of the element tag that triggered this function call.
+
+  int _type_ Set to XMLPARSE_TAG_START at the start of a tag and/or XMLPARSE_TAG_END at the end of a tag.
+
+  const char *k The contents of the 'k' attribute (or NULL if not defined).
+
+  const char *v The contents of the 'v' attribute (or NULL if not defined).
+
+  const char *message The contents of the 'message' attribute (or NULL if not defined).
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int LogErrorType_function(const char *_tag_,int _type_,const char *k,const char *v,const char *message)
+{
+ if(_type_&XMLPARSE_TAG_START)
+    AppendTaggingAction(current_list,k,v,TAGACTION_LOGERROR,message);
+
+ return(0);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   The XML tagging rules parser.
 
@@ -360,7 +455,7 @@ static int RelationType_function(const char *_tag_,int _type_)
 
 int ParseXMLTaggingRules(const char *filename)
 {
- FILE *file;
+ int fd;
  int retval;
 
  if(!ExistsFile(filename))
@@ -369,17 +464,14 @@ int ParseXMLTaggingRules(const char *filename)
     return(1);
    }
 
- file=fopen(filename,"r");
+ fd=OpenFile(filename);
 
- if(!file)
-   {
-    fprintf(stderr,"Error: Cannot open tagging rules file '%s' for reading.\n",filename);
-    return(1);
-   }
+ retval=ParseXML(fd,xml_toplevel_tags,XMLPARSE_UNKNOWN_ATTR_ERRNONAME);
 
- retval=ParseXML(file,xml_toplevel_tags,XMLPARSE_UNKNOWN_ATTR_ERRNONAME);
+ CloseFile(fd);
 
- fclose(file);
+ if(current_list_stack)
+    free(current_list_stack);
 
  if(retval)
     return(1);
@@ -403,22 +495,26 @@ void DeleteXMLTaggingRules(void)
 /*++++++++++++++++++++++++++++++++++++++
   Append a tagging rule to the list of rules.
 
-  TaggingRule *AppendTaggingRule Returns the latest rule (the just added one).
+  TaggingRuleList *AppendTaggingRule Returns the new TaggingRuleList inside the new TaggingRule.
 
   TaggingRuleList *rules The list of rules to add to.
 
   const char *k The tag key.
 
   const char *v The tag value.
+
+  int action Set to the type of action.
   ++++++++++++++++++++++++++++++++++++++*/
 
-TaggingRule *AppendTaggingRule(TaggingRuleList *rules,const char *k,const char *v)
+TaggingRuleList *AppendTaggingRule(TaggingRuleList *rules,const char *k,const char *v,int action)
 {
  if((rules->nrules%16)==0)
     rules->rules=(TaggingRule*)realloc((void*)rules->rules,(rules->nrules+16)*sizeof(TaggingRule));
 
  rules->nrules++;
 
+ rules->rules[rules->nrules-1].action=action;
+
  if(k)
     rules->rules[rules->nrules-1].k=strcpy(malloc(strlen(k)+1),k);
  else
@@ -429,43 +525,53 @@ TaggingRule *AppendTaggingRule(TaggingRuleList *rules,const char *k,const char *
  else
     rules->rules[rules->nrules-1].v=NULL;
 
- rules->rules[rules->nrules-1].nactions=0;
- rules->rules[rules->nrules-1].actions=NULL;
+ rules->rules[rules->nrules-1].message=NULL;
 
- return(&rules->rules[rules->nrules-1]);
+ rules->rules[rules->nrules-1].rulelist=(TaggingRuleList*)calloc(sizeof(TaggingRuleList),1);
+
+ return(rules->rules[rules->nrules-1].rulelist);
 }
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  Append a tagging action to a tagging rule.
+  Append a tagging action to the list of rules.
 
-  TaggingRule *rule The rule to add the action to.
+  TaggingRuleList *rules The list of rules to add to.
 
   const char *k The tag key.
 
   const char *v The tag value.
 
   int action Set to the type of action.
+
+  const char *message The message to use for the logerror action.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void AppendTaggingAction(TaggingRule *rule,const char *k,const char *v,int action)
+static void AppendTaggingAction(TaggingRuleList *rules,const char *k,const char *v,int action,const char *message)
 {
- if((rule->nactions%16)==0)
-    rule->actions=(TaggingAction*)realloc((void*)rule->actions,(rule->nactions+16)*sizeof(TaggingAction));
+ if((rules->nrules%16)==0)
+    rules->rules=(TaggingRule*)realloc((void*)rules->rules,(rules->nrules+16)*sizeof(TaggingRule));
 
- rule->nactions++;
+ rules->nrules++;
 
- rule->actions[rule->nactions-1].action=action;
+ rules->rules[rules->nrules-1].action=action;
 
  if(k)
-    rule->actions[rule->nactions-1].k=strcpy(malloc(strlen(k)+1),k);
+    rules->rules[rules->nrules-1].k=strcpy(malloc(strlen(k)+1),k);
  else
-    rule->actions[rule->nactions-1].k=NULL;
+    rules->rules[rules->nrules-1].k=NULL;
 
  if(v)
-    rule->actions[rule->nactions-1].v=strcpy(malloc(strlen(v)+1),v);
+    rules->rules[rules->nrules-1].v=strcpy(malloc(strlen(v)+1),v);
  else
-    rule->actions[rule->nactions-1].v=NULL;
+    rules->rules[rules->nrules-1].v=NULL;
+
+ if(message)
+    rules->rules[rules->nrules-1].message=strcpy(malloc(strlen(message)+1),message);
+ else
+    rules->rules[rules->nrules-1].message=default_logerror_message;
+
+ rules->rules[rules->nrules-1].rulelist=NULL;
 }
 
 
@@ -477,7 +583,7 @@ void AppendTaggingAction(TaggingRule *rule,const char *k,const char *v,int actio
 
 void DeleteTaggingRuleList(TaggingRuleList *rules)
 {
- int i,j;
+ int i;
 
  for(i=0;i<rules->nrules;i++)
    {
@@ -485,17 +591,14 @@ void DeleteTaggingRuleList(TaggingRuleList *rules)
        free(rules->rules[i].k);
     if(rules->rules[i].v)
        free(rules->rules[i].v);
+    if(rules->rules[i].message && rules->rules[i].message!=default_logerror_message)
+       free(rules->rules[i].message);
 
-    for(j=0;j<rules->rules[i].nactions;j++)
+    if(rules->rules[i].rulelist)
       {
-       if(rules->rules[i].actions[j].k)
-          free(rules->rules[i].actions[j].k);
-       if(rules->rules[i].actions[j].v)
-          free(rules->rules[i].actions[j].v);
+       DeleteTaggingRuleList(rules->rules[i].rulelist);
+       free(rules->rules[i].rulelist);
       }
-
-    if(rules->rules[i].actions)
-       free(rules->rules[i].actions);
    }
 
  if(rules->rules)
@@ -531,11 +634,8 @@ void DeleteTagList(TagList *tags)
     if(tags->v[i]) free(tags->v[i]);
    }
 
- if(tags->ntags)
-   {
-    free(tags->k);
-    free(tags->v);
-   }
+ if(tags->k) free(tags->k);
+ if(tags->v) free(tags->v);
 
  free(tags);
 }
@@ -553,14 +653,14 @@ void DeleteTagList(TagList *tags)
 
 void AppendTag(TagList *tags,const char *k,const char *v)
 {
- if((tags->ntags%16)==0)
+ if((tags->ntags%8)==0)
    {
     int i;
 
-    tags->k=(char**)realloc((void*)tags->k,(tags->ntags+16)*sizeof(char*));
-    tags->v=(char**)realloc((void*)tags->v,(tags->ntags+16)*sizeof(char*));
+    tags->k=(char**)realloc((void*)tags->k,(tags->ntags+8)*sizeof(char*));
+    tags->v=(char**)realloc((void*)tags->v,(tags->ntags+8)*sizeof(char*));
 
-    for(i=tags->ntags;i<(tags->ntags+16);i++)
+    for(i=tags->ntags;i<(tags->ntags+8);i++)
        tags->k[i]=tags->v[i]=NULL;
    }
 
@@ -631,101 +731,240 @@ void DeleteTag(TagList *tags,const char *k)
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  Apply a set of tagging rules to a set of tags.
+  Create a string containing all of the tags formatted as if HTML.
 
-  TagList *ApplyTaggingRules Returns the list of output tags after modification.
+  char *StringifyTag Returns a static pointer to the created string.
 
-  TaggingRuleList *rules The tagging rules to apply.
+  TagList *tags The list of tags to convert.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+char *StringifyTag(TagList *tags)
+{
+ static char *string=NULL;
+ int i,length=0,used=0;
+
+ for(i=0;i<tags->ntags;i++)
+   {
+    length+=strlen(tags->k[i]);
+    length+=strlen(tags->v[i]);
+
+    length+=16;
+   }
+
+ string=realloc((char*)string,length);
+
+ for(i=0;i<tags->ntags;i++)
+    used+=sprintf(string+used,"<tag k='%s' v='%s'>",tags->k[i],tags->v[i]);
+
+ return(string);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Apply a set of tagging rules to a set of node tags.
+
+  TagList *ApplyNodeTaggingRules Returns the list of output tags after modification.
 
   TagList *tags The tags to be modified.
 
-  node_t id The ID of the node, way or relation.
+  int64_t id The ID of the node.
   ++++++++++++++++++++++++++++++++++++++*/
 
-TagList *ApplyTaggingRules(TaggingRuleList *rules,TagList *tags,node_t id)
+TagList *ApplyNodeTaggingRules(TagList *tags,int64_t id)
 {
  TagList *result=NewTagList();
- int i,j;
 
- for(i=0;i<rules->nrules;i++)
-   {
-    if(rules->rules[i].k && rules->rules[i].v)
-      {
-       for(j=0;j<tags->ntags;j++)
-          if(!strcmp(tags->k[j],rules->rules[i].k) && !strcmp(tags->v[j],rules->rules[i].v))
-             apply_actions(rules,&rules->rules[i],j,tags,result,id);
-      }
-    else if(rules->rules[i].k && !rules->rules[i].v)
-      {
-       for(j=0;j<tags->ntags;j++)
-          if(!strcmp(tags->k[j],rules->rules[i].k))
-             apply_actions(rules,&rules->rules[i],j,tags,result,id);
-      }
-    else if(!rules->rules[i].k && rules->rules[i].v)
-      {
-       for(j=0;j<tags->ntags;j++)
-          if(!strcmp(tags->v[j],rules->rules[i].v))
-             apply_actions(rules,&rules->rules[i],j,tags,result,id);
-      }
-    else /* if(!rules->rules[i].k && !rules->rules[i].v) */
-      {
-       for(j=0;j<tags->ntags;j++)
-          apply_actions(rules,&rules->rules[i],j,tags,result,id);
-      }
-   }
+ current_id=id;
+ current_list=&NodeRules;
+
+ ApplyRules(current_list,tags,result,NULL,NULL);
 
  return(result);
 }
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  Apply a set of actions to a matching tag.
+  Apply a set of tagging rules to a set of way tags.
 
-  TaggingRuleList *rules The tagging rules to apply.
+  TagList *ApplyWayTaggingRules Returns the list of output tags after modification.
 
-  TaggingRule *rule The rule that matched (containing the actions).
+  TagList *tags The tags to be modified.
 
-  int match The matching tag number.
+  int64_t id The ID of the way.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+TagList *ApplyWayTaggingRules(TagList *tags,int64_t id)
+{
+ TagList *result=NewTagList();
+
+ current_id=id;
+ current_list=&WayRules;
+
+ ApplyRules(current_list,tags,result,NULL,NULL);
+
+ return(result);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Apply a set of tagging rules to a set of relation tags.
+
+  TagList *ApplyRelationTaggingRules Returns the list of output tags after modification.
+
+  TagList *tags The tags to be modified.
+
+  int64_t id The ID of the relation.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+TagList *ApplyRelationTaggingRules(TagList *tags,int64_t id)
+{
+ TagList *result=NewTagList();
+
+ current_id=id;
+ current_list=&RelationRules;
+
+ ApplyRules(current_list,tags,result,NULL,NULL);
+
+ return(result);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Apply a set of rules to a matching tag.
+
+  TaggingRuleList *rules The rules that are to be matched.
 
   TagList *input The input tags.
 
   TagList *output The output tags.
 
-  node_t id The ID of the node, way or relation.
+  const char *match_k The key matched at the higher level rule.
+
+  const char *match_v The value matched at the higher level rule.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void apply_actions(TaggingRuleList *rules,TaggingRule *rule,int match,TagList *input,TagList *output,node_t id)
+static void ApplyRules(TaggingRuleList *rules,TagList *input,TagList *output,const char *match_k,const char *match_v)
 {
- int i;
+ int i,j;
+ char *match_k_copy=NULL,*match_v_copy=NULL;
  
- for(i=0;i<rule->nactions;i++)
+ if(match_k)
+    match_k_copy=strcpy(malloc(strlen(match_k)+1),match_k);
+
+ if(match_v)
+    match_v_copy=strcpy(malloc(strlen(match_v)+1),match_v);
+
+ for(i=0;i<rules->nrules;i++)
    {
-    char *k,*v;
+    const char *k,*v;
 
-    if(rule->actions[i].k)
-       k=rule->actions[i].k;
-    else
-       k=input->k[match];
+    k=rules->rules[i].k;
 
-    if(rule->actions[i].v)
-       v=rule->actions[i].v;
-    else
-       v=input->v[match];
+    if(!k && rules->rules[i].action >= TAGACTION_INHERIT)
+       k=match_k_copy;
 
-    if(rule->actions[i].action==TAGACTION_SET)
-       ModifyTag(input,k,v);
-    if(rule->actions[i].action==TAGACTION_UNSET)
-       DeleteTag(input,k);
-    if(rule->actions[i].action==TAGACTION_OUTPUT)
-       ModifyTag(output,k,v);
-    if(rule->actions[i].action==TAGACTION_LOGERROR)
+    v=rules->rules[i].v;
+
+    if(!v && rules->rules[i].action >= TAGACTION_INHERIT)
+       v=match_v_copy;
+
+    switch(rules->rules[i].action)
       {
-       if(rules==&NodeRules)
-          logerror("Node %"Pnode_t" has an unrecognised tag value '%s' = '%s' (in tagging rules).\n",id,k,v);
-       if(rules==&WayRules)
-          logerror("Way %"Pway_t" has an unrecognised tag value '%s' = '%s' (in tagging rules).\n",id,k,v);
-       if(rules==&RelationRules)
-          logerror("Relation %"Prelation_t" has an unrecognised tag value '%s' = '%s' (in tagging rules).\n",id,k,v);
+      case TAGACTION_IF:
+       if(k && v)
+         {
+          for(j=0;j<input->ntags;j++)
+             if(!strcmp(input->k[j],k) && !strcmp(input->v[j],v))
+                ApplyRules(rules->rules[i].rulelist,input,output,input->k[j],input->v[j]);
+         }
+       else if(k && !v)
+         {
+          for(j=0;j<input->ntags;j++)
+             if(!strcmp(input->k[j],k))
+                ApplyRules(rules->rules[i].rulelist,input,output,input->k[j],input->v[j]);
+         }
+       else if(!k && v)
+         {
+          for(j=0;j<input->ntags;j++)
+             if(!strcmp(input->v[j],v))
+                ApplyRules(rules->rules[i].rulelist,input,output,input->k[j],input->v[j]);
+         }
+       else /* if(!k && !v) */
+         {
+          if(!input->ntags)
+             ApplyRules(rules->rules[i].rulelist,input,output,"","");
+          else
+             for(j=0;j<input->ntags;j++)
+                ApplyRules(rules->rules[i].rulelist,input,output,input->k[j],input->v[j]);
+         }
+       break;
+
+      case TAGACTION_IFNOT:
+       if(k && v)
+         {
+          for(j=0;j<input->ntags;j++)
+             if(!strcmp(input->k[j],k) && !strcmp(input->v[j],v))
+                break;
+
+          if(j!=input->ntags)
+             break;
+         }
+       else if(k && !v)
+         {
+          for(j=0;j<input->ntags;j++)
+             if(!strcmp(input->k[j],k))
+                break;
+
+          if(j!=input->ntags)
+             break;
+         }
+       else if(!k && v)
+         {
+          for(j=0;j<input->ntags;j++)
+             if(!strcmp(input->v[j],v))
+                break;
+
+          if(j!=input->ntags)
+             break;
+         }
+       else /* if(!k && !v) */
+         {
+          break;
+         }
+
+       ApplyRules(rules->rules[i].rulelist,input,output,k,v);
+       break;
+
+      case TAGACTION_SET:
+       ModifyTag(input,k,v);
+       break;
+
+      case TAGACTION_UNSET:
+       DeleteTag(input,k);
+       break;
+
+      case TAGACTION_OUTPUT:
+       ModifyTag(output,k,v);
+       break;
+
+      case TAGACTION_LOGERROR:
+       if(rules->rules[i].k && !rules->rules[i].v)
+          for(j=0;j<input->ntags;j++)
+             if(!strcmp(input->k[j],rules->rules[i].k))
+               {
+                v=input->v[j];
+                break;
+               }
+
+       if(current_list==&NodeRules)
+          logerror("Node %"Pnode_t" has an unrecognised tag '%s' = '%s' (in tagging rules); %s.\n",logerror_node(current_id),k,v,rules->rules[i].message);
+       if(current_list==&WayRules)
+          logerror("Way %"Pway_t" has an unrecognised tag '%s' = '%s' (in tagging rules); %s.\n",logerror_way(current_id),k,v,rules->rules[i].message);
+       if(current_list==&RelationRules)
+          logerror("Relation %"Prelation_t" has an unrecognised tag '%s' = '%s' (in tagging rules); %s.\n",logerror_relation(current_id),k,v,rules->rules[i].message);
       }
    }
+
+ if(match_k_copy) free(match_k_copy);
+ if(match_v_copy) free(match_v_copy);
 }
