@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2014 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -29,6 +29,7 @@
 #include "segments.h"
 #include "ways.h"
 #include "relations.h"
+#include "errorlog.h"
 
 #include "visualiser.h"
 
@@ -63,10 +64,12 @@ static double LonMax;
 static int limit_type=0;
 static Highway highways=Highway_None;
 static Transports transports=Transports_None;
+static Properties properties=Properties_None;
 
 /* Local functions */
 
 static void find_all_nodes(Nodes *nodes,callback_t callback);
+
 static void output_junctions(index_t node,double latitude,double longitude);
 static void output_super(index_t node,double latitude,double longitude);
 static void output_oneway(index_t node,double latitude,double longitude);
@@ -75,6 +78,7 @@ static void output_transport(index_t node,double latitude,double longitude);
 static void output_barrier(index_t node,double latitude,double longitude);
 static void output_turnrestriction(index_t node,double latitude,double longitude);
 static void output_limits(index_t node,double latitude,double longitude);
+static void output_property(index_t node,double latitude,double longitude);
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -152,7 +156,7 @@ static void output_junctions(index_t node,double latitude,double longitude)
  while(segmentp);
 
  if(count!=2 || difference)
-    printf("%.6f %.6f %d\n",radians_to_degrees(latitude),radians_to_degrees(longitude),count);
+    printf("node%"Pindex_t" %.6f %.6f %d\n",node,radians_to_degrees(latitude),radians_to_degrees(longitude),count);
 }
 
 
@@ -214,7 +218,7 @@ static void output_super(index_t node,double latitude,double longitude)
  if(!IsSuperNode(nodep))
     return;
 
- printf("%.6f %.6f n\n",radians_to_degrees(latitude),radians_to_degrees(longitude));
+ printf("node%"Pindex_t" %.6f %.6f\n",node,radians_to_degrees(latitude),radians_to_degrees(longitude));
 
  segmentp=FirstSegment(OSMSegments,nodep,1);
 
@@ -225,10 +229,10 @@ static void output_super(index_t node,double latitude,double longitude)
        index_t othernode=OtherNode(segmentp,node);
        double lat,lon;
 
-       GetLatLong(OSMNodes,othernode,&lat,&lon);
+       GetLatLong(OSMNodes,othernode,NULL,&lat,&lon);
 
        if(node>othernode || (lat<LatMin || lat>LatMax || lon<LonMin || lon>LonMax))
-          printf("%.6f %.6f s\n",radians_to_degrees(lat),radians_to_degrees(lon));
+          printf("segment%"Pindex_t" %.6f %.6f\n",IndexSegment(OSMSegments,segmentp),radians_to_degrees(lat),radians_to_degrees(lon));
       }
 
     segmentp=NextSegment(OSMSegments,segmentp,node);
@@ -304,12 +308,12 @@ static void output_oneway(index_t node,double latitude,double longitude)
          {
           double lat,lon;
 
-          GetLatLong(OSMNodes,othernode,&lat,&lon);
+          GetLatLong(OSMNodes,othernode,NULL,&lat,&lon);
 
           if(IsOnewayFrom(segmentp,node))
-             printf("%.6f %.6f %.6f %.6f\n",radians_to_degrees(latitude),radians_to_degrees(longitude),radians_to_degrees(lat),radians_to_degrees(lon));
+             printf("segment%"Pindex_t" %.6f %.6f %.6f %.6f\n",IndexSegment(OSMSegments,segmentp),radians_to_degrees(latitude),radians_to_degrees(longitude),radians_to_degrees(lat),radians_to_degrees(lon));
           else if(IsOnewayFrom(segmentp,othernode))
-             printf("%.6f %.6f %.6f %.6f\n",radians_to_degrees(lat),radians_to_degrees(lon),radians_to_degrees(latitude),radians_to_degrees(longitude));
+             printf("segment%"Pindex_t" %.6f %.6f %.6f %.6f\n",IndexSegment(OSMSegments,segmentp),radians_to_degrees(lat),radians_to_degrees(lon),radians_to_degrees(latitude),radians_to_degrees(longitude));
          }
       }
 
@@ -394,9 +398,9 @@ static void output_highway(index_t node,double latitude,double longitude)
             {
              double lat,lon;
 
-             GetLatLong(OSMNodes,othernode,&lat,&lon);
+             GetLatLong(OSMNodes,othernode,NULL,&lat,&lon);
 
-             printf("%.6f %.6f %.6f %.6f\n",radians_to_degrees(latitude),radians_to_degrees(longitude),radians_to_degrees(lat),radians_to_degrees(lon));
+             printf("segment%"Pindex_t" %.6f %.6f %.6f %.6f\n",IndexSegment(OSMSegments,segmentp),radians_to_degrees(latitude),radians_to_degrees(longitude),radians_to_degrees(lat),radians_to_degrees(lon));
             }
          }
       }
@@ -408,7 +412,7 @@ static void output_highway(index_t node,double latitude,double longitude)
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  Output the data for segments allowed for a paticular type of traffic.
+  Output the data for segments allowed for a particular type of traffic.
 
   Nodes *nodes The set of nodes to use.
 
@@ -482,9 +486,9 @@ static void output_transport(index_t node,double latitude,double longitude)
             {
              double lat,lon;
 
-             GetLatLong(OSMNodes,othernode,&lat,&lon);
+             GetLatLong(OSMNodes,othernode,NULL,&lat,&lon);
 
-             printf("%.6f %.6f %.6f %.6f\n",radians_to_degrees(latitude),radians_to_degrees(longitude),radians_to_degrees(lat),radians_to_degrees(lon));
+             printf("segment%"Pindex_t" %.6f %.6f %.6f %.6f\n",IndexSegment(OSMSegments,segmentp),radians_to_degrees(latitude),radians_to_degrees(longitude),radians_to_degrees(lat),radians_to_degrees(lon));
             }
          }
       }
@@ -496,7 +500,7 @@ static void output_transport(index_t node,double latitude,double longitude)
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  Output the data for nodes disallowed for a paticular type of traffic.
+  Output the data for nodes disallowed for a particular type of traffic.
 
   Nodes *nodes The set of nodes to use.
 
@@ -554,7 +558,7 @@ static void output_barrier(index_t node,double latitude,double longitude)
  Node *nodep=LookupNode(OSMNodes,node,1);
 
  if(!(nodep->allow&transports))
-    printf("%.6f %.6f\n",radians_to_degrees(latitude),radians_to_degrees(longitude));
+    printf("node%"Pindex_t" %.6f %.6f\n",node,radians_to_degrees(latitude),radians_to_degrees(longitude));
 }
 
 
@@ -631,14 +635,16 @@ static void output_turnrestriction(index_t node,double latitude,double longitude
     to_segmentp  =LookupSegment(OSMSegments,relation->to  ,2);
 
     from_node=OtherNode(from_segmentp,node);
-    to_node=OtherNode(to_segmentp,node);
+    to_node  =OtherNode(to_segmentp  ,node);
 
-    GetLatLong(OSMNodes,from_node,&from_lat,&from_lon);
-    GetLatLong(OSMNodes,to_node,&to_lat,&to_lon);
+    GetLatLong(OSMNodes,from_node,NULL,&from_lat,&from_lon);
+    GetLatLong(OSMNodes,to_node  ,NULL,&to_lat  ,&to_lon);
 
-    printf("%.6f %.6f %.6f %.6f %.6f %.6f\n",radians_to_degrees(from_lat),radians_to_degrees(from_lon),
-                                             radians_to_degrees(latitude),radians_to_degrees(longitude),
-                                             radians_to_degrees(to_lat),radians_to_degrees(to_lon));
+    printf("turn-relation%"Pindex_t" %.6f %.6f %.6f %.6f %.6f %.6f\n",
+           turnrelation,
+           radians_to_degrees(from_lat),radians_to_degrees(from_lon),
+           radians_to_degrees(latitude),radians_to_degrees(longitude),
+           radians_to_degrees(to_lat),radians_to_degrees(to_lon));
 
     turnrelation=FindNextTurnRelation1(OSMRelations,turnrelation);
    }
@@ -870,6 +876,7 @@ static void output_limits(index_t node,double latitude,double longitude)
 {
  Node *nodep=LookupNode(OSMNodes,node,1);
  Segment *segmentp,segmentps[MAX_SEG_PER_NODE];
+ index_t segments[MAX_SEG_PER_NODE];
  int limits[MAX_SEG_PER_NODE];
  int count=0;
  int i,j,same=0;
@@ -883,6 +890,7 @@ static void output_limits(index_t node,double latitude,double longitude)
        Way *wayp=LookupWay(OSMWays,segmentp->way,1);
 
        segmentps[count]=*segmentp;
+       segments [count]=IndexSegment(OSMSegments,segmentp);
 
        switch(limit_type)
          {
@@ -918,7 +926,7 @@ static void output_limits(index_t node,double latitude,double longitude)
 
  /* Display the interesting limits */
 
- printf("%.6f %.6f\n",radians_to_degrees(latitude),radians_to_degrees(longitude));
+ printf("node%"Pindex_t" %.6f %.6f\n",node,radians_to_degrees(latitude),radians_to_degrees(longitude));
 
  for(i=0;i<count;i++)
    {
@@ -931,24 +939,24 @@ static void output_limits(index_t node,double latitude,double longitude)
       {
        double lat,lon;
 
-       GetLatLong(OSMNodes,OtherNode(&segmentps[i],node),&lat,&lon);
+       GetLatLong(OSMNodes,OtherNode(&segmentps[i],node),NULL,&lat,&lon);
 
        switch(limit_type)
          {
          case SPEED_LIMIT:
-          printf("%.6f %.6f %d\n",radians_to_degrees(lat),radians_to_degrees(lon),speed_to_kph(limits[i]));
+          printf("segment%"Pindex_t" %.6f %.6f %d\n",segments[i],radians_to_degrees(lat),radians_to_degrees(lon),speed_to_kph(limits[i]));
           break;
          case WEIGHT_LIMIT:
-          printf("%.6f %.6f %.1f\n",radians_to_degrees(lat),radians_to_degrees(lon),weight_to_tonnes(limits[i]));
+          printf("segment%"Pindex_t" %.6f %.6f %.1f\n",segments[i],radians_to_degrees(lat),radians_to_degrees(lon),weight_to_tonnes(limits[i]));
           break;
          case HEIGHT_LIMIT:
-          printf("%.6f %.6f %.1f\n",radians_to_degrees(lat),radians_to_degrees(lon),height_to_metres(limits[i]));
+          printf("segment%"Pindex_t" %.6f %.6f %.1f\n",segments[i],radians_to_degrees(lat),radians_to_degrees(lon),height_to_metres(limits[i]));
           break;
          case WIDTH_LIMIT:
-          printf("%.6f %.6f %.1f\n",radians_to_degrees(lat),radians_to_degrees(lon),width_to_metres(limits[i]));
+          printf("segment%"Pindex_t" %.6f %.6f %.1f\n",segments[i],radians_to_degrees(lat),radians_to_degrees(lon),width_to_metres(limits[i]));
           break;
          case LENGTH_LIMIT:
-          printf("%.6f %.6f %.1f\n",radians_to_degrees(lat),radians_to_degrees(lon),length_to_metres(limits[i]));
+          printf("segment%"Pindex_t" %.6f %.6f %.1f\n",segments[i],radians_to_degrees(lat),radians_to_degrees(lon),length_to_metres(limits[i]));
           break;
          }
       }
@@ -956,6 +964,94 @@ static void output_limits(index_t node,double latitude,double longitude)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Output the data for segments that have a particular property.
+
+  Nodes *nodes The set of nodes to use.
+
+  Segments *segments The set of segments to use.
+
+  Ways *ways The set of ways to use.
+
+  Relations *relations The set of relations to use.
+
+  double latmin The minimum latitude.
+
+  double latmax The maximum latitude.
+
+  double lonmin The minimum longitude.
+
+  double lonmax The maximum longitude.
+
+  Property property The type of property.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void OutputProperty(Nodes *nodes,Segments *segments,Ways *ways,Relations *relations,double latmin,double latmax,double lonmin,double lonmax,Property property)
+{
+ /* Use local variables so that the callback doesn't need to pass them backwards and forwards */
+
+ OSMNodes=nodes;
+ OSMSegments=segments;
+ OSMWays=ways;
+ OSMRelations=relations;
+
+ LatMin=latmin;
+ LatMax=latmax;
+ LonMin=lonmin;
+ LonMax=lonmax;
+
+ /* Iterate through the nodes and process them */
+
+ properties=PROPERTIES(property);
+
+ find_all_nodes(nodes,(callback_t)output_property);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Process a single node and all connected one-way segments (called as a callback).
+
+  index_t node The node to output.
+
+  double latitude The latitude of the node.
+
+  double longitude The longitude of the node.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void output_property(index_t node,double latitude,double longitude)
+{
+ Node *nodep=LookupNode(OSMNodes,node,1);
+ Segment *segmentp;
+
+ segmentp=FirstSegment(OSMSegments,nodep,1);
+
+ do
+   {
+    if(IsNormalSegment(segmentp))
+      {
+       index_t othernode=OtherNode(segmentp,node);
+
+       if(node>othernode)
+         {
+          Way *wayp=LookupWay(OSMWays,segmentp->way,1);
+
+          if(wayp->props&properties)
+            {
+             double lat,lon;
+
+             GetLatLong(OSMNodes,othernode,NULL,&lat,&lon);
+
+             printf("segment%"Pindex_t" %.6f %.6f %.6f %.6f\n",IndexSegment(OSMSegments,segmentp),radians_to_degrees(latitude),radians_to_degrees(longitude),radians_to_degrees(lat),radians_to_degrees(lon));
+            }
+         }
+      }
+
+    segmentp=NextSegment(OSMSegments,segmentp,node);
+   }
+ while(segmentp);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   A function to iterate through all nodes and call a callback function for each one.
 
@@ -975,6 +1071,12 @@ static void find_all_nodes(Nodes *nodes,callback_t callback)
 
  /* Loop through all of the nodes. */
 
+ if(latminbin<0)                   latminbin=0;
+ if(latmaxbin>nodes->file.latbins) latmaxbin=nodes->file.latbins-1;
+
+ if(lonminbin<0)                   lonminbin=0;
+ if(lonmaxbin>nodes->file.lonbins) lonmaxbin=nodes->file.lonbins-1;
+
  for(latb=latminbin;latb<=latmaxbin;latb++)
     for(lonb=lonminbin;lonb<=lonmaxbin;lonb++)
       {
@@ -998,3 +1100,56 @@ static void find_all_nodes(Nodes *nodes,callback_t callback)
          }
       }
 }
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Output the data for error logs within the region.
+
+  ErrorLogs *errorlogs The set of error logs to use.
+
+  double latmin The minimum latitude.
+
+  double latmax The maximum latitude.
+
+  double lonmin The minimum longitude.
+
+  double lonmax The maximum longitude.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void OutputErrorLog(ErrorLogs *errorlogs,double latmin,double latmax,double lonmin,double lonmax)
+{
+ ll_bin_t latminbin=latlong_to_bin(radians_to_latlong(latmin))-errorlogs->file.latzero;
+ ll_bin_t latmaxbin=latlong_to_bin(radians_to_latlong(latmax))-errorlogs->file.latzero;
+ ll_bin_t lonminbin=latlong_to_bin(radians_to_latlong(lonmin))-errorlogs->file.lonzero;
+ ll_bin_t lonmaxbin=latlong_to_bin(radians_to_latlong(lonmax))-errorlogs->file.lonzero;
+ ll_bin_t latb,lonb;
+ index_t i,index1,index2;
+
+ /* Loop through all of the error logs. */
+
+ for(latb=latminbin;latb<=latmaxbin;latb++)
+    for(lonb=lonminbin;lonb<=lonmaxbin;lonb++)
+      {
+       ll_bin2_t llbin=lonb*errorlogs->file.latbins+latb;
+
+       if(llbin<0 || llbin>(errorlogs->file.latbins*errorlogs->file.lonbins))
+          continue;
+
+       index1=LookupErrorLogOffset(errorlogs,llbin);
+       index2=LookupErrorLogOffset(errorlogs,llbin+1);
+
+       if(index2>errorlogs->file.number_geo)
+          index2=errorlogs->file.number_geo;
+
+       for(i=index1;i<index2;i++)
+         {
+          ErrorLog *errorlogp=LookupErrorLog(errorlogs,i,1);
+
+          double lat=latlong_to_radians(bin_to_latlong(errorlogs->file.latzero+latb)+off_to_latlong(errorlogp->latoffset));
+          double lon=latlong_to_radians(bin_to_latlong(errorlogs->file.lonzero+lonb)+off_to_latlong(errorlogp->lonoffset));
+
+          if(lat>latmin && lat<latmax && lon>lonmin && lon<lonmax)
+             printf("errorlog%"Pindex_t" %.6f %.6f\n",i,radians_to_degrees(lat),radians_to_degrees(lon));
+         }
+      }
+}
