@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -28,6 +28,7 @@
 
 #include "types.h"
 
+#include "cache.h"
 #include "files.h"
 
 
@@ -84,10 +85,11 @@ struct _Ways
  off_t      namesoffset;        /*+ The offset of the names within the file. +*/
 
  Way        cached[3];          /*+ Two cached nodes read from the file in slim mode. +*/
- index_t    incache[3];         /*+ The indexes of the cached ways. +*/
 
  char      *ncached[3];         /*+ The cached way name. +*/
 
+ WayCache  *cache;              /*+ A RAM cache of ways read from the file. +*/
+
 #endif
 };
 
@@ -96,6 +98,8 @@ struct _Ways
 
 Ways *LoadWayList(const char *filename);
 
+void DestroyWayList(Ways *ways);
+
 int WaysCompare(Way *way1p,Way *way2p);
 
 
@@ -111,9 +115,23 @@ int WaysCompare(Way *way1p,Way *way2p);
 
 #else
 
-static Way *LookupWay(Ways *ways,index_t index,int position);
+static inline Way *LookupWay(Ways *ways,index_t index,int position);
 
-static char *WayName(Ways *ways,Way *wayp);
+static inline char *WayName(Ways *ways,Way *wayp);
+
+CACHE_NEWCACHE_PROTO(Way)
+CACHE_DELETECACHE_PROTO(Way)
+CACHE_FETCHCACHE_PROTO(Way)
+CACHE_INVALIDATECACHE_PROTO(Way)
+
+
+/* Inline functions */
+
+CACHE_STRUCTURE(Way)
+CACHE_NEWCACHE(Way)
+CACHE_DELETECACHE(Way)
+CACHE_FETCHCACHE(Way)
+CACHE_INVALIDATECACHE(Way)
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -130,12 +148,7 @@ static char *WayName(Ways *ways,Way *wayp);
 
 static inline Way *LookupWay(Ways *ways,index_t index,int position)
 {
- if(ways->incache[position-1]!=index)
-   {
-    SeekReadFile(ways->fd,&ways->cached[position-1],sizeof(Way),sizeof(WaysFile)+(off_t)index*sizeof(Way));
-
-    ways->incache[position-1]=index;
-   }
+ ways->cached[position-1]=*FetchCachedWay(ways->cache,index,ways->fd,sizeof(WaysFile));
 
  return(&ways->cached[position-1]);
 }
@@ -154,29 +167,22 @@ static inline Way *LookupWay(Ways *ways,index_t index,int position)
 static inline char *WayName(Ways *ways,Way *wayp)
 {
  int position=wayp-&ways->cached[-1];
-
  int n=0;
 
- SeekFile(ways->fd,ways->namesoffset+wayp->name);
-
  if(!ways->ncached[position-1])
-    ways->ncached[position-1]=(char*)malloc(32);
+    ways->ncached[position-1]=(char*)malloc(64);
 
- while(1)
+ while(!SlimFetch(ways->fd,ways->ncached[position-1]+n,64,ways->namesoffset+wayp->name+n))
    {
     int i;
-    int m=ReadFile(ways->fd,ways->ncached[position-1]+n,32);
 
-    if(m<0)
-       break;
-    
-    for(i=n;i<n+32;i++)
+    for(i=n;i<n+64;i++)
        if(ways->ncached[position-1][i]==0)
           goto exitloop;
 
-    n+=32;
+    n+=64;
 
-    ways->ncached[position-1]=(char*)realloc((void*)ways->ncached[position-1],n+32);
+    ways->ncached[position-1]=(char*)realloc((void*)ways->ncached[position-1],n+64);
    }
 
  exitloop:
