@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -27,6 +27,7 @@
 #include "ways.h"
 
 #include "typesx.h"
+#include "nodesx.h"
 #include "segmentsx.h"
 #include "waysx.h"
 
@@ -49,10 +50,9 @@ static SegmentsX *sortsegmentsx;
 /* Local functions */
 
 static int sort_by_id(WayX *a,WayX *b);
-static int deduplicate_by_id(WayX *wayx,index_t index);
+static int deduplicate_and_index_by_id(WayX *wayx,index_t index);
 
-static int sort_by_name(WayX *a,WayX *b);
-static int index_by_id(WayX *wayx,index_t index);
+static int sort_by_name(char *a,char *b);
 
 static int delete_unused(WayX *wayx,index_t index);
 static int sort_by_name_and_prop_and_id(WayX *a,WayX *b);
@@ -86,35 +86,34 @@ WaysX *NewWayList(int append,int readonly)
  if(append || readonly)
     if(ExistsFile(waysx->filename))
       {
-       off_t size,position=0;
+       FILESORT_VARINT waysize;
        int fd;
 
-       size=SizeFile(waysx->filename);
+       fd=ReOpenFileBuffered(waysx->filename);
 
-       fd=ReOpenFile(waysx->filename);
-
-       while(position<size)
+       while(!ReadFileBuffered(fd,&waysize,FILESORT_VARSIZE))
          {
-          FILESORT_VARINT waysize;
-
-          SeekReadFile(fd,&waysize,FILESORT_VARSIZE,position);
+          SkipFileBuffered(fd,waysize);
 
           waysx->number++;
-          position+=waysize+FILESORT_VARSIZE;
          }
 
-       CloseFile(fd);
+       CloseFileBuffered(fd);
 
        RenameFile(waysx->filename,waysx->filename_tmp);
       }
 
  if(append)
-    waysx->fd=OpenFileAppend(waysx->filename_tmp);
+    waysx->fd=OpenFileBufferedAppend(waysx->filename_tmp);
  else if(!readonly)
-    waysx->fd=OpenFileNew(waysx->filename_tmp);
+    waysx->fd=OpenFileBufferedNew(waysx->filename_tmp);
  else
     waysx->fd=-1;
 
+#if SLIM
+ waysx->cache=NewWayXCache();
+#endif
+
 
  waysx->nfilename_tmp=(char*)malloc(strlen(option_tmpdirname)+32);
 
@@ -145,6 +144,9 @@ void FreeWayList(WaysX *waysx,int keep)
  if(waysx->idata)
     free(waysx->idata);
 
+ if(waysx->odata)
+    free(waysx->odata);
+
  if(waysx->cdata)
     free(waysx->cdata);
 
@@ -152,6 +154,10 @@ void FreeWayList(WaysX *waysx,int keep)
 
  free(waysx->nfilename_tmp);
 
+#if SLIM
+ DeleteWayXCache(waysx->cache);
+#endif
+
  free(waysx);
 }
 
@@ -165,22 +171,31 @@ void FreeWayList(WaysX *waysx,int keep)
 
   Way *way The way data itself.
 
+  node_t *nodes The list of nodes for this way.
+
+  int nnodes The number of nodes for this way.
+
   const char *name The name or reference of the way.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void AppendWayList(WaysX *waysx,way_t id,Way *way,const char *name)
+void AppendWayList(WaysX *waysx,way_t id,Way *way,node_t *nodes,int nnodes,const char *name)
 {
  WayX wayx;
  FILESORT_VARINT size;
+ node_t nonode=NO_NODE_ID;
 
  wayx.id=id;
  wayx.way=*way;
 
- size=sizeof(WayX)+strlen(name)+1;
+ size=sizeof(WayX)+(nnodes+1)*sizeof(node_t)+strlen(name)+1;
 
- WriteFile(waysx->fd,&size,FILESORT_VARSIZE);
- WriteFile(waysx->fd,&wayx,sizeof(WayX));
- WriteFile(waysx->fd,name,strlen(name)+1);
+ WriteFileBuffered(waysx->fd,&size,FILESORT_VARSIZE);
+ WriteFileBuffered(waysx->fd,&wayx,sizeof(WayX));
+
+ WriteFileBuffered(waysx->fd,nodes  ,nnodes*sizeof(node_t));
+ WriteFileBuffered(waysx->fd,&nonode,       sizeof(node_t));
+
+ WriteFileBuffered(waysx->fd,name,strlen(name)+1);
 
  waysx->number++;
 
@@ -197,7 +212,7 @@ void AppendWayList(WaysX *waysx,way_t id,Way *way,const char *name)
 void FinishWayList(WaysX *waysx)
 {
  if(waysx->fd!=-1)
-    waysx->fd=CloseFile(waysx->fd);
+    waysx->fd=CloseFileBuffered(waysx->fd);
 }
 
 
@@ -277,24 +292,34 @@ void SortWayList(WaysX *waysx)
 
  /* Re-open the file read-only and a new file writeable */
 
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+ waysx->fd=ReOpenFileBuffered(waysx->filename_tmp);
 
  DeleteFile(waysx->filename_tmp);
 
- fd=OpenFileNew(waysx->filename_tmp);
+ fd=OpenFileBufferedNew(waysx->filename_tmp);
+
+ /* Allocate the array of indexes */
+
+ waysx->idata=(way_t*)malloc(waysx->number*sizeof(way_t));
+
+ logassert(waysx->idata,"Failed to allocate memory (try using slim mode?)"); /* Check malloc() worked */
 
  /* Sort the ways by ID and index them */
 
+ sortwaysx=waysx;
+
  xnumber=waysx->number;
 
  waysx->number=filesort_vary(waysx->fd,fd,NULL,
                                           (int (*)(const void*,const void*))sort_by_id,
-                                          (int (*)(void*,index_t))deduplicate_by_id);
+                                          (int (*)(void*,index_t))deduplicate_and_index_by_id);
+
+ waysx->knumber=waysx->number;
 
  /* Close the files */
 
- waysx->fd=CloseFile(waysx->fd);
- CloseFile(fd);
+ waysx->fd=CloseFileBuffered(waysx->fd);
+ CloseFileBuffered(fd);
 
  /* Print the final message */
 
@@ -327,16 +352,16 @@ static int sort_by_id(WayX *a,WayX *b)
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  Discard duplicate ways.
+  Discard duplicate ways and create and index of ids.
 
-  int deduplicate_by_id Return 1 if the value is to be kept, otherwise 0.
+  int deduplicate_and_index_by_id Return 1 if the value is to be kept, otherwise 0.
 
   WayX *wayx The extended way.
 
   index_t index The number of sorted ways that have already been written to the output file.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static int deduplicate_by_id(WayX *wayx,index_t index)
+static int deduplicate_and_index_by_id(WayX *wayx,index_t index)
 {
  static way_t previd=NO_WAY_ID;
 
@@ -347,7 +372,11 @@ static int deduplicate_by_id(WayX *wayx,index_t index)
     if(wayx->way.type==WAY_DELETED)
        return(0);
     else
+      {
+       sortwaysx->idata[index]=wayx->id;
+
        return(1);
+      }
    }
  else
     return(0);
@@ -355,17 +384,141 @@ static int deduplicate_by_id(WayX *wayx,index_t index)
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  Extract the way names from the ways and reference the list of names from the ways.
+  Split the ways into segments and way names.
+
+  SegmentsX *SplitWays Returns the set of segments that have been created.
 
   WaysX *waysx The set of ways to process.
 
+  NodesX *nodesx The set of nodes to use.
+
   int keep If set to 1 then keep the old data file otherwise delete it.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void ExtractWayNames(WaysX *waysx,int keep)
+SegmentsX *SplitWays(WaysX *waysx,NodesX *nodesx,int keep)
+{
+ SegmentsX *segmentsx;
+ index_t i;
+ int fd,nfd;
+ char *name=NULL;
+ int namelen=0;
+
+ /* Print the start message */
+
+ printf_first("Splitting Ways: Ways=0 Segments=0");
+
+ segmentsx=NewSegmentList();
+
+ /* Re-open the file read-only and a new file writeable */
+
+ waysx->fd=ReOpenFileBuffered(waysx->filename_tmp);
+
+ if(keep)
+    RenameFile(waysx->filename_tmp,waysx->filename);
+ else
+    DeleteFile(waysx->filename_tmp);
+
+ fd=OpenFileBufferedNew(waysx->filename_tmp);
+
+ nfd=OpenFileBufferedNew(waysx->nfilename_tmp);
+
+ /* Loop through the ways and create the segments and way names */
+
+ for(i=0;i<waysx->number;i++)
+   {
+    WayX wayx;
+    FILESORT_VARINT size;
+    node_t node,prevnode=NO_NODE_ID;
+    index_t index,previndex=NO_NODE;
+
+    ReadFileBuffered(waysx->fd,&size,FILESORT_VARSIZE);
+
+    ReadFileBuffered(waysx->fd,&wayx,sizeof(WayX));
+
+    waysx->allow|=wayx.way.allow;
+
+    while(!ReadFileBuffered(waysx->fd,&node,sizeof(node_t)) && node!=NO_NODE_ID)
+      {
+       index=IndexNodeX(nodesx,node);
+
+       if(prevnode==node)
+         {
+          logerror("Way %"Pway_t" contains node %"Pnode_t" that is connected to itself.\n",logerror_way(wayx.id),logerror_node(node));
+         }
+       else if(index==NO_NODE)
+         {
+          logerror("Way %"Pway_t" contains node %"Pnode_t" that does not exist in the Routino database.\n",logerror_way(wayx.id),logerror_node(node));
+         }
+       else if(previndex==NO_NODE)
+          ;
+       else
+         {
+          distance_t segment_flags=0;
+
+          if(wayx.way.type&Highway_OneWay)
+             segment_flags|=ONEWAY_1TO2;
+
+          if(wayx.way.type&Highway_Area)
+             segment_flags|=SEGMENT_AREA;
+
+          AppendSegmentList(segmentsx,i,previndex,index,segment_flags);
+         }
+
+       prevnode=node;
+       previndex=index;
+
+       size-=sizeof(node_t);
+      }
+
+    size-=sizeof(node_t)+sizeof(WayX);
+
+    if(namelen<size)
+       name=(char*)realloc((void*)name,namelen=size);
+
+    ReadFileBuffered(waysx->fd,name,size);
+
+    WriteFileBuffered(fd,&wayx,sizeof(WayX));
+
+    size+=sizeof(index_t);
+
+    WriteFileBuffered(nfd,&size,FILESORT_VARSIZE);
+    WriteFileBuffered(nfd,&i,sizeof(index_t));
+    WriteFileBuffered(nfd,name,size-sizeof(index_t));
+
+    if(!((i+1)%1000))
+       printf_middle("Splitting Ways: Ways=%"Pindex_t" Segments=%"Pindex_t,i+1,segmentsx->number);
+   }
+
+ FinishSegmentList(segmentsx);
+
+ if(name) free(name);
+
+ /* Close the files */
+
+ waysx->fd=CloseFileBuffered(waysx->fd);
+ CloseFileBuffered(fd);
+
+ CloseFileBuffered(nfd);
+
+ /* Print the final message */
+
+ printf_last("Split Ways: Ways=%"Pindex_t" Segments=%"Pindex_t,waysx->number,segmentsx->number);
+
+ return(segmentsx);
+}
+
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Sort the way names and assign the offsets to the ways.
+
+  WaysX *waysx The set of ways to process.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void SortWayNames(WaysX *waysx)
 {
  index_t i;
- int fd;
+ int nfd;
  char *names[2]={NULL,NULL};
  int namelen[2]={0,0};
  int nnames=0;
@@ -373,80 +526,88 @@ void ExtractWayNames(WaysX *waysx,int keep)
 
  /* Print the start message */
 
- printf_first("Sorting Ways by Name");
+ printf_first("Sorting Way Names");
 
- /* Re-open the file read-only and a new file writeable */
+ /* Re-open the file read-only and new file writeable */
 
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+ waysx->nfd=ReOpenFileBuffered(waysx->nfilename_tmp);
 
- if(keep)
-    RenameFile(waysx->filename_tmp,waysx->filename);
- else
-    DeleteFile(waysx->filename_tmp);
+ DeleteFile(waysx->nfilename_tmp);
 
- fd=OpenFileNew(waysx->filename_tmp);
+ nfd=OpenFileBufferedNew(waysx->nfilename_tmp);
 
- /* Sort the ways to allow separating the names */
+ /* Sort the way names */
 
- filesort_vary(waysx->fd,fd,NULL,
-                            (int (*)(const void*,const void*))sort_by_name,
-                            NULL);
+ waysx->nlength=0;
+
+ filesort_vary(waysx->nfd,nfd,NULL,
+                              (int (*)(const void*,const void*))sort_by_name,
+                              NULL);
 
  /* Close the files */
 
- waysx->fd=CloseFile(waysx->fd);
- CloseFile(fd);
+ waysx->nfd=CloseFileBuffered(waysx->nfd);
+ CloseFileBuffered(nfd);
 
  /* Print the final message */
 
- printf_last("Sorted Ways by Name: Ways=%"Pindex_t,waysx->number);
+ printf_last("Sorted Way Names: Ways=%"Pindex_t,waysx->number);
 
 
  /* Print the start message */
 
- printf_first("Separating Way Names: Ways=0 Names=0");
+ printf_first("Updating Ways with Names: Ways=0 Names=0");
 
- /* Re-open the file read-only and new files writeable */
+ /* Map into memory /  open the file */
 
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+#if !SLIM
+ waysx->data=MapFileWriteable(waysx->filename_tmp);
+#else
+ waysx->fd=SlimMapFileWriteable(waysx->filename_tmp);
+#endif
 
- DeleteFile(waysx->filename_tmp);
+ /* Re-open the file read-only and new file writeable */
 
- fd=OpenFileNew(waysx->filename_tmp);
+ waysx->nfd=ReOpenFileBuffered(waysx->nfilename_tmp);
 
- waysx->nfd=OpenFileNew(waysx->nfilename_tmp);
+ DeleteFile(waysx->nfilename_tmp);
 
- /* Copy from the single file into two files */
+ nfd=OpenFileBufferedNew(waysx->nfilename_tmp);
+
+ /* Update the ways and de-duplicate the names */
 
  for(i=0;i<waysx->number;i++)
    {
-    WayX wayx;
+    WayX *wayx;
+    index_t index;
     FILESORT_VARINT size;
 
-    ReadFile(waysx->fd,&size,FILESORT_VARSIZE);
+    ReadFileBuffered(waysx->nfd,&size,FILESORT_VARSIZE);
 
     if(namelen[nnames%2]<size)
        names[nnames%2]=(char*)realloc((void*)names[nnames%2],namelen[nnames%2]=size);
 
-    ReadFile(waysx->fd,&wayx,sizeof(WayX));
-    ReadFile(waysx->fd,names[nnames%2],size-sizeof(WayX));
+    ReadFileBuffered(waysx->nfd,&index,sizeof(index_t));
+    ReadFileBuffered(waysx->nfd,names[nnames%2],size-sizeof(index_t));
 
     if(nnames==0 || strcmp(names[0],names[1]))
       {
-       WriteFile(waysx->nfd,names[nnames%2],size-sizeof(WayX));
+       WriteFileBuffered(nfd,names[nnames%2],size-sizeof(index_t));
 
        lastlength=waysx->nlength;
-       waysx->nlength+=size-sizeof(WayX);
+       waysx->nlength+=size-sizeof(index_t);
 
        nnames++;
       }
 
-    wayx.way.name=lastlength;
+    wayx=LookupWayX(waysx,index,1);
 
-    WriteFile(fd,&wayx,sizeof(WayX));
+    wayx->way.name=lastlength;
+
+    PutBackWayX(waysx,wayx);
 
     if(!((i+1)%1000))
-       printf_middle("Separating Way Names: Ways=%"Pindex_t" Names=%"Pindex_t,i+1,nnames);
+       printf_middle("Updating Ways with Names: Ways=%"Pindex_t" Names=%"Pindex_t,i+1,nnames);
    }
 
  if(names[0]) free(names[0]);
@@ -454,68 +615,38 @@ void ExtractWayNames(WaysX *waysx,int keep)
 
  /* Close the files */
 
- waysx->fd=CloseFile(waysx->fd);
- CloseFile(fd);
+ waysx->nfd=CloseFileBuffered(waysx->nfd);
+ CloseFileBuffered(nfd);
 
- waysx->nfd=CloseFile(waysx->nfd);
+ /* Unmap from memory / close the files */
+
+#if !SLIM
+ waysx->data=UnmapFile(waysx->data);
+#else
+ waysx->fd=SlimUnmapFile(waysx->fd);
+#endif
 
  /* Print the final message */
 
- printf_last("Separated Way Names: Ways=%"Pindex_t" Names=%"Pindex_t,waysx->number,nnames);
-
-
- /* Print the start message */
-
- printf_first("Sorting Ways");
-
- /* Re-open the file read-only and a new file writeable */
-
- waysx->fd=ReOpenFile(waysx->filename_tmp);
-
- DeleteFile(waysx->filename_tmp);
-
- fd=OpenFileNew(waysx->filename_tmp);
-
- /* Allocate the array of indexes */
-
- waysx->idata=(way_t*)malloc(waysx->number*sizeof(way_t));
-
- logassert(waysx->idata,"Failed to allocate memory (try using slim mode?)"); /* Check malloc() worked */
-
- /* Sort the ways by ID */
-
- sortwaysx=waysx;
-
- filesort_fixed(waysx->fd,fd,sizeof(WayX),NULL,
-                                          (int (*)(const void*,const void*))sort_by_id,
-                                          (int (*)(void*,index_t))index_by_id);
-
- /* Close the files */
-
- waysx->fd=CloseFile(waysx->fd);
- CloseFile(fd);
-
- /* Print the final message */
-
- printf_last("Sorted Ways: Ways=%"Pindex_t,waysx->number);
+ printf_last("Updated Ways with Names: Ways=%"Pindex_t" Names=%"Pindex_t,waysx->number,nnames);
 }
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  Sort the ways into name order and then id order.
+  Sort the ways into name order.
 
   int sort_by_name Returns the comparison of the name fields.
 
-  WayX *a The first extended Way.
+  char *a The first way name.
 
-  WayX *b The second extended Way.
+  char *b The second way name.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static int sort_by_name(WayX *a,WayX *b)
+static int sort_by_name(char *a,char *b)
 {
  int compare;
- char *a_name=(char*)a+sizeof(WayX);
- char *b_name=(char*)b+sizeof(WayX);
+ char *a_name=a+sizeof(index_t);
+ char *b_name=b+sizeof(index_t);
 
  compare=strcmp(a_name,b_name);
 
@@ -526,24 +657,6 @@ static int sort_by_name(WayX *a,WayX *b)
 }
 
 
-/*++++++++++++++++++++++++++++++++++++++
-  Create the index of identifiers.
-
-  int index_by_id Return 1 if the value is to be kept, otherwise 0.
-
-  WayX *wayx The extended way.
-
-  index_t index The number of sorted ways that have already been written to the output file.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static int index_by_id(WayX *wayx,index_t index)
-{
- sortwaysx->idata[index]=wayx->id;
-
- return(1);
-}
-
-
 /*++++++++++++++++++++++++++++++++++++++
   Compact the way list, removing duplicated ways and unused ways.
 
@@ -572,11 +685,11 @@ void CompactWayList(WaysX *waysx,SegmentsX *segmentsx)
 
  /* Re-open the file read-only and a new file writeable */
 
- waysx->fd=ReOpenFile(waysx->filename_tmp);
+ waysx->fd=ReOpenFileBuffered(waysx->filename_tmp);
 
  DeleteFile(waysx->filename_tmp);
 
- fd=OpenFileNew(waysx->filename_tmp);
+ fd=OpenFileBufferedNew(waysx->filename_tmp);
 
  /* Sort the ways to allow compacting according to the properties */
 
@@ -589,16 +702,18 @@ void CompactWayList(WaysX *waysx,SegmentsX *segmentsx)
 
  /* Close the files */
 
- waysx->fd=CloseFile(waysx->fd);
- CloseFile(fd);
+ waysx->fd=CloseFileBuffered(waysx->fd);
+ CloseFileBuffered(fd);
+
+ /* Free the data */
+
+ free(segmentsx->usedway);
+ segmentsx->usedway=NULL;
 
  /* Print the final message */
 
  printf_last("Sorted and Compacted Ways: Ways=%"Pindex_t" Unique=%"Pindex_t,waysx->number,cnumber);
  waysx->number=cnumber;
-
- free(segmentsx->usedway);
- segmentsx->usedway=NULL;
 }
 
 
@@ -704,7 +819,7 @@ void SaveWayList(WaysX *waysx,const char *filename)
 {
  index_t i;
  int fd;
- int position=0;
+ index_t position=0;
  WayX wayx;
  WaysFile waysfile={0};
  highways_t   highways=0;
@@ -717,24 +832,24 @@ void SaveWayList(WaysX *waysx,const char *filename)
 
  /* Re-open the files */
 
- waysx->fd=ReOpenFile(waysx->filename_tmp);
- waysx->nfd=ReOpenFile(waysx->nfilename_tmp);
+ waysx->fd=ReOpenFileBuffered(waysx->filename_tmp);
+ waysx->nfd=ReOpenFileBuffered(waysx->nfilename_tmp);
 
  /* Write out the ways data */
 
- fd=OpenFileNew(filename);
+ fd=OpenFileBufferedNew(filename);
 
- SeekFile(fd,sizeof(WaysFile));
+ SeekFileBuffered(fd,sizeof(WaysFile));
 
  for(i=0;i<waysx->number;i++)
    {
-    ReadFile(waysx->fd,&wayx,sizeof(WayX));
+    ReadFileBuffered(waysx->fd,&wayx,sizeof(WayX));
 
     highways|=HIGHWAYS(wayx.way.type);
     allow   |=wayx.way.allow;
     props   |=wayx.way.props;
 
-    WriteFile(fd,&wayx.way,sizeof(Way));
+    WriteFileBuffered(fd,&wayx.way,sizeof(Way));
 
     if(!((i+1)%1000))
        printf_middle("Writing Ways: Ways=%"Pindex_t,i+1);
@@ -742,27 +857,27 @@ void SaveWayList(WaysX *waysx,const char *filename)
 
  /* Write out the ways names */
 
- SeekFile(fd,sizeof(WaysFile)+(off_t)waysx->number*sizeof(Way));
+ SeekFileBuffered(fd,sizeof(WaysFile)+(off_t)waysx->number*sizeof(Way));
 
  while(position<waysx->nlength)
    {
-    int len=1024;
+    size_t len=1024;
     char temp[1024];
 
     if((waysx->nlength-position)<1024)
        len=waysx->nlength-position;
 
-    ReadFile(waysx->nfd,temp,len);
+    ReadFileBuffered(waysx->nfd,temp,len);
 
-    WriteFile(fd,temp,len);
+    WriteFileBuffered(fd,temp,len);
 
     position+=len;
    }
 
  /* Close the files */
 
- waysx->fd=CloseFile(waysx->fd);
- waysx->nfd=CloseFile(waysx->nfd);
+ waysx->fd=CloseFileBuffered(waysx->fd);
+ waysx->nfd=CloseFileBuffered(waysx->nfd);
 
  /* Write out the header structure */
 
@@ -772,10 +887,10 @@ void SaveWayList(WaysX *waysx,const char *filename)
  waysfile.allow   =allow;
  waysfile.props   =props;
 
- SeekFile(fd,0);
- WriteFile(fd,&waysfile,sizeof(WaysFile));
+ SeekFileBuffered(fd,0);
+ WriteFileBuffered(fd,&waysfile,sizeof(WaysFile));
 
- CloseFile(fd);
+ CloseFileBuffered(fd);
 
  /* Print the final message */
 
