@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2012 Andrew M. Bishop
+ This file Copyright 2008-2013 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -30,6 +30,7 @@
 #include "typesx.h"
 #include "ways.h"
 
+#include "cache.h"
 #include "files.h"
 
 
@@ -54,6 +55,9 @@ struct _WaysX
  int      fd;                   /*+ The file descriptor of the open file (for the WaysX). +*/
 
  index_t  number;               /*+ The number of extended ways still being considered. +*/
+ index_t  knumber;              /*+ The number of extended ways kept for next time. +*/
+
+ transports_t allow;            /*+ The types of traffic that were seen when parsing. +*/
 
 #if !SLIM
 
@@ -64,9 +68,12 @@ struct _WaysX
  WayX     cached[3];            /*+ Three cached extended ways read from the file in slim mode. +*/
  index_t  incache[3];           /*+ The indexes of the cached extended ways. +*/
 
+ WayXCache *cache;              /*+ A RAM cache of extended ways read from the file. +*/
+
 #endif
 
  way_t   *idata;                /*+ The extended way IDs (sorted by ID). +*/
+ off_t   *odata;                /*+ The offset of the way in the file (used for error log). +*/
 
  index_t *cdata;                /*+ The compacted way IDs (same order as sorted ways). +*/
 
@@ -84,14 +91,16 @@ struct _WaysX
 WaysX *NewWayList(int append,int readonly);
 void FreeWayList(WaysX *waysx,int keep);
 
-void AppendWayList(WaysX *waysx,way_t id,Way *way,const char *name);
+void AppendWayList(WaysX *waysx,way_t id,Way *way,node_t *nodes,int nnodes,const char *name);
 void FinishWayList(WaysX *waysx);
 
 index_t IndexWayX(WaysX *waysx,way_t id);
 
 void SortWayList(WaysX *waysx);
 
-void ExtractWayNames(WaysX *waysx,int keep);
+SegmentsX *SplitWays(WaysX *waysx,NodesX *nodesx,int keep);
+
+void SortWayNames(WaysX *waysx);
 
 void CompactWayList(WaysX *waysx,SegmentsX *segmentsx);
 
@@ -104,13 +113,31 @@ void SaveWayList(WaysX *waysx,const char *filename);
 
 #define LookupWayX(waysx,index,position)  &(waysx)->data[index]
   
-#define PutBackWayX(waysx,wayx)           /* nop */
+#define PutBackWayX(waysx,wayx)           while(0) { /* nop */ }
 
 #else
 
-static WayX *LookupWayX(WaysX *waysx,index_t index,int position);
+/* Prototypes */
 
-static void PutBackWayX(WaysX *waysx,WayX *wayx);
+static inline WayX *LookupWayX(WaysX *waysx,index_t index,int position);
+
+static inline void PutBackWayX(WaysX *waysx,WayX *wayx);
+
+CACHE_NEWCACHE_PROTO(WayX)
+CACHE_DELETECACHE_PROTO(WayX)
+CACHE_FETCHCACHE_PROTO(WayX)
+CACHE_REPLACECACHE_PROTO(WayX)
+CACHE_INVALIDATECACHE_PROTO(WayX)
+
+
+/* Inline functions */
+
+CACHE_STRUCTURE(WayX)
+CACHE_NEWCACHE(WayX)
+CACHE_DELETECACHE(WayX)
+CACHE_FETCHCACHE(WayX)
+CACHE_REPLACECACHE(WayX)
+CACHE_INVALIDATECACHE(WayX)
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -127,7 +154,7 @@ static void PutBackWayX(WaysX *waysx,WayX *wayx);
 
 static inline WayX *LookupWayX(WaysX *waysx,index_t index,int position)
 {
- SeekReadFile(waysx->fd,&waysx->cached[position-1],sizeof(WayX),(off_t)index*sizeof(WayX));
+ waysx->cached[position-1]=*FetchCachedWayX(waysx->cache,index,waysx->fd,0);
 
  waysx->incache[position-1]=index;
 
@@ -147,7 +174,7 @@ static inline void PutBackWayX(WaysX *waysx,WayX *wayx)
 {
  int position1=wayx-&waysx->cached[0];
 
- SeekWriteFile(waysx->fd,&waysx->cached[position1],sizeof(WayX),(off_t)waysx->incache[position1]*sizeof(WayX));
+ ReplaceCachedWayX(waysx->cache,wayx,waysx->incache[position1],waysx->fd,0);
 }
 
 #endif /* SLIM */
