@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2010-2011 Andrew M. Bishop
+ This file Copyright 2010-2014 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -23,7 +23,8 @@
 #ifndef XMLPARSE_H
 #define XMLPARSE_H    /*+ To stop multiple inclusions. +*/
 
-#include <stdio.h>
+#include <stdint.h>
+#include <inttypes.h>
 
 
 /*+ The maximum number of attributes per tag. +*/
@@ -44,7 +45,7 @@ typedef struct _xmltag xmltag;
 /*+ A structure to hold the definition of a tag. +*/
 struct _xmltag
 {
- char *name;                            /*+ The name of the tag. +*/
+ char *name;                            /*+ The name of the tag - must be in lower case. +*/
 
  int  nattributes;                      /*+ The number of valid attributes for the tag. +*/
  char *attributes[XMLPARSE_MAX_ATTRS];  /*+ The valid attributes for the tag. +*/
@@ -68,9 +69,9 @@ struct _xmltag
 
 /* XML parser functions */
 
-int ParseXML(FILE *file,xmltag **tags,int options);
+int ParseXML(int fd,xmltag **tags,int options);
 
-unsigned long long ParseXML_LineNumber(void);
+uint64_t ParseXML_LineNumber(void);
 
 char *ParseXML_Decode_Entity_Ref(const char *string);
 char *ParseXML_Decode_Char_Ref(const char *string);
@@ -84,7 +85,7 @@ int ParseXML_IsFloating(const char *string);
 #define XMLPARSE_MESSAGE(tag,message) \
  do \
    { \
-    fprintf(stderr,"XML Parser: Error on line %llu: " message " in <%s> tag.\n",ParseXML_LineNumber(),tag); \
+    fprintf(stderr,"XML Parser: Error on line %" PRIu64 ": " message " in <%s> tag.\n",ParseXML_LineNumber(),tag); \
     return(1); \
    } \
     while(0)
@@ -92,7 +93,7 @@ int ParseXML_IsFloating(const char *string);
 #define XMLPARSE_INVALID(tag,attribute) \
  do \
    { \
-    fprintf(stderr,"XML Parser: Error on line %llu: Invalid value for '" #attribute "' attribute in <%s> tag.\n",ParseXML_LineNumber(),tag); \
+    fprintf(stderr,"XML Parser: Error on line %" PRIu64 ": Invalid value for '" #attribute "' attribute in <%s> tag.\n",ParseXML_LineNumber(),tag); \
     return(1); \
    } \
     while(0)
@@ -102,7 +103,7 @@ int ParseXML_IsFloating(const char *string);
    { \
     if(!attribute) \
       { \
-       fprintf(stderr,"XML Parser: Error on line %llu: '" #attribute "' attribute must be specified in <%s> tag.\n",ParseXML_LineNumber(),tag); \
+       fprintf(stderr,"XML Parser: Error on line %" PRIu64 ": '" #attribute "' attribute must be specified in <%s> tag.\n",ParseXML_LineNumber(),tag); \
        return(1); \
       } \
    } \
@@ -113,7 +114,7 @@ int ParseXML_IsFloating(const char *string);
    { \
     if(!attribute || !*attribute || !ParseXML_IsInteger(attribute)) \
       { \
-       fprintf(stderr,"XML Parser: Error on line %llu: '" #attribute "' attribute must be a integer in <%s> tag.\n",ParseXML_LineNumber(),tag); \
+       fprintf(stderr,"XML Parser: Error on line %" PRIu64 ": '" #attribute "' attribute must be a integer in <%s> tag.\n",ParseXML_LineNumber(),tag); \
        return(1); \
       } \
    } \
@@ -124,7 +125,7 @@ int ParseXML_IsFloating(const char *string);
    { \
     if(!attribute || !*attribute || !ParseXML_IsFloating(attribute)) \
       { \
-       fprintf(stderr,"XML Parser: Error on line %llu: '" #attribute "' attribute must be a number in <%s> tag.\n",ParseXML_LineNumber(),tag); \
+       fprintf(stderr,"XML Parser: Error on line %" PRIu64 ": '" #attribute "' attribute must be a number in <%s> tag.\n",ParseXML_LineNumber(),tag); \
        return(1); \
       } \
    } \
